/**
 *
 * ho ho ho .. here's my little stinky corn cadger 
 *
 * 'so what will we do today, brain' ?
 * '... the same thing we do every day, stinky. we try to take over the world!'
 * '.. har har har ...'
 *
 *
 * @author Martin Hoerning (mh@skyrix.com)
 * 
 */

package corncadger;

import corncadger.move.HamsterMover;

// that's all i need:
import algds.c2002.Hamster2;
import algds.c2002.Constants;

import java.util.Date;

/**
 * Describe class <code>CornCadger</code> here.
 * 
 * this is my hamster. its a sublass of Hamster2 to support flat and extended
 * mazes.
 *
 *
 * @author <a href="mailto:mh@skyrix.com?subject=CornCadger">Martin Hoerning</a>
 * @version 0.4.16
 * @date 2002-04-28
 * @see <a href="http://www.cs.uni-magdeburg.de/~gsemrau/h2002/index.html">Hamsterwettbewerb 2002</a>
 *
 */

public class CornCadger extends Hamster2 {

    // constants holding the version    
    private int MAJOR_VERSION    = 0;
    private int MINOR_VERSION    = 4;
    private int SUBMINOR_VERSION = 16;

    // Date to prevent a timeout exception
    private Date start;

    // the Map storing all information to navigate
    private Map map;

    // overwriting accessors
    public String getHamsterAuthor() {
        return "Martin Hoerning";
    }
    public String getHamsterVersion() {
        return MAJOR_VERSION+"."+MINOR_VERSION+"."+SUBMINOR_VERSION;
    }
    public String getHamsterDate() {
        return "2002-04-28";
    }
    
    // accessors
    private void setDirection(int _dir) {
        this.map.setCurrentDirection(_dir);
    }
    public int direction() {
        return this.map.currentDirection();
    }
    
    /* initialize the hamster values */
    public void prepareForLunch() {
        this.map   = new Map();
        this.start = new Date();
    }
    
    /* run ??? */
    public void run() {
        this.prepareForLunch();
        this.toddle(); // and nothing more
    }

    /* search the next path to go */
    private Path nextPathToGo() {
        boolean wantCorn = (this.getCarriedCorn() != Constants.CORNCARRY_MAX);
        return map.nextPathToExplore(wantCorn);
    }
    /* search path home */
    private Path pathHome() {
        return this.map.pathHome();
    }

    /* move */
    // the hamster mover for exploration
    private HamsterMover exploreMover = null;
    private void followPathExplore(Path _p) { 
        if (this.exploreMover == null) 
            this.exploreMover = HamsterMover.exploreMover(this, this.map);
        this.exploreMover.moveHamster(_p);
    }

    // the hamster mover for home moving
    private HamsterMover homeMover = null;
    private void followPathHome(Path _p) {
        if (this.homeMover == null) 
            this.homeMover = HamsterMover.homeMover(this, this.map);
        this.homeMover.moveHamster(_p);
    }

    // the hamster mover to move without stop
    private HamsterMover withoutStopMover = null;
    private void followPathWithoutStop(Path _p) {
        if (this.withoutStopMover == null) 
            this.withoutStopMover = 
                HamsterMover.withoutStopMover(this, this.map);
        this.withoutStopMover.moveHamster(_p);
    }

    /* sniff and do all u can in a field */
    private void takeAllUCanGet() {
        this.map.explore(this);
        if (!this.atHome())
            // if not at home take as much as u can carry
            this.take(Constants.CORNCARRY_MAX);
    }
    // is my mouth full or can i still put some corn in?
    private boolean gotEnough() {
        return (this.getCarriedCorn() == Constants.CORNCARRY_MAX);
    }
    // store food on home field
    public void storeFood() {
        Field here = this.map.currentField();
        if (!here.isHome()) {
            this.logln("I'm not at home!! I will not drop the corn!!");
            return;
        }
        this.dropHere(this.getCarriedCorn());
    }

    /* check the time to prevent a timeout */
    static long border = 1000 * 60 * 90; // 1000 * secs * mins == 1 1/2 hour
    public void checkTime() {
        long start = this.start.getTime();
        long time  = (new Date()).getTime();
        long div = time - start;
        if (div > border) {
            if (this.map.deepSearch()) this.map.setDeepSearch(false);
        }
        else if (!this.map.deepSearch()) {
            if (this.verbose())
                this.logln("Not much time left! switchin' to faster mode!");
            this.map.setDeepSearch(true);
        }
    }

    /* toddle !!! (ease down !!) */
    public void toddle() {
        // first explore the home field
        this.map.explore(this);

        // now the loop 
        collectLoop:
        while (true) {
            // get next path to explore
            Path p = this.nextPathToGo();
            if (p == null) {
                // didn't find another path to explore
                if (!this.map.isThereExplorePathFromHome())
                    // if there is realy nothing more to explore .. break loop
                    break collectLoop;
                // else try again with a full explore
                boolean wantCorn = 
                    (this.getCarriedCorn() != Constants.CORNCARRY_MAX);
                p = this.map.nextPathForFullExplore(wantCorn);
            }
            // no i got a path .. follow it
            this.followPathExplore(p);
            // and take all u can get
            this.takeAllUCanGet();
            //this.checkTime(); // not realy needed

            goHomeLoop:
            while (this.gotEnough()) {
                // if i got enough go home
                p = this.pathHome();
                if (p == null) {
                    // didn't find path
                    if (this.verbose())
                        this.logln("wanted to go back home, "+
                                   "but didn't find way "+
                                   "... still searching");
                    while (p == null) {
                        // trying a full search .. there must be a path
                        p = this.map.nextPathForFullExplore(false);
                        if (p == null)
                            // didn't find .. back to explore without 
                            // wanting corn
                            break goHomeLoop;
                        if (p.hasCliff(this.map)) {
                            // if path has cliff .. use home mover
                            // to drop the corn at cliff 
                            this.followPathHome(p);
                            // if could drop corn go back to exploring
                            if (!this.gotEnough()) break goHomeLoop;
                        }
                        else {
                            // has no cliffs .. take explore mover
                            // to explore a path home
                            this.followPathExplore(p);
                        }
                        // no look if there is a path home now
                        p = this.pathHome();
                    }
                }
                // finaly got a path home .. so take it 
                this.followPathHome(p);
                // if i reached home drop corn
                if (this.atHome())
                    this.storeFood();
            }
        } // endof: collectLoop
        // if im not at home, go home
        wayHome:
        while (!this.atHome()) {
            Path p = this.pathHome();
            if (p == null) {
                // there must be a way
                while (p == null) {
                    // again full explore
                    p = this.map.nextPathForFullExplore(false);
                    if (p == null) {
                        this.logln("wanted to go back home, "+
                                   "but didn't find way ");
                        break wayHome;
                    }
                    // just want to go home .. no more stops
                    this.followPathWithoutStop(p);
                    p = this.pathHome();
                }
            }
            // no more stops ...
            this.followPathWithoutStop(p);
        }
        // if i finaly reached home .. drop corn
        if (this.atHome())
            this.storeFood();
    }

    // overwriting .. yipie!!
    // the forward method .. to save the position in the map
    public boolean forward() {
        int wall = this.map.wallAt(this.direction());
        if (wall == Field.UNKNOWN)
            this.logln("forward() "+
                       "I won't go into the dark!! (wall unknown)");
        else if (wall == Field.WALL)
            this.logln("forward() "+
                       "Hey!! U'd like me to bang my head??? "+
                       "(there's a wall)");
        else if (wall == Field.CLIFFUP)
            this.logln("forward() "+
                       "Pahh!! Do I look like an ape?? (there's a cliffup)");
        else if (this.map.neighborAt(this.direction()) == null)
            this.logln("forward() "+
                       "Sorry, but I don't have a map for the next field! "+
                       "Please explore it!");
        else {
            // ready to move
            if (super.forward()) {
                if (!this.map.move(this.direction()))
                    this.logln("forward() "+
                               "RED ALERT!!!!! failed to move my finger on "+
                               "the map!! Now I'm in a jam, boy!");
                return true;
            }
            else {
                this.logln(" *%@%#^#@#^ .. hmpf .. just kissed the wall!!");
                // so just turned left
                this.setDirection((this.direction() == 0) 
                                  ? 5 : this.direction() - 1);
                return false;
            }
        }
        return false;
    }

    // the turn method .. to save the hamsters direction
    public void turn(int _dir) {
        int dir = this.direction();
        if (_dir == Constants.TURN_LEFT)
            this.setDirection((dir == 0) ? 5 : dir - 1);
        else if (_dir == Constants.TURN_RIGHT)
            this.setDirection((dir == 5) ? 0 : dir + 1);
        else {
            this.logln("turn(int) WARNING! Unknown direction");
            return;
        }
        super.turn(_dir);
    }

    // a help method to use the turn method more comfortable
    public void turnToDirection(int _dir) {
        int div = this.direction() - _dir;
        if (div < -3) div += 6;
        if (div > 3)  div -= 6;
        if (div > 0) 
            for (int i = 0; i < div; i++)
                this.turn(Constants.TURN_LEFT);
        else if (div < 0)
            for (int i = 0; i > div; i--)
                this.turn(Constants.TURN_RIGHT);
    }

    /* corn */
    // overwriting take .. to save the corn amounts
    public int take(int _n) {
        Field here      = this.map.currentField();
        int   available = here.corn();
        int   toTake    = Constants.CORNCARRY_MAX - this.getCarriedCorn();
        if (_n < toTake) toTake = _n;

        if (available == Field.UNKNOWN) {
            this.logln("take(int) unknown number of corns .. please explore");
            return 0;
        }
        else if (available < toTake)
            toTake = available;
        if (toTake == 0) return 0;
        int took = super.take(toTake);
        if (took != toTake) 
            this.logln("take(int) took unexected amount ("+
                       toTake+" vs. "+took+")");
        here.setCorn(available - took);
        this.checkNeededCorn();
        return took;
    }

    // overwriting .. to save the corn amounts
    public int dropHere(int _n) {
        Field here  = this.map.currentField();
        int toDrop  = (_n < this.getCarriedCorn()) 
            ? _n : this.getCarriedCorn();
        int dropped = super.dropHere(toDrop);
        int old     = here.corn();
        if (dropped < toDrop) // field is full
            here.setCorn(Constants.CORNFIELD_MAX);
        else if (old != Field.UNKNOWN)
            here.setCorn(old+dropped);
        else if (dropped > 0)
            here.setHasCorn(true);
        this.checkNeededCorn();
        return dropped;
    }

    // overwriting .. to save corn amounts
    public int dropForward(int _n) {
        Field forward  = this.map.neighborAt(this.direction());
        int toDrop     = (_n < this.getCarriedCorn()) 
            ? _n : this.getCarriedCorn();
        int dropped = super.dropForward(toDrop);
        int old     = forward.corn();
        if (dropped < toDrop) // field is full
            forward.setCorn(Constants.CORNFIELD_MAX);
        else if (old != Field.UNKNOWN)
            forward.setCorn(old+dropped);
        else if (dropped > 0)
            forward.setHasCorn(true);
        this.checkNeededCorn();
        return dropped;
    }

    // reset hamster needs
    private void checkNeededCorn() {
        int need = Constants.CORNCARRY_MAX - this.getCarriedCorn();
        this.map.hamsterNeeds().setCornNeeded(need);
    }

}
