
package corncadger;

import algds.c2002.*;

public class Field {

    private int x, y;    // startfelder
    private int height;  // hoehe relativ zum startfeld

    private boolean isHome;  // ist das startfeld

    // corn:
    private             int hasCorn; // hat das feld corn
    public static final int HASNOCORN = 0;
    public static final int HASCORN   = 1;
    public static final int UNKNOWN   = -100;
    private int corn;    // number of corn
    // this is the distance for the direct way from home to this field
    // so for home field it is 0
    private int radius = -1;

    // waende:
    /*          =N==
     *    NW  //    \\  NE
     *       //      \\
     *        \      /
     *.other.. \____/ .. fields ...
     *
     */
    private int wallNW;    // wand im nw
    private int wallN;     // wand im n
    private int wallNE;    // wand im no
    // is it a cliff in the view from the field ...
    public static final int CLIFFUP   = 1;
    public static final int CLIFFDOWN = 2;
    public static final int SLOPEUP   = 3;
    public static final int SLOPEDOWN = 4;
    public static final int WALL      = 5;
    public static final int WAY       = 0; // a normal way

    private boolean watched = false;
    private int     nextToHome;

    private int     crawlID = 0;

    /* constructor */
    public Field(int _x, int _y) {
        this.x = _x;
        this.y = _y;

        // first is know nothing .. buhh!!
        this.height     = Field.UNKNOWN;
        this.wallNW     = Field.UNKNOWN;
        this.wallN      = Field.UNKNOWN;
        this.wallNE     = Field.UNKNOWN;
        this.hasCorn    = Field.UNKNOWN;
        this.corn       = Field.UNKNOWN;
        this.isHome     = false;
        this.nextToHome = Field.UNKNOWN;
    }
    /* accessors */
    // setting the crawlID during search/crawl
    public void setCrawlID(int _crawlID) {
        this.crawlID = _crawlID;
    }
    public int crawlID() {
        return this.crawlID;
    }
    // setting the relative height
    public void setHeight(int _height) {
        this.height = _height;
    }
    public int height() {
        return this.height;
    }

    // setting the radius of the circle the field is on
    public void setRadius(int _r) {
        if (this.radius == -1)
            this.radius = _r;
    }
    public int radius() {
        return this.radius;
    }

    // watched
    public void setIsWatched() {
        this.watched = true;
    }
    public boolean watched() {
        return this.watched;
    }

    // home
    public void setIsHome() {
        this.isHome = true;
        this.height = 0;
    }
    public boolean isHome() {
        return this.isHome;
    }

    // next to home?
    // but in which direction
    public int nextToHome() {
        if (this.nextToHome == Field.UNKNOWN) {
            this.nextToHome = -1;
            if (this.radius != 1)
                return this.nextToHome;
            // home is 14, 59
            for (int i = 0; i < 6; i++) {
                int nx = this._xCoordForNeighbor(i);
                int ny = this._yCoordForNeighbor(i);
                if (nx == 14 && ny == 59) {
                    return (this.nextToHome = i);
                }
            }
        }
        return this.nextToHome;
    }

    // the corn
    public void setHasCorn(boolean _corn) {
        if (_corn) this.hasCorn = Field.HASCORN;
        else {
            this.hasCorn = Field.HASNOCORN;
            this.corn = 0;
        }
    }
    public int hasCorn() {
        return this.hasCorn;
    }
    public void setCorn(int _corn) {
        this.setHasCorn((_corn > 0));
        this.corn = (_corn < 0) ? 0 : _corn;
    }
    public int corn() {
        return this.corn;
    }

    // the walls
    public void setWallNW(int _wall) {
        this.wallNW = _wall;
    }
    public int wallNW() {
        return this.wallNW;
    }
    public void setWallN(int _wall) {
        this.wallN = _wall;
    }
    public int wallN() {
        return this.wallN;
    }
    public void setWallNE(int _wall) {
        this.wallNE = _wall;
    }
    public int wallNE() {
        return this.wallNE;
    }


    // getting the coords of neighbors in the array
    static int _xdivs1[] = {0, 1, 1, 0,  0,  0};
    static int _xdivs2[] = {0, 0, 0, 0, -1, -1};
    int _xCoordForNeighbor(int _dir) {
        if (_dir < 0 || _dir > 5) return Field.UNKNOWN;
        int tmp = ((this.y & 1) == 1)
            ? this.x + _xdivs1[_dir]
            : this.x + _xdivs2[_dir];
        return ((tmp < 0) || (tmp > 29)) ? Field.UNKNOWN : tmp;
    }
    static int _ydivs[] = {-2, -1, 1, 2,  1,  -1};
    int _yCoordForNeighbor(int _dir) {
        if (_dir < 0 || _dir > 5) return Field.UNKNOWN;
        int tmp = this.y + _ydivs[_dir];
        return ((tmp < 0) || (tmp > 119)) ? Field.UNKNOWN : tmp;
    }

    public int _x() { return this.x; }
    public int _y() { return this.y; }

    // overwrinting 
    public boolean equals(Object _other) {
        Field f = (Field)_other;
        if (f == this)        return true;
        if (this.x != f._x()) return false;
        if (this.y != f._y()) return false;
        return true;
    }

    public String toString() {
        return "Field: [ "+this.x+" ; "+this.y+" ]";
    }


    /*
     * this is for the fixed borders at a max maze with 30 fields radius
     */

    public static final int LEFT       = 1;
    public static final int UPPERLEFT  = 2;
    public static final int UPPERRIGHT = 4;
    public static final int RIGHT      = 8;
    public static final int LOWERRIGHT = 16;
    public static final int LOWERLEFT  = 32;

    /*
     *  bits:
     *  0 LEFT
     *  1 UPPERLEFT
     *  2 UPPERRIGHT
     *  3 RIGHT
     *  4 LOWERRIGHT
     *  5 LOWERLEFT
     *
     *  possibilities
     *  top:        means UPPERLEFT  and UPPERRIGHT --> 6
     *  upperright: means UPPERRIGHT                --> 4
     *  upperright-corner:
     *              means UPPERRIGHT and RIGHT      --> 
     *
     *
     */

    /*
     *  1   2   4   8   16   32

     private static final int[][] _borderCheck = {
     { 1, 1, 0, 0, 0, 1 }, // UPPERLEFT+UPPERRIGHT  (0)  == 35
     { 1, 1, 0, 0, 0, 0 }, // UPPERRIGHT            (1)  ==  3
     { 1, 1, 1, 0, 0, 0 }, // UPPERRIGHT+RIGHT      (2)  ==  7
     { 0, 1, 1, 0, 0, 0 }, // RIGHT                 (3)  ==  6
     { 0, 1, 1, 1, 0, 0 }, // RIGHT+LOWERRIGHT      (4)  == 14
     { 0, 0, 1, 1, 0, 0 }, // LOWERRIGHT            (5)  == 12
     { 0, 0, 1, 1, 1, 0 }, // LOWERRIGHT+LOWERLEFT  (6)  == 28
     { 0, 0, 0, 1, 1, 0 }, // LOWERLEFT             (7)  == 24
     { 0, 0, 0, 1, 1, 1 }, // LOWERLEFT+LEFT        (8)  == 56
     { 0, 0, 0, 0, 1, 1 }, // LEFT                  (9)  == 48
     { 1, 0, 0, 0, 1, 1 }, // LEFT+UPPERLEFT        (10) == 49
     { 1, 0, 0, 0, 0, 1 }, // UPPERLEFT             (11) == 33
     { 0, 0, 0, 0, 0, 0 }  // <nothing>             (12) ==  0
     };
    */

    private int     myBorders;
    private boolean bordersChecked = false;

    public int borders() {
        if (this.bordersChecked)
            return this.myBorders;
        this.myBorders = this._computeBorders();
        this.bordersChecked = true;
        return this.myBorders;
    }

    public boolean isMaxBorder(int _dir) {
        int check = this.borders() >> _dir;
        return ((check & 1) == 1);
    }
    
    private int _computeBorders() {
        if (this.radius != 29) return 0;

        // some fixed points
        if (this.x == 14) {
            if (this.y == 1)
                // UPPERLEFT | UPPERRIGHT;
                return 35;
            if (this.y == 117)
                // LOWERLEFT | LOWERRIGHT;
                return 28;
            if ((this.y & 1) == 1)
                return 0;
        }

        if (this.x == 0) {
            if (this.y == 30)
                // LEFT | UPPERLEFT;
                return 49;

            if (this.y == 88)
                // LEFT | LOWERLEFT;
                return 56;

            if ((this.y & 1) == 0)
                // LEFT
                return 48;
        }

        if (this.x == 29) {
            if (this.y == 30)
                // RIGHT | UPPERRIGHT;
                return 7;

            if (this.y == 88)
                // RIGHT | LOWERRIGHT;
                return 14;

            if ((this.y & 1) == 0)
                // RIGHT;
                return 6;
        }

        // now the diagonale
        if ((this.y & 1) == 1) {
            // y & 2 would be 1

            if ((15 - ((this.y+1)/2)) == this.x) 
                // the upper left border
                // UPPERLEFT;
                return 33;

            if ((this.x - 14) == ((this.y-1)/2))
                // UPPERRIGHT;
                return 3;

            if ((this.x + 44) == ((this.y-1)/2))
                // LOWERLEFT;
                return 24;

            if ((72 - this.x) == ((this.y-1)/2))
                // LOWERRIGHT;
                return 12;

            return 0;
        }

        // y % 2 would be 0
        if ((15 - (this.y/2)) == this.x)
            // the upper left
            // UPPERLEFT;
            return 33;

        if ((this.x - 14) == ((this.y)/2))
            // UPPERRIGHT;
            return 3;

        if ((this.x + 44) == ((this.y)/2))
            // LOWERLEFT;
            return 24;

        if ((72 - this.x) == ((this.y-2)/2))
            // LOWERRIGHT;
            return 12;

        return 0;
    }


    /*
      public double distanceValueToHomeField() {
      // home is 14, 59
      double xdiv, ydiv;
      
      if (this.x > 14)
      xdiv = (this.x - 14.0) * 2.0 - 1.0 + (this.y & 1);
      else 
      xdiv = (14.0 - this.x) * 2.0 + 1.0 - (this.y & 1);
      ydiv = (this.y < 59) ? (59.0 - this.y) : (this.y - 59.0);
      
      return xdiv*xdiv*1.0 + ydiv*ydiv*0.25;
      }
    */
}
