
package corncadger;

import corncadger.search.PathSearch;
import corncadger.search.PathCache;
import corncadger.search.MapInit;
import corncadger.search.LockAround;

import algds.c2002.Constants;
import java.util.Vector;
import java.util.Enumeration;

public class Map {

    // the field array        
    private Field fields[][] = new Field[30][120]; // the maximum size of map

    // the current field
    private Field current    = null;
    // the current direction of the hamster
    private int   currentDir = 0;
    // the home field
    private Field home       = null;
    
    // whether to perform a deep search
    private boolean deepSearch = true;

    // caching paths
    private PathCache pathCache = null;
    private Path      todoPath  = null;

    // the hamster needs
    private HamsterNeeds hamsterNeeds = null;
    
    // initialize the map
    public Map() {
        this._initFields();
        this.hamsterNeeds = new HamsterNeeds();
    }
    
    // accessors
    public void  setCurrentDirection(int _dir) { this.currentDir = _dir; }
    public int   currentDirection() { return this.currentDir; }
    public Field currentField()     { return this.current;    }
    public Field homeField()        { return this.home;       }
    
    public void setDeepSearch(boolean _flag) { this.deepSearch = _flag; }
    public boolean deepSearch()              { return this.deepSearch;  }

    public HamsterNeeds hamsterNeeds() { return this.hamsterNeeds; }
    
    /* accessors */
    /**
     * Describe <code>isFieldWatched</code> method here.
     *
     * a field is explored, when u know wether there is corn and 
     * u know the height
     *
     * @param _field a <code>Field</code> value
     * @return true, when explored, false when not
     */
    public boolean isFieldWatched(Field _field) {
        // its watched when u know the height and if there is corn
        return _field.watched();
        //if (_field.hasCorn() == Field.UNKNOWN) return false;
        //if (_field.height()  == Field.UNKNOWN) return false;
        //return true;
    }
    public boolean isFieldWatchedOverWall(Field _field, int _wall) {
        // if already watched from somewhere else .. it's actualy watched
        if ((_field != null) && (_field.watched())) return true;
        // if wall unknown .. unwatched
        if (_wall == Field.UNKNOWN) return false;
        // if wall .. nothing to see .. so watched
        if (_wall == Field.WALL)    return true; 
        // 'cause nothing to see over a wall

        // no instance exsistant .. so unwatched
        if (_field == null) return false;
        if ((_wall == Field.CLIFFUP) || (_wall == Field.SLOPEUP)) {
            // can only see the height
            if (_field.height()  == Field.UNKNOWN) return false;
            return true;
        }
        return this.isFieldWatched(_field);
    }

    public boolean isFieldExplored(Field _field) {
        if (!this.isFieldWatched(_field)) return false;
        // ok .. i know the corn, i know the height .. 
        // .. but wait .. what about my neigbors .. lets take a look
        return this.areNeighborsWatched(_field);
    }
    // for full explored it must be explored and all walls must be known
    public boolean isFieldFullExplored(Field _field) {
        if (!this.isFieldExplored(_field)) return false;
        return (this.numberOfKnownWalls(_field) == 6);
    }

    public boolean areNeighborsExplored() {
        return areNeighborsExplored(this.current);
    }
    // take a look around .. 
    public boolean areNeighborsExplored(Field _f) {
        return (new LockAround(this, _f)).allExplored();
    }
    

    /**
     * Describe <code>isCurrentFieldExplored</code> method here.
     *
     * did i see all i can get??
     *
     * @return a <code>boolean</code> value
     */
    public boolean isCurrentFieldExplored() {
        return this.isFieldExplored(this.current);
    }
    /**
     * Describe <code>areNeighborsWatched</code> method here.
     *
     * a neighbor is explored, if i know wether there is corn and
     * if i know the height
     * a neighbor only needs to be explored if it is reachable
     * (so if there is no wall or cliffup)
     *
     * @param _field a <code>Field</code> value
     * @return a <code>boolean</code> value
     */
    public boolean areNeighborsWatched(Field _field) {
        for (int i = 0; i < 6; i++) {
            int   wall = this.wallAt(_field, i);
            Field n    = neighborAt(_field, i);
            if (!this.isFieldWatchedOverWall(n, wall)) return false;
            // ok .. neighbor is checked
        }
        return true;
    }

    // well .. what u see is wat u get
    public int numberOfWatchedNeighbors(Field _field) {
        int count = 0;
        for (int i = 0; i < 6; i++) {
            if (!neighborValidAt(_field, i)) continue;
            int   wall = this.wallAt(_field, i);
            Field n    = neighborAt(_field, i);
            if (this.isFieldWatchedOverWall(n, wall))
                count++;
        }
        return count;
    }
    // humm .. the same i think .. is it?
    public int numberOfKnownWalls(Field _field) {
        int count = 0;
        turnAround:
        for (int i = 0; i < 6; i++) {
            int   wall = this.wallAt(_field, i);
            if (wall == Field.UNKNOWN) continue turnAround;
            count++;
        }
        return count;
    }

    /**
     * Describe <code>neighborAt</code> method here.
     *
     * the directions:
     *     0
     *     _
     * 5  / \  1
     * 4  \_/  2
     *     3
     *
     * if unknown or invalid return null;
     *
     * @param _field the field 
     * @param _dir the direction
     * @return the neighbor
     */
    public Field neighborAt(Field _field, int _dir) {
        int nX = _field._xCoordForNeighbor(_dir);
        int nY = _field._yCoordForNeighbor(_dir);
        if (!this._checkCoords(nX, nY)) return null;
        return this.fields[nX][nY];
    }
    // check wether coords are valid
    public boolean neighborValidAt(Field _field, int _dir) {
        int nX = _field._xCoordForNeighbor(_dir);
        int nY = _field._yCoordForNeighbor(_dir);
        return this._checkCoords(nX, nY);
    }
    public Field neighborAt(int _dir) {
        return this.neighborAt(this.current, _dir);
    }
    // checking coords
    private boolean _checkCoords(int _x, int _y) {
        if (_x == Field.UNKNOWN) return false;
        if (_y == Field.UNKNOWN) return false;
        return true;
    }

    // check wether a neighbor field is home
    public int isNeighborFieldHome(Field _field) {
        return _field.nextToHome();
    }
    public int isCurrentNeighborFieldHome() {
        return this.isNeighborFieldHome(this.current);
    }

    // storing wall information
    public void setWallAt(Field _field, int _dir, int _wall) {
        if      (_dir == 5) _field.setWallNW(_wall);
        else if (_dir == 0) _field.setWallN(_wall);
        else if (_dir == 1) _field.setWallNE(_wall);
        else if (!this.neighborValidAt(_field, _dir)) return; // a WALL !!!
        else if (_dir == 2) {
            Field n = this.neighborAt(_field, 2);
            if (n != null) n.setWallNW(this._reverseWall(_wall));
        }
        else if (_dir == 3) {
            Field n = this.neighborAt(_field, 3);
            if (n != null) n.setWallN(this._reverseWall(_wall));
        }
        else if (_dir == 4) {
            Field n = this.neighborAt(_field, 4);
            if (n != null) n.setWallNE(this._reverseWall(_wall));
        }
    }
    public void setWallAt(int _dir, int _wall) {
        this.setWallAt(this.current, _dir, _wall);
    }

    // accessing wall information
    public int wallAt(Field _field, int _dir) {
        if (this._wallAtMaxBorder(_field, _dir))
            return Field.WALL;
        if (_dir == 5) return _field.wallNW();
        if (_dir == 0) return _field.wallN();
        if (_dir == 1) return _field.wallNE();
        if (!this.neighborValidAt(_field, _dir)) return Field.WALL;
        if (_dir == 2) {
            Field n = this.neighborAt(_field, 2);
            return (n == null) ? Field.UNKNOWN : this._reverseWall(n.wallNW());
        }
        if (_dir == 3) {
            Field n = this.neighborAt(_field, 3);
            return (n == null) ? Field.UNKNOWN : this._reverseWall(n.wallN());
        }
        if (_dir == 4) {
            Field n = this.neighborAt(_field, 4);
            return (n == null) ? Field.UNKNOWN : this._reverseWall(n.wallNE());
        }
        return Field.UNKNOWN;
    }
    public int wallAt(int _dir) {
        return this.wallAt(this.current, _dir);
    }

    /* exploring */
    public Field createNeighborAt(int _dir) {
        return this.createNeighborAt(this.current, _dir);
    }
    public Field createNeighborAt(Field _f, int _dir) {
        int nX = _f._xCoordForNeighbor(_dir);
        int nY = _f._yCoordForNeighbor(_dir);
        if (!this._checkCoords(nX, nY)) return null;
        Field n = new Field(nX, nY);
        this.fields[nX][nY] = n;
        return n;
    }
    // well .. just explore 
    public void explore(CornCadger _cc) {
        int dir     = this.currentDirection();
        int wallCnt = 0;

        this.current.setCorn(_cc.getAvailableCorn());
        this.current.setIsWatched();

        turnAround:
        for (int i = 0; i < 6; i++) {
            Field n = neighborAt(i);
            int   w = wallAt(i);
            if (n == null)
                n = createNeighborAt(i);
            if ((w == Field.UNKNOWN) || 
                (!this.isFieldWatchedOverWall(n, w))) {
                // if field isn't watched .. turn and look at it
                _cc.turnToDirection(i);
                wallCnt++;
                int info = _cc.lookForward();
                int wall = (info & Constants.WALLMASK);
                // map the wall information
                if (wall == Constants.WALL)  
                    // there is a wall
                    this.setWallAt(i, Field.WALL);
                else if (wall == Constants.CLIFFDOWN) 
                    // there is a cliff going down
                    this.setWallAt(i, Field.CLIFFDOWN);
                else if (wall == Constants.CLIFFUP) 
                    // there is a cliff going up
                    this.setWallAt(i, Field.CLIFFUP);
                else if (wall == Constants.SLOPEDOWN) 
                    // there is a slope going down
                    this.setWallAt(i, Field.SLOPEDOWN);
                else if (wall == Constants.SLOPEUP) 
                    // there is a slope going up
                    this.setWallAt(i, Field.SLOPEUP);
                else {
                    // nothing of this .. so its just simple a way
                    this.setWallAt(i, Field.WAY);
                }
                if (n == null) 
                    // neighbor not valid at this field
                    continue turnAround; 
                w = wallAt(i);
                if (w != Field.WALL) {
                    // can see the height
                    n.setHeight(_cc.getHeightDifference()+
                                this.current.height());
                    if ((w != Field.CLIFFUP) && (w != Field.SLOPEUP)) {
                        // can see if there is corn
                        n.setHasCorn
                            ((info & Constants.CORNMASK) == Constants.CORN);
                        // it's fully watched
                        n.setIsWatched();
                    }
                }
            }
        }
        // turn back to old direction
        _cc.turnToDirection(dir);
        if (wallCnt != 0) {
            // new walls discovered -> maybe new/faster paths available
            // clear caches
            this.pathCache = null;
            this.todoPath  = null;
        }
    }

    /* moving */
    // move the finger on the map ...
    public boolean move(int _dir) {
        Field next = neighborAt(_dir);
        if (next == null) return false;
        this.current = next;
        return true;
    }

    // get's next path in full explore mode
    public Path nextPathForFullExplore(boolean _wantCorn) {
        return PathSearch.exploreSearchFull(this, _wantCorn).bestPath();
    }

    // check whether there is still corn available from home field
    public boolean isThereExplorePathFromHome() {
        return (PathSearch.exploreSearchFromHome(this, true).
                bestPath() != null);
    }
    
    /* getting all the best paths */
    public Path nextPathToExplore(boolean _wantCorn) {
        if (this.current.isHome()) {
            Path  p;
            // doing special cache stuff for explore paths from home
            // look if there is s.th. on the todo list
            p = this.todoPath;
            this.todoPath = null; // reseting todo path
            if (p == null) // todopath was null
                p = PathSearch.exploreSearch(this, _wantCorn).bestPath();
            if (p != null) {
                Field end = p.end();
                int   needed = this.hamsterNeeds.cornNeeded();
                if (end.corn() > (Constants.CORNCARRY_MAX+needed))
                    // cache it (cash it ?? :-)
                    this.todoPath = p;
            }
            return p;
        }
        Path p = PathSearch.exploreSearch(this, _wantCorn).bestPath();
        
        // reseting hamsters needs
        this.hamsterNeeds.resetHamsterNeeds(false);
        return p;
    }
    public Path pathHome() {
        Path p = null;
        if (this.pathCache != null)
            p = this.pathCache.pathBetween(this.current, this.home);
        if (p == null) {
            p = PathSearch.homeSearch(this).bestPath();
            if (p != null) {
                if (this.pathCache == null) this.pathCache = new PathCache();
                // cache it (hihi .. cash it :-)
                this.pathCache.addPath(p);
            }
        }
        return p;
    }

    /* private methods */
    // just return the wall how it is seen from the other side
    private int _reverseWall(int _wall) {
        switch (_wall) {
            case (Field.CLIFFUP):   return Field.CLIFFDOWN;
            case (Field.CLIFFDOWN): return Field.CLIFFUP;
            case (Field.SLOPEUP):   return Field.SLOPEDOWN;
            case (Field.SLOPEDOWN): return Field.SLOPEUP;
        }
        return _wall;
    }

    // init field array
    private void _initFields() {
        for (int y = 0; y < 120; y++)
            for (int x = 0; x < 30; x++)
                this.fields[x][y] = null;
        // init start field
        this.home = this.current = new Field(14, 59); // index begins at 0
        this.current.setIsHome();
        this.fields[14][59] = this.current;

        MapInit.initMap(this);
    }

    private boolean _wallAtMaxBorder(Field _field,int _dir) {
        //return (_field.borders()[_dir] == 1);
        // is it a field at a radius of 30 ?
        return _field.isMaxBorder(_dir);
    }
}
