/**
 * a path connects fields
 *
 * it has a start and an end.. so the way leeds from start to end
 *
 * so just a path
 *
 */

package corncadger;

import java.util.Vector;
import algds.c2002.Constants;

// one node in the path (which is a linked list)
class FieldNode {
    // the field
    private Field     field = null;
    // the next node
    private FieldNode next  = null;
    // the direction which leads to the next field
    private int       dir   = Field.UNKNOWN;


    // the constructors
    FieldNode(Field _field, int _dir, FieldNode _next) {
        this.field = _field;
        this.dir   = _dir;
        this.next  = _next;
    }
    FieldNode(Field _field) {
        this(_field, Field.UNKNOWN, null);
    }

    // accessors
    void setNext(int _dir, FieldNode _next) {
        this.next = _next;
        this.dir  = _dir;
    }
    FieldNode next() {
        return this.next;
    }
    Field field() {
        return this.field;
    }
    int direction() { // from 0 to 5 .. or Field.UNKNOWN
        return this.dir;
    }
}

public class Path {
    // the start and end of the path
    private FieldNode start  = null, end = null;
    // the length of the path
    private int       length = 0;

    /* some caching */
    // -1 unknown, 0 for no, 1 for yes
    private int hasCliffCache             = -1;
    private int lastDirectionCache        = -1;
    private int hasCliffAtEndCache        = -1;
    private int hasFirstCliffAtEndCache   = -1;
    private int distanceToFirstCliffCache = -2;
    private int numberOfCliffDownsCache   = -1;
    private int maxRadiusCache            = -1;
    
    private Field canDropCornOnNextCliffDownCache = null;

    // clearing the cache
    private void addedClearCache() {
        if (this.hasCliffCache == 0) this.hasCliffCache = -1;
        this.hasCliffAtEndCache      = -1;
        this.hasFirstCliffAtEndCache = -1;
        this.canDropCornOnNextCliffDownCache = null;
        this.distanceToFirstCliffCache = -2;
        this.numberOfCliffDownsCache   = -1;
        this.maxRadiusCache            = -1;
    }
    private void removedClearCache() {
        if (this.hasCliffCache == 1) this.hasCliffCache = -1;
        this.lastDirectionCache = -1;
        this.hasCliffAtEndCache = -1;

        if (this.hasFirstCliffAtEndCache == 1)
            this.hasFirstCliffAtEndCache = 0;
        else
            this.hasFirstCliffAtEndCache = -1;

        this.canDropCornOnNextCliffDownCache = null;
        this.distanceToFirstCliffCache = -2;
        this.numberOfCliffDownsCache   = -1;
        this.maxRadiusCache            = -1;
    }

    // constructor
    public Path(Field _start) {
        this.start = this.end = new FieldNode(_start);
    }

    /* manipulating */
    public void append(Field _field, int _dir) {
        FieldNode next = new FieldNode(_field);
        this.end.setNext(_dir, next);
        this.end = next;
        this.length++;
        this.addedClearCache();
        this.lastDirectionCache = _dir;
    }
    // concat to paths
    public Path concat(Path _path) {
        Path concated = (Path)this.clone();
        if (this.end() != _path.start()) {
            System.out.println("Can only concat paths with "+
                               "one field where they meet");
            return concated;
        }
        FieldNode n = _path.start;
        while (n.next() != null) {
            concated.append(n.next().field(), n.direction());
            n = n.next();
        }
        return concated;
    }
    // remove the last field
    public void removeLast() {
        FieldNode node = this.start;
        if (node == this.end) return; // dont remove the start
        while (node.next() != this.end)
            node = node.next();
        this.end = node;
        this.end.setNext(Field.UNKNOWN, null);
        this.length--;
        this.removedClearCache();
    }

    /* accessors */
    public Field start() {
        return this.start.field();
    }
    public Field end() {
        return this.end.field();
    }
    public int length() { // length is number of nodes - 1
        return this.length;
    }

    // the steps which the hamster has to take
    // a int-array of directions
    public int[] steps() {
        int       steps[] = new int[this.length];
        FieldNode n       = this.start;
        int       idx     = 0;
        while (n.next() != null) {
            steps[idx++] = n.direction();
            n = n.next();
        }
        return steps;
    }
    // returns also 0 if length == 0
    public int lastDirection() {
        if (this.lastDirectionCache != -1) {
            int dir = 0;
            FieldNode n = this.start;
            while (n.next() != null) {
                dir = n.direction();
                n = n.next();
            }
            this.lastDirectionCache = dir;
        }
        return this.lastDirectionCache;
    }

    // some accessors
    public boolean hasCliff(Map _m) {
        if (this.hasCliffCache == -1) {
            FieldNode n = this.start;
            this.hasCliffCache = 0;
            while (n.next() != null) {
                if (_m.wallAt(n.field(), n.direction()) == Field.CLIFFDOWN) {
                    this.hasCliffCache = 1;
                    return true;
                }
                n = n.next();
            }
        }
        return (this.hasCliffCache == 1);
    }
    public boolean hasCliffAtEnd(Map _m) {
        if (this.hasCliffAtEndCache == -1) {
            FieldNode n = this.start;
            int       w = Field.UNKNOWN;
            while (n.next() != null) {
                w = _m.wallAt(n.field(), n.direction());
                n = n.next();
            }
            this.hasCliffAtEndCache = (w == Field.CLIFFDOWN) ? 1 : 0;
        }
        return (this.hasCliffAtEndCache == 1);
    }
    public boolean hasFirstCliffAtEnd(Map _m) {
        if (this.hasFirstCliffAtEndCache == -1) {
            FieldNode n = this.start;
            int       w = Field.UNKNOWN;
            this.hasFirstCliffAtEndCache = 0;
            while (n.next() != null) {
                w = _m.wallAt(n.field(), n.direction());
                n = n.next();
                if (w == Field.CLIFFDOWN) {
                    this.hasFirstCliffAtEndCache = 
                        (n == this.end) ? 1 : 0;
                    return true;
                }
            }
        }
        return (this.hasFirstCliffAtEndCache == 1);
    }

    public boolean canDropCornOnNextCliffDown(Map _m) {
        if (this.canDropCornOnNextCliffDownCache == null) {
            FieldNode n = this.start;
            int       w = Field.UNKNOWN;
            fieldloop:
            while (n.next() != null) {
                w = _m.wallAt(n.field(), n.direction());
                n = n.next();
                if (w == Field.CLIFFDOWN) {
                    this.canDropCornOnNextCliffDownCache = n.field();
                    break fieldloop;
                }
            }
            return false;
        }
        Field f = this.canDropCornOnNextCliffDownCache;
        return (f.isHome() || (f.corn() != Constants.CORNFIELD_MAX));
    }

    public int distanceToFirstCliff(Map _m) {
        if (this.distanceToFirstCliffCache == -2) {
            FieldNode n   = this.start;
            int       cnt = 0;
            this.distanceToFirstCliffCache = -1;
            while (n.next() != null) {
                if (_m.wallAt(n.field(), n.direction()) == Field.CLIFFDOWN) {
                    this.distanceToFirstCliffCache = cnt;
                    return cnt;
                }
                cnt++;
                n = n.next();
            }
        }
        return this.distanceToFirstCliffCache;
    }

    public int numberOfKnownNeighbors(Map _m) {
        int       cnt = 0;
        FieldNode n   = this.start;
        while (n != null) {
            cnt += _m.numberOfWatchedNeighbors(n.field());
            n    = n.next();
        }
        return cnt;
    }
    public int numberOfKnownWalls(Map _m) {
        int       cnt = 0;
        FieldNode n   = this.start;
        while (n != null) {
            cnt += _m.numberOfKnownWalls(n.field());
            n    = n.next();
        }
        return cnt;
    }
    // -1 if no unknown walls
    public int stepsTillUnknownWalls(Map _m) {
        int       cnt = 0;
        FieldNode n   = this.start;
        while (n != null) {
            if (_m.numberOfKnownWalls(n.field()) != 6) return cnt;
            n    = n.next();
            cnt++;
        }
        return 50000; // a big number to scare the hamster :-)
    }

    public int numberOfFieldsWithKnownNumberOfCorn() {
        int       cnt = 0;
        FieldNode n   = this.start;
        while (n != null) {
            if (n.field().corn() != Field.UNKNOWN) cnt++;
            n = n.next();
        }
        return cnt;
    }

    public int numberOfCliffDowns(Map _m) {
        if (this.numberOfCliffDownsCache == -1) {
            int       cnt = 0;
            FieldNode n   = this.start;
            while (n.next() != null) {
                if (_m.wallAt(n.field(), n.direction()) == Field.CLIFFDOWN) 
                    cnt++;
                n = n.next();
            }
            this.numberOfCliffDownsCache = cnt;
        }
        return this.numberOfCliffDownsCache;
    }

    public int maxRadius() {
        if (this.maxRadiusCache == -1) {
            int       max = 0, tmp;
            FieldNode n   = this.start;
            while (n != null) {
                if ((tmp = n.field().radius()) > max) 
                    max = tmp;
                n = n.next();
            }
            this.maxRadiusCache = max;
        }
        return this.maxRadiusCache;
    }

    public boolean leadsOverHome() {
        FieldNode n = this.start;
        while (n != null) {
            if (n.field().isHome()) return true;
            n = n.next();
        }
        return false;
    }

    public boolean containsField(Field _f) {
        FieldNode n = this.start;
        while (n != null) {
            if (n.field() == _f) return true;
            n = n.next();
        }
        return false;
    }

    /* cloning */
    public Object clone() {
        Path      copy = new Path(this.start());
        FieldNode node = this.start;
        while (node.next() != null) {
            copy.append(node.next().field(), node.direction());
            node = node.next();
        }
        return copy;
    }

    // overwriting
    public boolean equals(Object _other) {
        if (_other == this) return true;
        Path other = (Path)_other;
        if (other.start() != this.start()) return false;
        int steps[]  = this.steps();
        int steps2[] = other.steps();
        if (steps.length != steps2.length) return false;
        for (int i = 0; i < steps.length; i++)
            if (steps[i] != steps2[i]) return false;
        return true;
    }

    public String toString() {
        return "Path from "+this.start()+" to "+this.end()+" with length "+
            this.length;
    }

    /* the way back */
    // and check for cliffs
    public Path reverse(boolean _ignoreCliffs, Map _m) {
        Path      p = new Path(this.start());
        FieldNode n = this.start;
        while (n.next() != null) {
            int dir = n.direction();
            if (!_ignoreCliffs) {
                if (_m.wallAt(dir) == Field.CLIFFDOWN) {
                    return null;
                }
            }
            dir = (dir > 2) ? dir - 3 : dir + 3;
            n = n.next();
            p._appendInFront(n.field(), dir);
        }
        return p;
    }
    // this is done, ignoring cliffs
    public Path reverse() {
        Path      p = new Path(this.start());
        FieldNode n = this.start;
        while (n.next() != null) {
            int dir = n.direction();
            dir = (dir < 3) ? (dir + 3) : (dir - 3);
            n = n.next();
            p._appendInFront(n.field(), dir);
        }
        return p;
    }

    /* private stuff */
    void _appendInFront(Field _field, int _dir) {
        this.start = new FieldNode(_field, _dir, this.start);
        this.length++;
        this.addedClearCache();
    }

    /* help method for map */
    public void _removeFieldsSinceCliff(Vector _selected, Map _m) {
        FieldNode n = this.start;
        boolean   c = false;
        while (n != null) {
            if (!c)
                c = (_m.wallAt(n.field(), n.direction()) == Field.CLIFFDOWN);
            else 
                _selected.removeElement(n.field());
            n = n.next();
        }
    }
}
