

package corncadger.move;

import corncadger.Path;
import corncadger.Map;
import corncadger.CornCadger;
import corncadger.Field;
import corncadger.HamsterNeeds;

public abstract class HamsterMover {
    // the map
    private Map        map     = null;
    // the hamster
    private CornCadger hamster = null;

    // the constructor
    protected HamsterMover(CornCadger _hamster, Map _map) {
        this.hamster = _hamster;
        this.map     = _map;
    }

    // accessors
    protected Map        map()     { return this.map;     }
    protected CornCadger hamster() { return this.hamster; }

    // static instances
    // for explorer
    private static HamsterMover explorer = null;
    public static HamsterMover exploreMover(CornCadger _hamster, Map _map) {
        if (explorer == null) explorer = new ExploreMover(_hamster,_map);
        return explorer;
    }
    // for home mover
    private static HamsterMover home = null;
    public static HamsterMover homeMover(CornCadger _hamster, Map _map) {
        if (home == null) home = new HomeMover(_hamster,_map);
        return home;
    }
    // for without stop mover
    private static HamsterMover withoutStop = null;
    public static HamsterMover withoutStopMover(CornCadger _hamster, 
                                                Map _map) {
        if (withoutStop == null) 
            withoutStop = new WithoutStopMover(_hamster,_map);
        return withoutStop;
    }

    /* this to be implemented in subclasses */
    public abstract void moveHamster(Path _p);

    // help method
    protected boolean storeFood(Map _m, CornCadger _c) {
        if (_m.currentField().isHome()) {   
            // can easyly drop it
            _c.storeFood();
            return true;
        }
        else {
            int dir = _m.isCurrentNeighborFieldHome();
            if (dir != -1) {
                int wall = _m.wallAt(dir);
                if (wall == Field.CLIFFDOWN) {
                    _c.turnToDirection(dir);
                    _c.dropForward(_c.getCarriedCorn());
                    return true;
                }
            }
        }
        return false;
    }

}


// subclass to explore
class ExploreMover extends HamsterMover {

    // constructor
    ExploreMover(CornCadger _hamster, Map _map) {
        super(_hamster, _map);
    }

    // implementing moveHamster
    public void moveHamster(Path _p) {
        CornCadger c          = this.hamster();
        Map        m          = this.map();
        int        steps[]    = _p.steps();
        int        max        = steps.length;
        boolean    moved      = false;
        int        knownWalls = 0;
        for (int i = 0; i < max; i++) {
            // store food if available
            if (c.getCarriedCorn() != 0) {
                if (this.storeFood(m,c))
                    // if food stored stop movin
                    return;
            }

            // turn and step forward
            c.turnToDirection(steps[i]);
            moved = c.forward();

            // get number of known walls and explore anyway
            knownWalls = m.numberOfKnownWalls(m.currentField());
            m.explore(c);

            if (knownWalls != 6) {
                if (c.getCarriedCorn() != 0) {
                    if (this.storeFood(m,c))
                        // if food stored stop movin
                        return;
                }
                // just explored new walls
                // maybe now there is a better way
                return;
            }
            if (!moved) 
                // didn't move .. better stop
                return;
        }
        if (c.getCarriedCorn() != 0) {
            if (this.storeFood(m,c))
                // if food stored stop movin
                return;
        }
    }

}


// subclass to walk without stop
class WithoutStopMover extends HamsterMover {

    // constructor
    WithoutStopMover(CornCadger _hamster, Map _map) {
        super(_hamster,_map);
    }

    // implementing moveHamster
    public void moveHamster(Path _p) {
        CornCadger c          = this.hamster();
        Map        m          = this.map();
        int        steps[]    = _p.steps();
        int        max        = steps.length;
        boolean    moved      = false;
        int        knownWalls = 0;

        for (int i = 0; i < max; i++) {
            // turn and step forward
            c.turnToDirection(steps[i]);
            moved = c.forward();

            // get number of known walls and explore anyway
            knownWalls = m.numberOfKnownWalls(m.currentField());
            m.explore(c);

            if (knownWalls != 6) 
                // just explored new walls
                // maybe now there is a better way
                return;
            if (!moved) 
                // didn't move .. better stop
                return;
        }
    }

}
