
package corncadger.move;

import corncadger.CornCadger;
import corncadger.Field;
import corncadger.Map;
import corncadger.Path;
import corncadger.search.LockAround;

import algds.c2002.Constants;

class HomeMover extends HamsterMover {

    HomeMover(CornCadger _hamster, Map _map) {
        super(_hamster,_map);
    }

    /**
     * Describe <code>checkForExplore</code> method here.
     *
     * @param _f a <code>Field</code> value
     * @return true if stop, false if move on
     */
    private boolean checkForExplore(Field _f, int _nextDir) {
        CornCadger c    = this.hamster();
        Map        m    = this.map();

        if (_f.isHome()) return false;
        
        int corn = _f.corn();
        if (corn == 0) {
            LockAround l = new LockAround(m,_f);
            if (!l.allExplored()) {
                // can drop corn on current field to continue exploring
                Path p = l.neighborToExplore(_nextDir);
                if (p != null) {
                    c.dropHere(c.getCarriedCorn());
                    HamsterMover.withoutStopMover(c,m).moveHamster(p);
                    
                    return true;
                }
            }
        }
        return false;
    }

    // implementing moveHamster
    public void moveHamster(Path _p) {
        CornCadger c          = this.hamster();
        Map        m          = this.map();
        int        steps[]    = _p.steps();
        int        max        = steps.length;
        boolean    moved      = false;
        int        knownWalls = 0;
        int        maxRadius  = _p.maxRadius();
        
        boolean    checkForExplore = (max < 99);
        
        for (int i = 0; i < max; i++) {
            c.turnToDirection(steps[i]);
            
            // check for cliffs, and if i can drop corn over it
            if (m.wallAt(c.direction()) == Field.CLIFFDOWN) {
                // yeah a cliff
                Field target = m.neighborAt(c.direction());
                int   corn   = target.corn();
                if (target.isHome() || (corn != Constants.CORNFIELD_MAX)) {
                    // can drop corn !!
                    c.dropForward(c.getCarriedCorn());
                    // stop movin
                    return;
                }
            }
            
            // check if there are unknown neighbors
            // if so, drop corn if possible and go on exploring
            // turn and step forward
            Field f = m.currentField();
            if (checkForExplore) {
                if (this.checkForExplore(f, steps[i]))
                    return;
            }
            
            moved = c.forward();

            // get number of known walls and explore anyway
            knownWalls = m.numberOfKnownWalls(m.currentField());
            m.explore(c);
            
            if (!moved) 
                // didn't move .. better stop
                return;
            
            if (c.getCarriedCorn() != 0) {
                // store food if possible
                if (this.storeFood(m,c))
                    // if food stored stop movin
                    return;
            }
            if (knownWalls != 6) 
                // just explored new walls
                // maybe now there is a better way
                return;
        }
    }
    
}
