
package corncadger.search;

import corncadger.Path;
import corncadger.Map;

class CriteriaTree {
    CriteriaTreeNode root = null;

    CriteriaTree(CriteriaTreeNode _root) {
        this.root = _root;
    }

    // won't return null, just a if they are equal
    Path betterPath(Path _a, Path _b, Map _m) {
        if (this.root == null) return _a;
        Path better = this.root.betterPath(_a,_b,_m);
        return (better == null) ? _a : better;
    }

    private static CriteriaTree explorationTree = null;
    public static CriteriaTree explorationTree() {
        if (explorationTree == null)
            explorationTree = CriteriaTree._explorationTree();
        return explorationTree;
    }

    // building the criteria tree
    private static CriteriaTree _explorationTree() {
        CriteriaTreeNode a, b, c, d, e, tmp;

        e = CriteriaTreeNode.treeLeaf(Criteria.moreUnwatchedWallsCriteria);
        e = CriteriaTreeNode.elseNode(Criteria.higherLevelCriteria, e);
        e = CriteriaTreeNode.elseNode(Criteria.lessCliffsCriteria, e);
        e = CriteriaTreeNode.
            elseNode(Criteria.nearestToUnknownWallsCriteria, e);
        e = CriteriaTreeNode.elseNode(Criteria.maybeNearerToHomeCriteria, e);
        
        d = CriteriaTreeNode.elseNode(Criteria.lowerNumberOfCornCriteria, e);
        
        tmp = CriteriaTreeNode.
            choiceNode(FlagCriteria.hasMatchingNumberOfCornCriteria, d, d);
        tmp = CriteriaTreeNode.
            choiceNode(FlagCriteria.hasUnknownNumberOfCornCriteria, e, tmp);
        
        c = CriteriaTreeNode.
            choiceNode(FlagCriteria.hasNoCornCriteria, e, tmp);
        
        b = CriteriaTreeNode.
            choiceNode(FlagCriteria.hasUnknownCornCriteria, e, c);
        
        tmp = CriteriaTreeNode.
            elseNode(Criteria.moreUnwatchedNeighborsCriteria, c);
        tmp = CriteriaTreeNode.
            choiceNode(FlagCriteria.hasNoCornCriteria, tmp, tmp);
        
        a = CriteriaTreeNode.
            elseNode(Criteria.moreUnwatchedNeighborsCriteria, e);
        a = CriteriaTreeNode.elseNode(Criteria.higherLevelCriteria, a);
        
        a = CriteriaTreeNode.
            choiceNode(FlagCriteria.hasUnknownCornCriteria, a, tmp);
        
        tmp = CriteriaTreeNode.
            choiceNode(FlagCriteria.isUnexploredCriteria, a, b);
        tmp = CriteriaTreeNode.elseNode(Criteria.shorterPathCriteria, tmp);
        
        return new CriteriaTree(tmp);
    }
}

/*
 *
 * the exploration tree:
 *
 * +- shorterPathCriteria
 * +- isUnexploredCriteria
 *     +--both--> (A)
 *     +--none--> (B)
 *
 *  (A): 
 *   +- hasUnknownCorn
 *       +--both--> higherLevelCriteria
 *       |           +- moreUnwatchedNeighbors
 *       |           +- (E)
 *       +--none--> hasNoCornCriteria
 *                   +--both--> (A2)
 *                   +--none--> (A2)
 *
 *  (A2):
 *   +- moreUnwatchedNeighbors
 *   +- (C)
 *
 *  (B):
 *   +- hasUnknownCorn
 *       +--both--> (E)
 *       +--none--> (C)
 *
 *  (C):
 *   +- hasNoCornCriteria
 *       +--both--> (E)
 *       +--none--> hasUnknownNumberOfCorn
 *                   +--both--> (E)
 *                   +--none--> hasMatchingNumberOfCorn
 *                              +--both--> (D)
 *                              +--none--> (D)
 *
 *  (D):
 *   +- lowerNumberOfCorn
 *   +- (E)
 *
 *  (E):
 *   +- maybeNearerToHomeCriteria
 *   +- nearestToUnknownWallsCriteria
 *   +- lessCliffsCriteria
 *   +- higherLeveCriteria
 *   +- moreUnwatchedWalls
 *
 */
