

package corncadger.search;

import corncadger.Path;
import corncadger.Map;

// a node in the criteria tree

abstract class CriteriaTreeNode {
    private Criteria c;
    
    // constructor
    protected CriteriaTreeNode(Criteria _c) { this.c = _c; }
    // accessor    
    Criteria criteria() { return this.c; }
    // implemented in subclasses
    abstract Path betterPath(Path _a, Path _b, Map _m);
    
    // creating nodes:
    public static CriteriaTreeNode treeLeaf(Criteria _c) {
        return elseNode(_c, null);
    }
    
    public static CriteriaTreeNode elseNode(Criteria _c,
                                            CriteriaTreeNode _else) {
        return new ElseCriteriaNode(_c, _else);
    }
    
    public static CriteriaTreeNode choiceNode(FlagCriteria _c,
                                              CriteriaTreeNode _both,
                                              CriteriaTreeNode _none) {
        return new FlagCriteriaNode(_c, _both, _none);
    }
    
}

class FlagCriteriaNode extends CriteriaTreeNode {
    // can handle 4 cases:
    // a fullfilles criteria and b not
    // b fullfilles criteria and a not
    // both fullfill criteria
    // none fullfills criteria
    CriteriaTreeNode noneNode;
    CriteriaTreeNode bothNode;

    FlagCriteriaNode(FlagCriteria _c, 
                     CriteriaTreeNode _bothNode,
                     CriteriaTreeNode _noneNode) 
    {
        super(_c);
        this.noneNode = _noneNode;
        this.bothNode = _bothNode;
    }
    private FlagCriteria flagCriteria() {
        return (FlagCriteria)(super.criteria());
    }
    Path betterPath(Path _a, Path _b, Map _m) {
        FlagCriteria c  = this.flagCriteria();
        boolean      bA = c.fullfilled(_a, _m);
        boolean      bB = c.fullfilled(_b, _m);
        if (bA && !bB) return _a;
        if (bB && !bA) return _b;
        if (bA && bB)  return (this.bothNode == null) 
                           ? null : this.bothNode.betterPath(_a, _b, _m);
        return (this.noneNode == null)
            ? null : this.noneNode.betterPath(_a, _b, _m);
    }
}

class ElseCriteriaNode extends CriteriaTreeNode {
    // handels 3 cases
    // a is better than b, otherwise or both are same
    CriteriaTreeNode elseNode;
    ElseCriteriaNode(Criteria _c, CriteriaTreeNode _else) {
        super(_c);
        this.elseNode = _else;
    }
    Path betterPath(Path _a, Path _b, Map _m) {
        Path better = this.criteria().betterPath(_a,_b,_m);
        return (better == null) 
            ? ((this.elseNode == null) 
               ? null   : this.elseNode.betterPath(_a,_b,_m))
            : better;
    }
}
