
package corncadger.search;

import corncadger.Field;
import corncadger.Map;
import corncadger.Path;

import java.util.Vector;
import java.util.Enumeration;

/*
 * this class has a function to find all the paths from this field to anywhere
 */

abstract class FieldPathCrawl {
    // something i got to know
    private Map     map;
    private Field   field;
    private PathSet paths;

    // this is an special id, every visited field gets it
    private   int     crawlID;
    protected int     targetCrawlID = -1;
    private   boolean targetReached = false;
    protected boolean checkTarget   = false;
    protected boolean noCliffs      = false;

    private Vector visitedLast = new Vector();

    // the constructor
    public FieldPathCrawl(Map _m, Field _f, int _crawlID,
                          boolean _noCliffs) {
        this.map      = _m;
        this.field    = _f;
        this.crawlID  = _crawlID;
        this.noCliffs = _noCliffs;

        this.paths = this.createPathSet();
        this.paths.add(new Path(this.field));
        this.field.setCrawlID(_crawlID);
    }

    // accessors
    int crawlID() { return this.crawlID; }
    Map map()     { return this.map;     }
    
    boolean targetReached() { return this.targetReached; }

    public void setTargetCrawlID(int _crawlID) {
        this.checkTarget = true;
        if (this.targetCrawlID == -1)
            this.targetCrawlID = _crawlID;
    }

    //private static final int _exploreDirs[] = { 0, 1, 2, 5, 4, 3 }; // 1
    private static final int _exploreDirs[] = { 0, 1, 5, 2, 4, 3 }; // 2
    // the directions to check
    private static int nextDir(int _from, int _dir) {
        int dir = _from + _exploreDirs[_dir];
        while (dir < 0) dir += 6;
        while (dir > 5) dir -= 6;
        return dir;
    }

    // is the neighbor passable
    boolean isPassable(Field _field, int _dir, 
                       int _wall, Field _n) {
        Map   m       = this.map();
        int   wall    = _wall;
        Field n       = _n;
        int   crawlID;

        if (n == null) 
            return false; // not explored
        crawlID = n.crawlID();
        if (crawlID == this.crawlID())
            return false; // already visited

        if ((wall == Field.WALL) || (wall == Field.CLIFFUP)) 
            return false; // not possible

        if (this.noCliffs && (wall == Field.CLIFFDOWN)) 
            return false; // no cliffs

        if (wall == Field.UNKNOWN) {
            // there is a special case, which might happen:
            // field is only reachable over cliffdowns
            // and there is one unknown wall
            // so this must be passable, to leave this field!
            boolean passable = true;
            search1:
            for (int i = 0; i < 6; i++) {
                if (i == _dir) continue search1;
                wall = m.wallAt(_field, i);
                if (wall == Field.UNKNOWN) {
                    passable = false;
                    break search1;
                }
                else if ((wall == Field.CLIFFDOWN) ||
                         (wall == Field.SLOPEDOWN) ||
                         (wall == Field.WAY) ||
                         (wall == Field.SLOPEUP)) {
                    passable = false;  // there is another way out
                    break search1;
                }
            }
            // found no other way out
            // must be leaveable!!
            if (passable) return true;
            // now i look at my neighbor field
            // if this is a field with only ways in and the only unknown
            // wall is this the one that leads to this field
            // and the level is equal or below me
            // than this field is also passable from here !!
            if (n == null) return false;
            // the other way
            int otherWay = _dir-3;
            if (otherWay < 0) otherWay += 6;
            passable = true;
            search2:
            for (int i = 0; i < 6; i++) {
                if (i == otherWay) continue search2;
                wall = m.wallAt(n, i);
                if (wall == Field.UNKNOWN) {
                    passable = false;
                    break search2;
                }
                else if ((wall == Field.CLIFFDOWN) ||
                         (wall == Field.SLOPEDOWN) ||
                         (wall == Field.WAY) ||
                         (wall == Field.SLOPEUP)) {
                    passable = false;  // there is another way out
                    break search2;
                }
            }
            if (passable) {
                // might be passable
                // now check the height
                int h1 = _field.height();
                int h2  = n.height();
                if ((h1 == Field.UNKNOWN) || (h2 == Field.UNKNOWN))
                    return false;
                if (h2 <= h1) return true;
            }
            return false;
        }
        return true;
    }

    // returns number of added paths
    /**
     * Describe <code>extendPathByOneStep</code> method here.
     *
     * @param _path the path to extend
     * @param _result the set where the results are added
     * @param _deep wether to go deep with search
     * @param _visited the vector of visited fields
     * @return number of added paths
     */
    int extendPathByOneStep(Path _path, PathSet _result,
                            boolean _deep, Vector _visited)
    {
        Field last = _path.end();
        Map   m    = this.map();
        int   from = (_path.length() == 0) 
            ? m.currentDirection() 
            : _path.lastDirection();
        int   cnt  = 0;

        int   dir;
        int   wall;
        Field n;
        turnAround:
        for (int i = 0; i < 6; i++) {
            dir  = nextDir(from, i);
            wall = m.wallAt(last, dir);
            n    = m.neighborAt(last, dir);

            if (!isPassable(last, dir, wall, n))
                continue turnAround;

            if (n == null)
                if ((n = m.neighborAt(last, dir)) == null)
                    continue turnAround;

            // add the field to the path and to visited fields
            _path.append(n, dir);
            if (_result.add(_path, m)) cnt++;  // add the path
            if (_deep)
                // add the field, so it is later marked as visited
                _visited.add(n);               
            else 
                // if not deep mark field as already visited
                this.checkAndMarkField(n);

            _path.removeLast();
        }
        return cnt;
    }

    // creates the PathSet needed for this crawl
    // implemented in subclasses
    abstract PathSet createPathSet();


    /**
     * Describe <code>extendPathsByOneStep</code> method here.
     *
     * @param _deep wether to go deep
     * @return the actual result
     */
    public PathSet extendPathsByOneStep(boolean _deep) {
        Enumeration e      = paths.paths();
        PathSet     result = this.createPathSet();
        while (e.hasMoreElements()) {
            this.extendPathByOneStep((Path)e.nextElement(), result,
                                     _deep, this.visitedLast);
        }
        this.paths = result;
        return result;
    }
    public PathSet currentPaths() {
        return this.paths;
    }
    void checkAndMarkField(Field _one) {
        if (this.checkTarget && (_one.crawlID() == this.targetCrawlID))
            this.targetReached = true;
        else _one.setCrawlID(this.crawlID);
    }
    public void markFieldsAsVisited() {
        Enumeration e = visitedLast.elements();
        Field       one;
	
        while (e.hasMoreElements())
            this.checkAndMarkField((Field)e.nextElement());
        this.visitedLast = new Vector();
    }
} /* FieldPathCrawl */
