
package corncadger.search;

import corncadger.Path;
import corncadger.Map;
import corncadger.Field;
import corncadger.HamsterNeeds;

// checks the fullfilled criteria for every path and compares
abstract class FlagCriteria extends Criteria {

    /* has cliff*/
    public static final FlagCriteria hasCliffCriteria 
        = new HasCliffCriteria();
    
    /* has cliff at end */
    public static final FlagCriteria hasCliffAtEndCriteria 
        = new HasCliffAtEndCriteria();
    
    /* has first cliff at end */
    public static final FlagCriteria hasFirstCliffAtEndCriteria 
        = new HasFirstCliffAtEndCriteria();
    
    /* can drop corn at next cliff down */
    public static final FlagCriteria canDropCornOnNextCliffDownCriteria
        = new CanDropCornOnNextCliffDownCriteria();
    
    /* has corn */
    public static final FlagCriteria hasCornCriteria 
        = new HasCornCriteria();
    
    /* has no corn */
    public static final FlagCriteria hasNoCornCriteria 
        = new HasNoCornCriteria();
    
    /* has unknown number of corn, but has corn */
    public static final FlagCriteria hasUnknownNumberOfCornCriteria
        = new HasUnknownNumberOfCornCriteria();
    
    /* has no or unknown corn */
    public static final FlagCriteria hasNoOrUnknownCornCriteria 
        = new HasNoOrUnknownCornCriteria();
    public static final FlagCriteria hasUnknownCornCriteria 
        = new HasUnknownCornCriteria();
    
    /* end field is unexplored */
    public static final FlagCriteria isUnexploredCriteria
        = new IsUnexploredCriteria();
    /* corn matches */
    public static final FlagCriteria hasMatchingNumberOfCornCriteria
        = new HasMatchingNumberOfCornCriteria();
    
    public Path betterPath(Path _a, Path _b, Map _m) {
        boolean fa = this.fullfilled(_a,_m);
        boolean fb = this.fullfilled(_b,_m);
        if (fa && !fb) return _a;
        if (fb && !fa) return _b;
        return null;
    }

    public abstract boolean fullfilled(Path _a, Map _m);

}

class HasCliffCriteria extends FlagCriteria {
    public HasCliffCriteria() {}
    public boolean fullfilled(Path _a, Map _m) { return _a.hasCliff(_m);  }
}

class HasCliffAtEndCriteria extends FlagCriteria {
    public HasCliffAtEndCriteria() {}
    public boolean fullfilled(Path _a, Map _m) { return _a.hasCliffAtEnd(_m); }
}

class HasFirstCliffAtEndCriteria extends FlagCriteria {
    public HasFirstCliffAtEndCriteria() {}
    public boolean fullfilled(Path _a, Map _m) { 
        return _a.hasFirstCliffAtEnd(_m); 
    }
}

class CanDropCornOnNextCliffDownCriteria extends FlagCriteria {
    public CanDropCornOnNextCliffDownCriteria() {}
    public boolean fullfilled(Path _a, Map _m) {
        return _a.canDropCornOnNextCliffDown(_m);
    }
}

class HasCornCriteria extends FlagCriteria {
    public HasCornCriteria() {}
    public boolean fullfilled(Path _a, Map _m) { 
        return (_a.end().hasCorn() == Field.HASCORN);
    }
}

class HasNoCornCriteria extends FlagCriteria {
    public HasNoCornCriteria() {}
    public boolean fullfilled(Path _a, Map _m) {
        return (_a.end().hasCorn() == Field.HASNOCORN);
    }
}

class IsUnexploredCriteria extends FlagCriteria {
    public IsUnexploredCriteria() {}
    public boolean fullfilled(Path _a, Map _m) {
        return (!_m.isFieldExplored(_a.end()));
    }
}

class HasNoOrUnknownCornCriteria extends FlagCriteria {
    public HasNoOrUnknownCornCriteria() {}
    public boolean fullfilled(Path _a, Map _m) { 
        int has = _a.end().hasCorn();
        return ((has == Field.HASNOCORN) || (has == Field.UNKNOWN));
    }
}

class HasUnknownCornCriteria extends FlagCriteria {
    public HasUnknownCornCriteria() {}
    public boolean fullfilled(Path _a, Map _m) { 
        int has = _a.end().hasCorn();
        return (has == Field.UNKNOWN);
    }
}

class HasUnknownNumberOfCornCriteria extends FlagCriteria {
    public HasUnknownNumberOfCornCriteria() {}
    public boolean fullfilled(Path _a, Map _m) { 
        Field end  = _a.end();
        int   has  = end.hasCorn();
        int   corn = end.corn();
        return ((has == Field.HASCORN) && (corn == Field.UNKNOWN));
    }
}

class HasMatchingNumberOfCornCriteria extends FlagCriteria {
    public HasMatchingNumberOfCornCriteria() {}
    public boolean fullfilled(Path _a, Map _m) { 
        int   corn = _m.hamsterNeeds().cornNeeded();
        Field end  = _a.end();
        return (end.corn() == corn);
    }
}
