
package corncadger.search;

import corncadger.Field;
import corncadger.Path;
import corncadger.Map;

abstract class InverseFieldPathCrawl extends FieldPathCrawl {

    public InverseFieldPathCrawl(Map _m, Field _f, int _crawlID,
				 boolean _noCliffs) 
    {
	super(_m, _f, _crawlID, _noCliffs);
    }
    /*
     * CAREFUL!! this must all be seen inverse!!!
     *
     *
     */
    boolean isPassable(Field _field, int _dir, 
		       int _wall, Field _n) {
	Map   m       = this.map();
        int   wall    = _wall;
	Field n       = _n;
	int   crawlID;

	if (n == null) 
	    return false; // not explored

	crawlID = n.crawlID();
	if (crawlID == this.crawlID())
	    return false; // already visited

        if ((wall == Field.WALL) || (wall == Field.CLIFFDOWN)) 
	    return false; // not possible

	if (this.noCliffs && (wall == Field.CLIFFUP)) 
	    return false; // no cliffs

        if (wall == Field.UNKNOWN) {
            // now i look at my neighbor field
            // if this is a field with only ways in and the only unknown
            // wall is this the one that leads to this field
            // and the level is equal or below me
            // than this field is also passable from here !!
            // the other way
            int otherWay = (_dir < 3) ? (_dir + 3) : (_dir - 3);
            boolean passable = false;
            search1:
            for (int i = 0; i < 6; i++) {
                if (i == otherWay) continue search1;
                wall = m.wallAt(n, i);
                if (wall == Field.UNKNOWN) {
                    passable = false;
                    break search1;
                }
                else if ((wall == Field.CLIFFDOWN) ||
                         (wall == Field.SLOPEDOWN) ||
                         (wall == Field.WAY) ||
			 (wall == Field.SLOPEUP)) {
                    passable = false;  // there is another way out
                    break search1;
                }
		else if (wall == Field.CLIFFUP) {
		    passable = true;
		}
            }
            if (passable) {
                // might be passable
                // now check the height
                int h1 = _field.height();
                int h2  = n.height();
                if ((h1 == Field.UNKNOWN) || (h2 == Field.UNKNOWN))
                    return false;
                if (h2 >= h1) return true;
            }
            return false; // it's not passable
        }
        return true; // all is ok to pass
    }

}
