

package corncadger.search;

import corncadger.Map;
import corncadger.Path;
import corncadger.Field;

import java.util.Enumeration;
import java.util.Vector;

// take a look around
public class LockAround {
    // what i need to know
    FieldPathCrawl crawl = null;
    PathSet        paths = null;
    Map            map   = null;
    Field          start = null;

    // constructor
    public LockAround(Map _m, Field _f) {
        this.start = _f;
        this.map   = _m;
        this.crawl = 
            new NeighborCheckCrawl(_m, start, PathSearch.nextCrawlID());
    }

    // the neighbots
    PathSet neighbors() {
        if (this.paths == null) {
            this.paths = this.crawl.extendPathsByOneStep(false);
        }
        return this.paths;
    }

    // some other accessors
    public boolean allExplored() {
        // take one step and check the fields
        Enumeration e = this.neighbors().paths();
        Path        one;
        Field       f;
        // for now, only check neighbors that aren't behind cliff downs 
        while (e.hasMoreElements()) {
            one = (Path)(e.nextElement());
            f   = one.end();
            if (!this.map.isFieldExplored(f)) {
                return false;
            }
        }
        return true;
    }

    public Vector unexploredPaths() {
        Enumeration e = this.neighbors().paths();
        Path        one;
        Field       f;
        Vector      v = new Vector();
        // for now, only check neighbors that aren't behind cliff downs 
        while (e.hasMoreElements()) {
            one = (Path)(e.nextElement());
            f   = one.end();
            if (!this.map.isFieldExplored(f)) {
                v.addElement(one);
            }
        }
        return v;
    }

    public Path neighborToExplore(int _nextDir) {
        // nextDir is the direction the hamster would go next
        /*        _
         *      _/b\_
         *     /a\_/c\
         *     \_/1\_/
         *     /d\_/e\
         *     \_/2\_/
         *       \_/
         *
         *   if hamster would go from 1 to 2
         *   a,b,c would be ok to explore next
         *   d,e would be ok, if they are not reachable from 2
         *
         */
        PathSet     set    = PathSet.explorePathSet();
        Enumeration e      = this.neighbors().paths();
        Path        one    = null;
        Field       f      = null;
        Map         m      = this.map;
        int         dir    = 0;
        int         dirDiv = 0;

        neighbors:
        while (e.hasMoreElements()) {
            one = (Path)e.nextElement();
            f   = one.end();

            dir    = one.lastDirection();
            dirDiv = dir - _nextDir;
            while (dirDiv < 0) dirDiv += 6;

            if (dirDiv == 0) continue neighbors;

            if (!m.isFieldExplored(f)) {
                if ((dirDiv == 2) || (dirDiv == 3) || (dirDiv == 4))
                    set.add(one,m);
                else {
                    // dirDiv is 1 or 5, which means it's a 
                    // neighbor to the next field (2)
                    // now check if it is reachable from (2)
                    Field next = m.neighborAt(this.start, _nextDir);
                    int   tmp  = 0;
                    if (dirDiv == 1)
                        tmp = _nextDir + 2;
                    else // dirDiv is 5
                        tmp = _nextDir - 2;
                    while (tmp < 0) tmp += 6;
                    while (tmp > 5) tmp -= 6;
                    int wall = m.wallAt(next, tmp);
                    if ((wall == Field.CLIFFUP) || (wall == Field.WALL) ||
                        (wall == Field.UNKNOWN)) {
                        // neighbor seems not reachable from next field
                        // so add to explore set
                        set.add(one);
                    }
                }
            }
        }
        return set.bestPath(m);
    }

}
