/**
 * the path search
 * main class wenn es um das suchen von pfaden geht :-)
 */

package corncadger.search;

import corncadger.Path;
import corncadger.Field;
import corncadger.Map;

import corncadger.DebugWindow;

import java.util.Enumeration;

// super class of all pathSearch classes
public abstract class PathSearch {
    private Map map;
    static private int crawlCount = 1;
    // accessor
    protected static int nextCrawlID() {
        return crawlCount++;
    }

    // initializing subclasses
    public static PathSearch exploreSearch(Map _m, boolean _wantCorn) {
        if (!_wantCorn) return new ExploreHomePathSearch(_m, _wantCorn);
        return new ExplorePathSearch(_m, _wantCorn);
    }
    public static PathSearch exploreSearchFull(Map _m, boolean _wantCorn) {
        if (!_wantCorn) return new ExploreHomePathSearch(_m, _wantCorn, true);
        return new ExplorePathSearch(_m, _wantCorn, true);
    }
    public static PathSearch homeSearch(Map _m) {
        return new HomePathSearch(_m);
    }

    protected static PathSearch exploreSearchWithoutCliffs(Map _m, 
                                                           boolean _wantCorn) 
    {
        return new ExplorePathSearch(_m, _m.currentField(), 
                                     true,      // noCliffs
                                     _wantCorn, // wantCorn
                                     false);    // fullSearch
    }
    public static PathSearch exploreSearchFromHome(Map _m, boolean _wantCorn)
    {
        return new ExplorePathSearch(_m, _m.homeField(), 
                                     false,      // noCliffs
                                     _wantCorn, // wantCorn
                                     false);    // fullSearch
    }
    protected static PathSearch fastestSearch(Map _m, Field _from, Field _to)
    {
        return new FastestPathSearch(_m, _from, _to);
    }
    protected static PathSearch cliffAlternativeSearch(Map _m, PathSet _paths,
                                                       boolean _wantCorn)
    {
        return new AlternativesToCliffSearch(_m, _paths, _wantCorn);
    }

    // constructor
    protected PathSearch(Map _m) {
        this.map = _m;
    }

    // accessors
    static FieldPathCrawl mapInitCrawl(Map _m) {
        return new MapCreationCrawl(_m, _m.homeField(), nextCrawlID(), false);
    }
    Map map() { return this.map; }
    public Path bestPath() {
        return this.bestPaths().bestPath(this.map());
    }

    protected abstract PathSet bestPaths();

} /* PathSearch */


// conctete subclass for searching explore paths
class ExplorePathSearch extends PathSearch {
    FieldPathCrawl crawl;
    boolean        wantCorn;
    boolean        noCliffs;
    boolean        fromHome;
    boolean        fullExplore;
    ExplorePathSearch(Map _m) { super(_m); }
    public ExplorePathSearch(Map _m, Field _start, 
                             boolean _noCliffs, boolean _wantCorn,
                             boolean _fullExplore) {
        super(_m);
        this.crawl = new ExploreFieldPathCrawl(_m, _start, 
                                               nextCrawlID(), _noCliffs);
        this.wantCorn    = _wantCorn;
        this.noCliffs    = _noCliffs;
        this.fromHome    = _start.isHome();
        this.fullExplore = _fullExplore;
    }

    /*    public ExplorePathSearch(Map _m) {
          this(_m, _m.currentField(), false, true, false);
          }*/
    // constructor
    public ExplorePathSearch(Map _m, boolean _wantCorn) {
        this(_m, _m.currentField(), false, _wantCorn, false);
    }
    public ExplorePathSearch(Map _m, boolean _wantCorn, boolean _fullExplore) {
        this(_m, _m.currentField(), false, _wantCorn, _fullExplore);
    }

    // filter the pathset
    PathSet filterPaths(PathSet _set) {
        return (this.fullExplore)
            ? _set.unexploredPaths(this.map(), this.wantCorn, true)
            : _set.unexploredPaths(this.map(), this.wantCorn);
    }

    public PathSet bestPaths() {
        PathSet paths = this.crawl.currentPaths();
        while (true) {
            if (paths.count() == 0) {
                return paths;
            }
            paths = this.filterPaths(paths);
            if (paths.count() != 0) {
                // found paths to explore
                if (this.noCliffs || this.fromHome || this.allowCliffs()) {
                    return paths;
                }

                // check cliffs
                PathSearch ps = 
                    PathSearch.cliffAlternativeSearch(this.map(), paths, 
                                                      this.wantCorn);
                PathSet tmp = ps.bestPaths();
                return tmp;
            }
            paths = this.crawl.extendPathsByOneStep(true);
            this.crawl.markFieldsAsVisited();
        }
    }

    public Path bestPath() {
        Path best = super.bestPath();
        if (!this.fromHome)
            // no more checking
            return best;

        if (best == null) return best;

        if (this.fullExplore) // no more filtering
            return best;

        // path comes from home
        // 20 corn max * 10 points per corn makes 200 points
        // so max length is 100 'cause this path and back would be 200 steps
        // but may be there are cliffs
        // so max is for the path and the way back 200
        if (best.length() > 99) {
            // now check the way back
            int back = Distance.shortestDistance(this.map(), 
                                                 best.end(), 
                                                 this.map().homeField());
            if (back == -1) 
                // now way back know until now ...
                // be curious :-)
                return best;

            if (back+best.length() > 199) {
                // that wouldn be clever .. max would be +-0 points
                if (this.fullExplore)
                    return null;

                // so check the way with a full explore
                best = 
                    (new ExplorePathSearch(this.map(), best.start(),
                                           // with cliffs, wantCorn and full
                                           false , true, true)).bestPath();
                if (best != null) {
                    if (best.length() < 99) {
                        // there is a better way to an field with 
                        // unexplored walls
                        // just be curious and explore :-)
                        return best;
                    }
                }
                return null;
            }

            // seems as the way to explore is longer than 99 steps but
            // the way back is much shorter
        }
        return best;
    }

    public boolean allowCliffs() { return false; }

} /* ExplorePathSearch */

// concrete subclass to search alternatives for cliffs
class AlternativesToCliffSearch extends PathSearch {
    PathSet        paths;
    boolean        wantCorn;

    public AlternativesToCliffSearch(Map _m, PathSet _paths, 
                                     boolean _wantCorn) 
    {
        super(_m);
        this.paths    = _paths;
        this.wantCorn = _wantCorn;
    }

    public PathSet bestPaths() {
        Map     m       = this.map();
        PathSet without = this.paths.pathsWithoutCliffs(m);
        if (without != null) {
            return without;
        }
        // so all result paths have cliffs
        // so now search for the next explore paths
        PathSearch ps = 
            PathSearch.exploreSearchWithoutCliffs(m, this.wantCorn);
        without = ps.bestPaths();
        if (without.count() == 0) return this.paths; // no way without cliffs
        // found alternatives 
        // check all the non-cliff ways
        Enumeration e   = without.paths();
        Enumeration e2;
        Path        one;
        PathSet     with = PathSet.explorePathSet();
        
        while (e.hasMoreElements()) {
            one = (Path)e.nextElement();
            // compare every non-cliff result with every over-cliff result
            e2   = this.paths.paths();
            Path        over;
            overCliffs:
            while (e2.hasMoreElements()) {
                over = (Path)e2.nextElement();
                if (over.end() == one.end()) {
                    // the same end
                    with.add(over,m);
                    continue overCliffs;
                }
                // check the path between overcliff and alternative
                int shortcut = 
                    Distance.shortestDistance(m, over.end(), one.end());
                if (shortcut == -1)
                    // no path found
                    continue overCliffs;
                if ((over.length()+shortcut) < one.length()) {
                    // the over-cliff+shortcut is shorter than alternative
                    with.add(over,m);
                    continue overCliffs;
                }
            }
        }
        // no check if we've found s.th.
        if (with.count() != 0)
            // yeah, we had
            return with;
        // only alternatives left
        // might be they are reached shorter over cliffs
        // so i'll just search the direct paths
        PathSet result = PathSet.wayPathSet();
        e = without.paths();
        while (e.hasMoreElements()) {
            one = (Path)e.nextElement();
            ps  = PathSearch.fastestSearch(m, one.start(), one.end());
            e2 = ps.bestPaths().paths();
            while (e2.hasMoreElements())
                result.add((Path)e2.nextElement(), m);
        }
        return result;
    }
}

/*
 * this is not the fastest way to search a path, but will return the fastest
 * path :-)
 */
class FastestPathSearch extends PathSearch {
    protected FieldPathCrawl startCrawl;
    protected FieldPathCrawl endCrawl;
    protected int            startCrawlID;
    protected int            endCrawlID;

    protected FastestPathSearch(Map _m) {
        super(_m);
    }
    public FastestPathSearch(Map _m, Field _start, Field _end) {
        this(_m);
        this.startCrawlID = nextCrawlID();
        this.endCrawlID   = nextCrawlID();

        this.startCrawl = 
            new FastestPathCrawl(_m, _start, this.startCrawlID);
        this.startCrawl.setTargetCrawlID(this.endCrawlID);
        this.endCrawl   = 
            new InverseFastestPathCrawl(_m, _end, this.endCrawlID);
        this.endCrawl.setTargetCrawlID(this.startCrawlID);
    }

    private PathSet filterPaths(PathSet _paths, int _crawlID) {
        return _paths.pathsWithCrawlIDAtEnd(_crawlID, this.map());
    }

    private PathSet concatToMatching(PathSet _paths, PathSet _set) {
        PathSet     result = PathSet.wayPathSet();
        Enumeration e      = _paths.paths();
        Path        one, other;
        Map         m      = this.map();
        while (e.hasMoreElements()) {
            one   = (Path)e.nextElement();
            other = _set.pathEndingWith(one.end());
            other = other.reverse();
            one   = one.concat(other);
            result.add(one, m);
        }
        return result;
    }

    private PathSet reverseAndConcatToMatching(PathSet _paths, PathSet _set) {
        PathSet     result = PathSet.wayPathSet();
        Enumeration e      = _paths.paths();
        Path        one, other;
        Map         m      = this.map();
        while (e.hasMoreElements()) {
            one   = (Path)e.nextElement();
            other = _set.pathEndingWith(one.end());
            one   = one.reverse();
            other = other.concat(one);
            result.add(other, m);
        }
        return result;
    }

    public PathSet bestPaths() {
        PathSet startPaths = this.startCrawl.currentPaths();
        PathSet endPaths   = this.endCrawl.currentPaths();

        while (true) {
            startPaths = this.startCrawl.extendPathsByOneStep(true);
            this.startCrawl.markFieldsAsVisited();
            if (startPaths.count() == 0) return startPaths;

            if (this.startCrawl.targetReached()) {
                // reached targetCrawlID
                PathSet paths = this.filterPaths(startPaths, this.endCrawlID);
                return this.concatToMatching(paths, endPaths);
            }

            /*
              endPaths   = this.endCrawl.extendPathsByOneStep(true);
              this.endCrawl.markFieldsAsVisited();
              if (endPaths.count() == 0) return endPaths;
              if (this.endCrawl.targetReached()) {
              // reached targetCrawlID
              PathSet paths = this.filterPaths(endPaths, this.startCrawlID);
              return this.reverseAndConcatToMatching(paths, startPaths);
              }
            */
        }
    }

}

// subclass to search the path home
class HomePathSearch extends FastestPathSearch {

    public HomePathSearch(Map _m) {
        super(_m);
        this.startCrawlID = nextCrawlID();
        this.endCrawlID   = nextCrawlID();

        this.startCrawl = 
            new HomePathCrawl(_m, this.startCrawlID);
        this.startCrawl.setTargetCrawlID(this.endCrawlID);
        this.endCrawl   = 
            new InverseHomePathCrawl(_m, this.endCrawlID);
        this.endCrawl.setTargetCrawlID(this.startCrawlID);
    }

}

// subclass to explore a path home
class ExploreHomePathSearch extends ExplorePathSearch {

    public ExploreHomePathSearch(Map _m, Field _start, 
                                 boolean _noCliffs, boolean _wantCorn,
                                 boolean _fullExplore) {
        super(_m);
        this.crawl = new ExploreHomePathCrawl(_m, _start, 
                                              nextCrawlID(), _noCliffs);
        this.wantCorn    = _wantCorn;
        this.noCliffs    = _noCliffs;
        this.fromHome    = _start.isHome();
        this.fullExplore = _fullExplore;
    }
    public ExploreHomePathSearch(Map _m) {
        this(_m, _m.currentField(), false, true, false);
    }
    public ExploreHomePathSearch(Map _m, boolean _wantCorn) {
        this(_m, _m.currentField(), false, _wantCorn, false);
    }
    public ExploreHomePathSearch(Map _m, boolean _wantCorn, 
                                 boolean _fullExplore) {
        this(_m, _m.currentField(), false, _wantCorn, _fullExplore);
    }

    public boolean allowCliffs() { return false; }
}
