/**
 * PathSet .. just a set of paths .. mostly same start
 *
 *
 */

package corncadger.search;

import corncadger.Path;
import corncadger.Field;
import corncadger.Map;
import corncadger.HamsterNeeds;

import java.util.Vector;
import java.util.Enumeration;

public abstract class PathSet {
    // holding the paths in a vector
    protected Vector  paths        = new Vector();

    // some static subclasses
    public static PathSet standardPathSet() {
        return new StandardPathSet();
    }
    public static PathSet explorePathSet() {
        return new ExplorePathSet();
    }
    public static PathSet exploreHomePathSet() {
        return new ExploreHomePathSet();
    }
    public static PathSet wayPathSet() {
        return new WayPathSet();
    }
    public static PathSet inverseWayPathSet() {
        return new InverseWayPathSet();
    }
    public static PathSet homePathSet() {
        return new HomePathSet();
    }
    public static PathSet inverseHomePathSet() {
        return new InverseHomePathSet();
    }

    /**
     * Describe <code>add</code> method here.
     *
     * adds a copy of path to the set
     *
     * @param _path a <code>Path</code> value
     * @return a <code>boolean</code> value
     */
    protected boolean add(Path _path) {
        Path cp = (Path)_path.clone();
        this.paths.addElement(cp);
        return true;
    }

    // this must be overwritten
    public abstract Path    bestPath(Map _m);
    public abstract boolean add(Path _path, Map _m);
    public abstract PathSet newInstance();
    public abstract boolean isWalkable();

    // accessors
    public Enumeration paths() {
        return this.paths.elements();
    }
    public boolean contains(Path _path) {
        return this.paths.contains(_path);
    }
    public boolean containsPathTo(Field _target) {
        Enumeration e = this.paths.elements();
        Path one;
        while ((e.hasMoreElements())) {
            one = (Path)e.nextElement();
            if (one.end() == _target) return true;
        }
        return false;
    }
    public int count() { return this.paths.size(); }

    // some accessors to get the capabilities of the paths
    // and to create new sets
    public PathSet pathsWithoutCliffs(Map _map) {
        int     max      = this.paths.size();
        PathSet without  = this.newInstance();
        Path    one;
        for (int i = 0; i < max; i++) {
            one = (Path)this.paths.elementAt(i);
            if (!one.hasCliff(_map))
                without.add(one);
        }
        return (without.count() == 0) ? null : without;
    }

    public PathSet pathsWithCliffs(Map _map) {
        int     max   = this.paths.size();
        PathSet with  = this.newInstance();
        Path    one;
        for (int i = 0; i < max; i++) {
            one = (Path)this.paths.elementAt(i);
            if (one.hasCliff(_map))
                with.add(one);
        }
        return (with.count() == 0) ? null : with;
    }

    public PathSet unexploredPaths(Map _m, boolean _wantCorn) {
        return this.unexploredPaths(_m, _wantCorn, false);
    }
    public PathSet unexploredPaths(Map _m, boolean _wantCorn, boolean _full) {
        int          max = this.paths.size();
        PathSet      set = this.newInstance();
        Path         one;
        Field        f;
        boolean      explored;
        HamsterNeeds needs       = _m.hamsterNeeds();
        if (needs.onlyExplore()) {
            _wantCorn = false;
        }
        allFields:
        for (int i = 0; i < max; i++) {
            one = (Path)this.paths.elementAt(i);
            f   = one.end();

            explored = (_full) 
                ? _m.isFieldFullExplored(f) 
                : _m.isFieldExplored(f);
            if (!explored) {
                set.add(one, _m);
                continue allFields;
            }
            if (_wantCorn && (f.hasCorn() == Field.HASCORN) && 
                (!f.isHome())) {
                set.add(one, _m);
                continue allFields;
            }
        }
        return set;
    }

    public PathSet pathsWithCrawlIDAtEnd(int _crawlID, Map _m) {
        int     max = this.paths.size();
        PathSet set = this.newInstance();
        Path    one;
        Field   f;
        for (int i = 0; i < max; i++) {
            one = (Path)this.paths.elementAt(i);
            f   = one.end();
            if (f.crawlID() == _crawlID) 
                set.add(one, _m);
        }
        return set;
    }

    public Path firstPath() {
        if (this.paths.size() > 0)
            return (Path)this.paths.elementAt(0);
        return null;
    }

    public Path pathEndingWith(Field _end) {
        int     max = this.paths.size();
        Path    one;
        Field   f;
        for (int i = 0; i < max; i++) {
            one = (Path)this.paths.elementAt(i);
            f   = one.end();
            if (f == _end) return one;
        }
        return null;
    }

    /* ****
     *  some private methods 
     * ***** 
     */

    // returns better path or null if equal
    protected Path _betterPathAfterCriterias(Vector _criterias, 
                                             Path _a, Path _b, 
                                             Map _m)
    {
        int      max  = _criterias.size();
        Criteria one;
        Path     best;
        for (int i = 0; i < max; i++) {
            one  = (Criteria)_criterias.elementAt(i);
            best = one.betterPath(_a, _b, _m);
            if (best != null) return best;
        }
        return null;
    }

    protected boolean _replacePath(Path _one, Path _path) {
        this.paths.removeElement(_one);
        return this.add(_path);
    }
    protected Path _checkCriteriaAndReplace(Criteria _c,
                                            Path _one, Path _path, Map _m)
    {
        Path better = _c.betterPath(_one, _path, _m);
        if (better == null) return null;
        if (better == _one) // path is not better
            return _one;
        this._replacePath(_one, _path);
        return _path;
    }

}
