/*******************************************************************************
/
/ class ObjectList - Dynamische Objectverwaltung (hnelt Delphi-Implementation)
/
/   // Liefert die aktuelle Kapazitt der Liste
/   public int getCapacity()
/
/   // Setzt die Kapazitt der Liste auf den gegebenen Wert (darf nicht kleiner
/   // als getCount sein; die Liste wird automatisch erweitert, wenn deren
/   // Kapazitt nicht meht ausreicht; ein setzen dieses Wertes ist nur sinnvoll
/   // wenn man vorher weiss wie viele Elemente die Liste fassen soll
/   public void setCapacity(int Size)
/
/   // Liefert die Anzahl der in der Liste enthaltenen Objekte
/   public int getCount()
/
/   // Liefert das Objekt am gegebenem Index der Liste
/   public Object getItem(int index)
/
/   // das Objekt am gegeben im Index wird durch das gegebene Objekt ersetzt
/   public void setItem(int index, Object obj)
/
/   // Fgt ein Objekt an das Ende der Liste hinzu; Methode liefert den neuen
/   // Index des hinzugefgten Elements
/   public int add(Object obj)
/
/   // Lscht alle Elemente aus der Liste
/   public void clear()
/
/   // Lscht das Element im gegebenen Index; der Index aller darberliegende
/   // Elemente sinkt um eins
/   public void delete(int index)
/
/   // tauscht die Positionen zweier Objekte
/   public void exchange(int index1, int index2)
/
/   // bestimmt die Position eines Objektes oder liefert -1 zurck, falls es
/   // nicht Element der Liste ist
/   public int indexOf(Object obj)
/
/   // Fgt ein Objekt an dem gegebenem Index ein; der Index aller darber-
/   // liegenden Objekte erhht sich um eins
/   public void insert(int index, Object obj)
/
/   // Entfernt alle Objekte deren Wert null ist aus der Liste
/   public void pack()
/
/   // Entfernt ein gegebenes Objekt aus der Liste; Methode liefert Position des
/   // Objekts zurck bzw. -1 wenn das Objekt nicht Element der Liste ist.
/   public int remove(Object obj)
/
/   // Startet eine Sortierung (Quicksort). Um die Vergleichsfunktions festzu-
/   // legen muss die abtrakte Methode "SortFunc" in "ListSortCompare" ber-
/   // schrieben werden, in der zwei Elemente verglichen werden. Wenn Object1
/   // "grer" als Objekt2 ist, muss dass Ergebnis der Funktion positiv sein,
/   // bei "Gleichheit" beider Objekte 0, ansonsten negativ.
/   public void sort(ListSortCompare Compare)
*******************************************************************************/
package dighty;

public class ObjectList {

  public Object Tag = null;

  public Object[] FItems = null;
  protected int FCapacity = 0;
  protected int FCount = 0;

  private void grow() {
    int Delta = 4;
    if (this.FCapacity > 64)
      Delta = (FCapacity / 4);
    else
     if (FCapacity > 8)
       Delta = 16;
    setCapacity(FCapacity + Delta);
  }

  private void quicksort(int l, int r, ListSortCompare Compare) {
    Object P, T = null;
    int i, j = 0;
    do {
      i = l;
      j = r;
      P = FItems[(l + r) >> 1];
      do {
        while (Compare.CompareFunc(FItems[i], P) > 0)
          i++;
        while (Compare.CompareFunc(FItems[j], P) < 0)
          j--;
        if (i <= j) {
          T = FItems[i];
          FItems[i] = FItems[j];
          FItems[j] = T;
          i++;
          j--;
        }
      }
      while (i <= j);
      if (l < j)
        quicksort(l, j, Compare);
      l = i;
    }
    while (i < r);
  }

  public int getCapacity() {
    return FCapacity;
  }

  public void setCapacity(int Size){
    Object[] NewList = new Object[Size];
    if (FItems != null)
      System.arraycopy(FItems, 0, NewList, 0, FCount);
    FItems = NewList;
    FCapacity = Size;
    FCount = Math.min(Size, FCount);
  }

  public int getCount() {
    return FCount;
  }

  public Object getItem(int index){
    return FItems[index];
  }
  public Object getLast() { return FItems[FCount - 1];}

  public void setItem(int index, Object obj) {
    FItems[index] = obj;
  }

  public int add(Object obj) {
    if (FCount == FCapacity) grow();
    FItems[FCount] = obj;
    FCount++;
    return (FCount - 1);
  }
  public ObjectList Clone() {
    ObjectList cClone = new ObjectList();
    cClone.setCapacity(getCount());
    System.arraycopy(FItems, 0, cClone.FItems, 0, getCount());
    return cClone;
  }
  public void CloneTo(ObjectList cClone) {
    cClone.FItems = null;
    cClone.setCapacity(getCount());
    cClone.FCount = FCount;
    if (FCount > 0)
      System.arraycopy(FItems, 0, cClone.FItems, 0, getCount());
  }

  public void clear() {
    for (int i = 0; i < FCount; i++)
      FItems[i] = null;
    FCount = 0;
  }

  public void delete(int index) {
//    FItems[index] = null;
    System.arraycopy(FItems, index + 1, FItems, index, FCount - index - 1);
    FCount--;
    FItems[FCount] = null;
  }
  static Object ExchangeTempVar;
  public void exchange(int index1, int index2) {
    if (index1 != index2) {
      ExchangeTempVar = FItems[index1];
      FItems[index1] = FItems[index2];
      FItems[index2] = ExchangeTempVar;
    }
  }

  public int indexOf(Object obj) {
    for (int i = 0; i < FCount; i++)
      if (FItems[i] == obj)
        return i;
    return -1;
  }

  public void insert(int index, Object obj) {
    if (FCount == FCapacity) grow();
    if (index != FCount)
      System.arraycopy(FItems, index, FItems, index + 1, FCount - index);
    FItems[index] = obj;
    FCount++;
  }

  public void pack() {
    int iTakePos = 0;
    int iPlacePos = 0;
    while (iTakePos < FCount) if (FItems[iTakePos] == null) break; else iTakePos++;
    if (iTakePos >= FCount) return;
    iPlacePos = iTakePos;
    for (iTakePos++; iTakePos < FCount; iTakePos++)
    {
      if (FItems[iTakePos] != null){
        FItems[iPlacePos] = FItems[iTakePos];
        iPlacePos++;
      }

    }
    for (int i = iPlacePos; i < iTakePos; i++)
      FItems[i] = null;
    FCount -= (iTakePos - iPlacePos);
/*    for (int i = FCount-1; i >=0; i--)
      if (FItems[i] == null)
        delete(i);*/
  }

  public int remove(Object obj) {
    int result = indexOf(obj);
    if (result != -1)
      delete(result);
    return result;
  }

  public void sort(ListSortCompare Compare) {
    if (getCount() > 0)
      quicksort(0, FCount - 1, Compare);
  }
}

