package fidel;
import algds.c2002.Hamster;
import java.util.*;

/**           F I D E L           **/
/**          ~~~~~~~~~~~          **/
/** entworfen von Gunni & Troschi **/

/** Grundkonzept:
die Grundidee ist recht einfach: Fidel tastet sich per Zufall durch das
Labyrinth. In jedem Zug, wir die Richtung zufllig verndert. Dadurch bleibt
Fidel nicht irgendwo stecken, bzw. luft im Kreis. Das 'Finden' des Heimwegs
ist auf die einfachste Art gelst: Jede Drehung und jeder Schritt werden in
einem Stack gespeichert, der dann seitenverkehrt wieder abgebaut wird. Das
erzeugt zwar viele (leider) unntige Schritte, aber auf diese Weise findet
Fidel auf jeden Fall nach Hause*/

public class Fidel extends Hamster {
	
	boolean fertig = false;			//wird auf true gesetzt, wenn der Hamster stopt
	boolean voll = false;			//true, wenn Fidel die Backen voll hat (wird derzeit nicht verwendet)
	int zufall = 0;					//Inititalisierung der Zufallszahl
	Stack pfad = new Stack();		//Stack fr den Heimweg
	Object re = new Object();		//wird bei Rechtsdrehung im Stack gespeichert
	Object li = new Object();		//wird bei Linksdrehung im Stack gespeichert
	Object vor = new Object();		//wird bei Vorwrtsbewegung im Stack gespeichert
	
	
	/** Fidel dreht sich erstmal einmal rum, um zu berprfen, ob das Startfeld 
	nur von Mauern umgeben ist, rennt dann solange, bis fertig auf true gesetzt 
	wird, und zwar wird eine Zufallszahl bestimmt, die spter noch fr die 
	Drehung gebraucht wird, gesammelt, ein Schritt vorwrts gemacht (oder eine
	Drehung) und dann wird noch einmal zufllig gedreht. */
	public void run() {
		drehEinMalRum();
		while(!fertig) {
			zufall = (int) (3*Math.random());
			sammeln();	
			vorwaerts();
			drehZufall();
		}//while
	}//run
	
	/** Um zu berprfen, ob Fidel nicht auf dem Startfeld eingesperrt ist,
	dreht er sich erstmal um die eigene Achse. Wenn berall nur Wnde sind, 
	wre es ja sinnlos, berhaupt loszulaufen...  */
	private void drehEinMalRum(){
		int i = 0;
		while (i < 6){
			i++;
			if ((lookForward() & WALLMASK)!=WALL) return;
			if ((lookForward() & WALLMASK)==WALL) turn(TURN_LEFT);			
		}//while
		if (i >= 6) fertig = true;
	}//drehEinMalRum
	
	/** Vorwrts. Hier wird die Bewegung und die erste Drehung vollzogen. Wenn
	keine Wand im Weg ist, vor! sollte eine da sein, dreht sich Fidel entsprechend
	der vorher ermittelten Zufallszahl. Sollte die Anzahl der bisher gemachten
	Schritte grer als 5000 sein, geht Fidel nach Hause und macht Schlu. Dies
	verhindert, dass es zu viele Minuspunkte gibt. */
	private void vorwaerts(){
		if ((lookForward() & WALLMASK)!=WALL){ forward(); pfad.push(vor); }
		if ((lookForward() & WALLMASK)==WALL && zufall == 1){ turn(TURN_LEFT); pfad.push(li); }
		if ((lookForward() & WALLMASK)==WALL && zufall == 2){ turn(TURN_RIGHT); pfad.push(re); }
		if (getSteps() >= 5000) { nachHause(); fertig = true; }		
	}//vorwaerts
	
	/** Die Heimkehrmethode ist sehr simpel. Jede Bewegung wurde vorher in einem
	Stack gespeichert und wird umgekehrt wieder ausgefhrt. Drei mal turn wendet 
	den Hamster, aus links wird rechts und umgekehrt. Sobald der Hamster wieder
	auf dem Startfeld ist, wird alles abgeladen und zu Sicherheit der Stack geleert. */
	private void nachHause(){
		turn(TURN_LEFT);turn(TURN_LEFT);turn(TURN_LEFT);
		while (!atHome()){			
			if (!pfad.empty() && pfad.peek().equals(vor)){ forward(); pfad.pop(); }
			if (!pfad.empty() && pfad.peek().equals(li)){ turn(TURN_RIGHT);	pfad.pop(); }
			if (!pfad.empty() && pfad.peek().equals(re)){ turn(TURN_LEFT); pfad.pop(); }
			if (pfad.empty()){ break; }		
		}//while
		if (atHome()){ ablegen(); pfad.clear();
			}//if
	}//nachHause
	
	/** Eigentlich selbsterklrend; wenn Fidel zu Hause ist, legt er alle Krner
	ab, die er so mit sich rumtrgt. */
	private void ablegen(){ 
		if (atHome()) dropHere(CORNCARRY_MAX);
	}//ablegen
			
	/** Die Zufallsdrehung: Zuerst wird eine Zufallszahl bestimmt. Dann wird 
	anhand dieser Zahl eine von drei mglichen Aktionen ausgefhrt: Nichts; 
	Drehung nach links oder Drehung nach rechts. */
	private void drehZufall(){
		zufall = (int) (3*Math.random());
		//if (verbose()) log(zufall + " ");
		if (zufall == 0){ return; }
		if (zufall == 1){ turn(TURN_LEFT); pfad.push(li); }
		if (zufall == 2){ turn(TURN_RIGHT); pfad.push(re); }
	}//zufall
	
	/** Sammeln, futtern, bunkern,... 
	Sollte Fidel grade zu Hause sein, legt er natrlich alles ab, ansonsten
	packt er sich erstmal alles in die Backen, was er tragen kann. Wenn kein
	Platz mehr fr weitere Krner ist, geht er nach Hause. */
	private void sammeln(){
		if (atHome()){ ablegen(); pfad.clear(); }
		else if (!atHome()) take(CORNCARRY_MAX);
		if (getCarriedCorn() == CORNCARRY_MAX) { voll = true; nachHause(); }	
	}//sammeln
	
	/** Autoren des Hamsters. */
	public String getHamsterAuthor() { return "Troschi und Gunni"; }
	/** Datum der Hamsterversion. */
	public String getHamsterDate() { return "29.04.2002"; }
	/** Hamsterversion. */
	public String getHamsterVersion() { return "0.058"; }
		
}//class