package frisswienix;
import algds.c2002.Hamster;
import algds.c2002.Constants;
import java.util.Vector;

public class HamsterMap {

    /**
     * Die sechs Richtungen, in die es gehen kann.
     * Norden, Nordosten, Sdosten, Sden, Sdwesten, Nordwesten
     * Das Sechseck im Uhrzeigersinn halt.
     */

    public static final int DIRECTION_NO = 0;
    public static final int DIRECTION_NE = 1;
    public static final int DIRECTION_SE = 2;
    public static final int DIRECTION_SO = 3;
    public static final int DIRECTION_SW = 4;
    public static final int DIRECTION_NW = 5;

    /**
     * Zweidimensionales Array fr die Karte.
     */

    protected HamsterCell map[][];

    /**
     * Man glaube ja nicht, dass dieser Code noch
     * geht, wenn MAZELENGTH_MAX pltzlich nicht 30 ist:
     * Startpunkt des Hamsters ist 29, 29.
     */

    public HamsterMap() {
	int max = Constants.MAZELENGTH_MAX * 2 - 1;
	map = new HamsterCell[max][max];
	for (int x = 0; x < max; x++)
	    for (int y = 0; y < max; y++)
		map[x][y] = new HamsterCell(this, x, y);
    }

    /**
     * Gibt die Hamsterzelle an der Position x, y zurck.
     * Wird man wohl nie brauchen...
     */

    public HamsterCell getCell(int x, int y) {
	return map[x][y];
    }

    /**
     * Gibt die Zuhausezelle zurck.
     */

    public HamsterCell getHomeCell() {
	int max = Constants.MAZELENGTH_MAX;
	return getCell(max - 1, max - 1);
    }

    /**
     * Liefert die an die bergebende Zelle angrenzende Zelle
     * in der angegebenen Richtung zurck.
     * @throws IllegalArgumentException wenn direction falsche Konstante
     * @returns Nachbarfeld oder Null, wenn direction ausserhalb Spielfeld
     */

    public HamsterCell getNeighbour(HamsterCell cell, int direction) {
	int x = cell.getX();
	int y = cell.getY();

	switch (direction) {
	case DIRECTION_NO: y--; break;
	case DIRECTION_NE: x++; break;
	case DIRECTION_SE: x++; y++; break;
	case DIRECTION_SO: y++; break;
	case DIRECTION_SW: x--; break;
	case DIRECTION_NW: x--; y--; break;
	default: throw new IllegalArgumentException("falsche Richtung");
	}

	try {
	    // getCell gibt Null zurck wenn x oder y zwar nicht grob
	    // falsch aber dennoch ausserhalb sind
	    return getCell(x, y);
	} catch (ArrayIndexOutOfBoundsException a) {
	    // hier fangen wir die Exception und geben auch einfach Null
	    return null;
	}
    }

    /**
     * Lscht die Flags aller Zellen. Die Flags knnen
     * von Algorithmen beliebig benutzt werden.
     */

    public void clearFlags() {
	for (int i = 0; i < 59; i++)
	    for (int j = 0; j < 59; j++)
		if (map[i][j] != null)
		    map[i][j].clearFlag();
    }
    
    /**
     * Berechnet fr alle Zellen auf Basis des aktuellen
     * Erforschungszustandes die Home-Distance neu.
     */

    public void calculateHomeDistances() {
	/* Home Distance zurcksetzen */
	for (int i = 0; i < 59; i++)
	    for (int j = 0; j < 59; j++)
		if (map[i][j] != null)
		    map[i][j].setHomeDistance(-1);

	/* Schlange */
	Vector cells = new Vector();

	/* Heimatzelle als erstes Element in Schlange */
	HamsterCell cell = getHomeCell();
	cells.addElement(cell);
	cell.setHomeDistance(0);

	while (cells.size() > 0) {
	    /* vordersten Eintrag aus Schlange holen */
	    cell = (HamsterCell)cells.elementAt(0);
	    cells.removeElementAt(0);

	    int dist = cell.getHomeDistance() + 1;

	    /* alle Nachbarn in Schlange einreihen */
	    for (int direction = 0; direction < 6; direction++)
		/* geht's in die Richtung? */
		if (cell.getWall(direction) == cell.WALL_NO) {
		    HamsterCell neighbour = getNeighbour(cell, direction);
		    /* gibt es eine Nachbarzelle? ist sie schon erreichbar? */
		    if (neighbour != null &&
			neighbour.getHomeDistance() == -1) {
			neighbour.setHomeDistance(dist);
			cells.addElement(neighbour);
		    }
		}
	}
    }

    /**
     * Bestimmt die Richtung, in die der Hamster gehen muss, um
     * von cell1 nach cell2, die aneinander grenzen mssen, gehen
     * muss.
     */

    public int getCellOrientation(HamsterCell cell1, HamsterCell cell2) {
	int dx = cell2.getX() - cell1.getX();
	int dy = cell2.getY() - cell1.getY();

	if (dy == -1 && dx == 0)
	    return DIRECTION_NO;
	else if (dy == 0 && dx == 1)
	    return DIRECTION_NE;
	else if (dy == 1 && dx == 1)
	    return DIRECTION_SE;
	else if (dy == 1 && dx == 0)
	    return DIRECTION_SO;
	else if (dy == 0 && dx == -1)
	    return DIRECTION_SW;
	else if (dy == -1 && dx == -1)
	    return DIRECTION_NW;
	else
	    throw new IllegalArgumentException("Zellen nicht nebeneinander");
    }

    /**
     * Kehrt eine Richtung um.
     */

    public int getOppositeDirection(int direction) {
	switch (direction) {
	case DIRECTION_NO: return DIRECTION_SO;
	case DIRECTION_NE: return DIRECTION_SW;
	case DIRECTION_SE: return DIRECTION_NW;
	case DIRECTION_SO: return DIRECTION_NO;
	case DIRECTION_SW: return DIRECTION_NE;
	case DIRECTION_NW: return DIRECTION_SE;
	default: throw new IllegalArgumentException("falsche Richtung");
	}
    }

}
