package h2d2c;
import algds.c2002.Hamster2;
import java.util.*;
import java.io.*;
import h2d2c.util.*;

/**
 * @author Christian Semrau
 */
public class H2D2 extends
Hamster2
implements H2D2Constants {

protected Map map;
protected Cell homecell;
protected Cell curcell;
protected IntPoint position;
protected int direction;
protected boolean dontknowwayhome = false;

protected boolean FOLLOW_WALLS = false;
protected int[] DIR_PREFERENCE = (int[])(SAMEDIR.clone());

public static H2D2 currenthamster;

public static boolean VERBOSE = false;
public static boolean DEBUGSTEPS = false;
public static boolean DEBUGHOME = false;

public void run() {
  currenthamster = this;

  map = new Map();
  direction = N;
  position = map.homePoint;
  homecell = map.homecell;
  curcell = homecell;

  homecell.wayhome.set(0,-1);
  homecell.waymin.set(0,-1);

  boolean v = VERBOSE;
//  VERBOSE = false;
  map.updateWayMin(homecell);
  VERBOSE = v;

  try{
    control();
  }catch(DoStopException exc){
    if (exc.getMessage()!=null && exc.getMessage().length()>0) {
      StringWriter st = new StringWriter();
      exc.printStackTrace(new PrintWriter(st));
      logln(st.toString());
    }
  }
}

public void control() {
  VERBOSE = true;
  lookaround();
  while (true) {
//    if (getSteps()>=1130) VERBOSE = true;
//    if (getSteps()>=1160) VERBOSE = true;
    gotointerestingneighbor();
    if (getavailablecorn()>0) {
      dotake();
    }
    if (getCarriedCorn()==CORNCARRY_MAX) {
      bringhome(false);  // nur laufen, wenn ein Weg bekannt ist
    }
  }
}

// Liefert die Richtung in die der Hamster als naechstes gehen soll,
// wenn -1 geliefert wird, muss der Hamster weiter weg suchen
protected int explorationDirection() {
  int nexttowall = -1, corn = -1, notseenneighbor = -1;
  int notseenCount = 0;
  int wallAndUnknown = -1;
  for (int i=0; i<NUMDIR; i++) {
    int dir = DIR_PREFERENCE[i];
    if (curcell.isWall(dir) || curcell.isCliffUp(dir)) continue;

    Cell n = curcell.getLink(dir);
    checkifneedtovisit(n);
    if (n.getState()==ENTERED && (dontknowwayhome || n.getCorn()==0)) continue;

//    if (!curcell.isFree(dir)) continue;

    boolean c = false, u = false, w = false;
    if (!dontknowwayhome &&
      (n.getCorn()>0 || n.getCorn()==UNKNOWNCORNAMOUNT)) { c = true; }
    if (n.countNotseenNeighbor() > notseenCount) {
      notseenCount = n.countNotseenNeighbor();
      u = true;
    }
    if (FOLLOW_WALLS &&
      (curcell.isWall(LEFTDIR[dir]) || curcell.isWall(RIGHTDIR[dir]))
      ) { w = true; }

    if (c) corn = dir;
    if (u) notseenneighbor = dir;
    if (w) nexttowall = dir;
    if (u&w) wallAndUnknown = dir;
  }

  if (corn>=0) return(corn);
  if (wallAndUnknown>=0) return(wallAndUnknown);
  if (notseenneighbor>=0) return(notseenneighbor);
  return -1;
}

/** Geht zum naechsten interessanten Feld. */
protected void gotointerestingneighbor() {

  int dir = explorationDirection();
  if (dir>=0) {
    turnto(dir);
    goforward();
  }else{

    boolean f = gotointerestingfield();
    if (!f){
      findhome(true);
      if (dontknowwayhome) {
        if (curcell.possibleway.distance==INFINITEDISTANCE) {
          logln("Nichtmal ansatzweise ist ein Heimweg auffindbar!");
          throw new DoStopException("Boeser User! Sperrt mich ein!");
        }
        gotofield3();  // geht possibleway bis zum ersten nur gesehenen Feld
      }else{
        logln("Nicht interessantes mehr. Gehe heim.");
        bringhome(true);
        throw new DoStopException();
      }
    }

  }

}

/** Geht zum naechsten interessanten Feld. */
protected boolean gotointerestingfield() {
  boolean cutDistance = atHome();

  map.clearKnownWay();
  PriorityQueue queue = new PriorityQueue(CompareKnownWay.COMPARATOR);
  curcell.knownway.set(0,-1);
  queue.addFirst(curcell);

  LinkedList candidates = new LinkedList();
  int candidateDist = INFINITEDISTANCE;

  while(!queue.isEmpty()) {
    Cell c = (Cell)queue.removeMin();

    int currdist = c.knownway.distance;

    checkifneedtovisit(c);
    if ((c.getState()<ENTERED && c.hasNotseenNeighbor()) || (!dontknowwayhome && c.getCorn()!=0)) {
      if (candidateDist>currdist)
        candidateDist=currdist;
      else
      if (candidateDist<currdist)
        break; // alle Kandidaten gefunden

      candidates.addFirst(c);
    }

    if (cutDistance && (currdist > MAX_DISTANCE)) {
      logln("Kein interessantes Feld in akzeptabler Entfernung.");
      return false;
      // die Suche wird abgebrochen
    }

    for(int i=0; i<NUMDIR; i++) {
      int dir = DIR_PREFERENCE[i];
      if (!c.isKnown(dir)) { continue; }
      if (c.isWall(dir) || c.isCliffUp(dir)) {
        continue;
      }
      Cell n = c.getLink(dir);

      if (n==null) {
        logln("gotointerestingfield: Fehler: Nachbar null obwohl freier Weg!");
        logln("HamsterPosition: "+position+", Richtung: "+DIRSTRING[direction]);
        logln("SuchPosition: "+c.getPosition()+", Richtung: "+DIRSTRING[dir]);
        throw new DoStopException("Nachbar null obwohl freier Weg!");
      }
/*
      if (c.isCliffDown(dir) && n.getState()<ENTERED) {
        continue;
      }
*/
      Way nw = n.knownway;
      if (nw.distance>currdist+1){
        nw.distance=currdist+1;
        nw.direction=ANTIDIR[dir];
        queue.addLast(n);
      }
    }
  }

  if (candidates.size()==0)
    return false;

  Cell dest;

  Cell corn = null, free = null;
  for (Iterator e = candidates.iterator(); e.hasNext();) {
    dest = (Cell)e.next();
    if (dest.getCorn()>0 || dest.getCorn()==UNKNOWNCORNAMOUNT) corn = dest;
    free = dest;
  }
  if (corn!=null) dest = corn; else dest = free;


  map.clearPossibleWay();
  queue = new PriorityQueue(ComparePossibleWay.COMPARATOR);
  curcell.possibleway.set(0,-1);
  queue.addFirst(curcell);

  while(!queue.isEmpty()) {
    Cell c = (Cell)queue.removeMin();

    int currdist = c.possibleway.distance;

    checkifneedtovisit(c);
    if (c==dest) {
      break; // Kandidat gefunden
    }

    for(int i=0; i<NUMDIR; i++) {
      int dir = DIR_PREFERENCE[i];
      // if (!c.isKnown(dir)) { continue; }
      if (c.isWall(dir) || c.isCliffUp(dir)) {
        continue;
      }
      Cell n = c.getLink(dir);

      if (n==null) {
        logln("gotointerestingfield: Fehler: Nachbar null obwohl freier Weg!");
        logln("HamsterPosition: "+position+", Richtung: "+DIRSTRING[direction]);
        logln("SuchPosition: "+c.getPosition()+", Richtung: "+DIRSTRING[dir]);
        throw new DoStopException("Nachbar null obwohl freier Weg!");
      }
      if (n.getState()==NOTSEEN) continue;
/*
      if (c.isCliffDown(dir) && n.getState()<ENTERED) {
        continue;
      }
*/
      Way nw = n.possibleway;
      if (nw.distance>currdist+1){
        nw.distance=currdist+1;
        nw.direction=ANTIDIR[dir];
        queue.addLast(n);
      }
    }
  }

  int knowndist = dest.knownway.distance;
  int possibledist = dest.possibleway.distance;

//  // Beide Methoden brechen den Pfad ab, sobald ein neues Feld erkundet wurde
//  if (possibledist < knowndist)
//    gotofield3(dest);  // possibleway gehen
//  else
      gotofield(dest);  // knownway gehen

  return true;
}

/** Geht zum angegebenen Feld, indem der Pfad zurueckverfolgt wird.
 * Es wird davon ausgegangen, dass der Hamster auf dem Feld mit
 * knownway.distance==0 steht.
 */
protected void gotofield(Cell c){
  Way w = c.knownway;
  int dist = w.distance;
  if (dist==INFINITEDISTANCE) {
    logln("gotofield: no knownway!");
    throw new DoStopException("No knownway found!");
  }
  int[] dirs = new int[dist];

  for(int i=0; i<dist; i++) {
    dirs[i] = w.direction;
    c = c.getLink(w.direction);
    w = c.knownway;
  }

  for(int i=dist-1; i>=0; i--) {
    turnto(ANTIDIR[dirs[i]]);
    goforward();
  }
}

/** Geht zum angegebenen Feld, indem der Pfad verfolgt wird,
 * aber nur solange keine neuen Felder betreten werden.
 * Sobald ein neues Feld betreten wurde, wird der Pfad abgebrochen
 * und false geliefert.
 * Es wird davon ausgegangen, dass der Hamster auf dem Feld mit
 * knownway.distance==0 steht und nach c will.
 */
protected boolean gotofield2(Cell c){
  Way w = c.knownway;
  int dist = w.distance;
  if (dist==INFINITEDISTANCE) {
    logln("gotofield2: no knownway!");
    throw new DoStopException("No knownway found!");
  }
  int[] dirs = new int[dist];

  for(int i=0; i<dist; i++) {
    dirs[i] = w.direction;
    c = c.getLink(w.direction);
    w = c.knownway;
  }

  for(int i=dist-1; i>=0; i--) {
    turnto(ANTIDIR[dirs[i]]);
    boolean newinfo = (curcell.getLink(direction).getState()<ENTERED);
    goforward();
    if (newinfo) return false;
  }
  return true;
}

// selbe wie gotofield2, aber benutzt possibleway statt knownway
protected boolean gotofield3(Cell c){
  Way w = c.possibleway;
  int dist = w.distance;
  if (dist==INFINITEDISTANCE) {
    logln("gotofield3: no possibleway!");
    throw new DoStopException("No possibleway found!");
  }
  int[] dirs = new int[dist];

  for(int i=0; i<dist; i++) {
    dirs[i] = w.direction;
    c = c.getLink(w.direction);
    w = c.possibleway;
  }

  for(int i=dist-1; i>=0; i--) {
    turnto(ANTIDIR[dirs[i]]);
    boolean newinfo = (curcell.getLink(direction).getState()<ENTERED);
    if ((lookForward()&WALLMASK)==WALL || (lookForward()&WALLMASK)==CLIFFUP)
      return false; // sofort stoppen wenn wir vor einer Wand stehen
    goforward();
    if (newinfo) return false; // neues Feld erkundet, also neu orientieren
  }
  return true;
}

/** Geht zum angegebenen Feld, indem der Pfad verfolgt wird,
 * aber nur solange keine neuen Felder betreten werden.
 * Sobald ein neues Feld betreten wurde, wird der Pfad abgebrochen
 * und false geliefert.
 * Es wird davon ausgegangen, dass der Hamster zum Feld mit
 * knownway.distance==0 will.
 */
protected boolean gotofield2(){
  Way w = curcell.knownway;
  int dist = w.distance;
  if (dist==INFINITEDISTANCE) {
    logln("gotofield2: no knownway!");
    throw new DoStopException("No knownway found!");
  }
  int[] dirs = new int[dist];

  Cell c = curcell;
  for(int i=0; i<dist; i++) {
    dirs[i] = w.direction;
    c = c.getLink(w.direction);
    w = c.knownway;
  }

  for(int i=0; i<dist; i++) {
    turnto(dirs[i]);
    boolean newinfo = (curcell.getLink(direction).getState()<ENTERED);
    goforward();
    if (newinfo) return false;
  }
  return true;
}

// selbe wie gotofield2, aber benutzt possibleway statt knownway
protected boolean gotofield3(){
  Way w = curcell.possibleway;
  int dist = w.distance;
  if (dist==INFINITEDISTANCE) {
    logln("gotofield3: no possibleway!");
    throw new DoStopException("No possibleway found!");
  }

  int[] dirs = new int[dist];

  Cell c = curcell;
  for(int i=0; i<dist; i++) {
    dirs[i] = w.direction;
    c = c.getLink(w.direction);
    w = c.possibleway;
  }

  for(int i=0; i<dist; i++) {
    turnto(dirs[i]);
    boolean newinfo = (curcell.getLink(direction).getState()<ENTERED);
    if ((lookForward()&WALLMASK)==WALL || (lookForward()&WALLMASK)==CLIFFUP)
      return false; // sofort stoppen wenn wir vor einer Wand stehen
    goforward();
    if (newinfo) return false; // neues Feld erkundet, also neu orientieren
  }
  return true;
}

// selbe wie gotofield2, aber benutzt wayhome statt knownway
protected boolean gotofield4(){
  Way w = curcell.wayhome;
  int dist = w.distance;
  if (dist==INFINITEDISTANCE) {
    logln("gotofield4: no wayhome at "+curcell);
    throw new DoStopException("No wayhome found!");
  }

  int[] dirs = new int[dist];

  Cell c = curcell;
  for(int i=0; i<dist; i++) {
    dirs[i] = w.direction;
    c = c.getLink(w.direction);
    w = c.wayhome;
    if (w.distance == INFINITEDISTANCE) {
      logln("Ploetzlich kein wayhome mehr da bei "+c);
      return false;
    }
  }

  for(int i=0; i<dist; i++) {
    turnto(dirs[i]);
    if ((lookForward()&WALLMASK)==WALL || (lookForward()&WALLMASK)==CLIFFUP)
      return false; // sofort stoppen wenn wir vor einer Wand stehen
    goforward();
  }
  return true;
}

protected void bringhome(boolean tryIfNotKnown) {
  if (atHome()){
    dodrophere();
    return;
  }
/*
  if (curcell.wayhome.distance < INFINITEDISTANCE
  && curcell.wayhome.distance == curcell.waymin.distance) {
    if (verbose() && DEBUGHOME) logln("Going home, way home dist="+curcell.wayhome.distance);
    gotofield4();
    if (!atHome()){
      if (verbose() && DEBUGHOME) logln("Nicht zuhause angekommen, versuche normale Routine.");
    }else
      return;
  }
*/
  if (findhome(tryIfNotKnown)){
    int currdist = curcell.knownway.distance;
    int possibledist = curcell.possibleway.distance;

//    if ((currdist>50 || currdist!=possibledist) && verbose()) {
      if (verbose() && DEBUGHOME) logln("Going home, possible dist="+possibledist+", known dist="+currdist+", wayhome dist="+curcell.wayhome.distance);
//    }

    int trialcount = 0;
    boolean trypossible = false;
    int lastpossibledist = possibledist;
    do {

      if (possibledist==INFINITEDISTANCE) {
        if (verbose() && DEBUGHOME) logln("No way home found.");
        dodrophere();
        return;
      }

      boolean amziel;
/*
      if (curcell.wayhome.distance < INFINITEDISTANCE
      && curcell.wayhome.distance == curcell.waymin.distance) {
        if (verbose() && DEBUGHOME) logln("Going home, wayhome dist="+curcell.wayhome.distance);
        gotofield4();
        if (!atHome()) {
          findhome(tryIfNotKnown);
          continue;
        }else
          break;
      }
*/
      boolean triedpossible;
      trypossible = trialcount<3 && possibledist<currdist && lastpossibledist>=possibledist;
      if (trypossible || curcell.knownway.distance==INFINITEDISTANCE) {
        if (verbose() && DEBUGHOME) logln("try possible way");
        amziel = gotofield3();
        triedpossible = true;
      }else{
        if (verbose() && DEBUGHOME) logln("go known way");
        amziel = gotofield2();
        triedpossible = false;
      }

      if (!amziel){
        int olddir = triedpossible ? curcell.possibleway.direction : curcell.knownway.direction;
        findhome(tryIfNotKnown);
        int newdir = triedpossible ? curcell.possibleway.direction : curcell.knownway.direction;
        lastpossibledist = possibledist;
        currdist = curcell.knownway.distance;
        possibledist = curcell.possibleway.distance;
        if (trypossible && olddir!=newdir)
          trialcount++;
        if (verbose() && DEBUGHOME) logln("Still going home, possible dist="+possibledist+", known dist="+currdist+", trial="+trialcount);
      }else
      if (!atHome()){
        if (verbose() && DEBUGHOME) logln("bringhome: Vorzeitig gestoppt...");
        break;
      }
    } while (!atHome());

    dodrophere();
  }else{
    if (verbose()) logln("Kein Heimweg gefunden!");
    dodrophere();
  }
}

/**
 * Ermittelt den Heimweg mit einer Breitensuche.
 */
protected boolean findhome(boolean tryIfNotKnown) {

  if (curcell.wayhome.distance<INFINITEDISTANCE
  && curcell.wayhome.distance == curcell.waymin.distance) {
    dontknowwayhome = false;
    Cell c = curcell;
    Way w = c.wayhome;

  //-----------------------
//  setDeveloperFlag(c.position.x, c.position.y, 4, w.direction);
  //-----------------------

    while (w.distance>0) {
      c.knownway.set(w.distance, w.direction);
      c.possibleway.set(w.distance, w.direction);
      c = c.getLink(w.direction);
      w = c.wayhome;
    }
    return true;
  }

  map.clearKnownWay();
  map.clearPossibleWay();

  PriorityQueue queue = new PriorityQueue(CompareKnownWay.COMPARATOR);
  homecell.knownway.set(0,-1);
  queue.addFirst(homecell);

  boolean homefound = false;
  while(!queue.isEmpty()) {
    Cell c = (Cell)queue.removeMin();
    int currdist = c.knownway.distance;

    if (c==curcell) {
      dontknowwayhome = false;
      homefound = true;
      break;
    }

    for(int dir=MINDIR; dir<=MAXDIR; dir++) {
      if (!c.isKnown(dir)) continue;
      if (c.isWall(dir) || c.isCliffDown(dir)) continue;
      Cell n = c.getLink(dir);

      if (n==null) {
        logln("findhome: Leerer Nachbarlink bei "+c+", "+DIRSTRING[dir]+"!");
        continue;
      }

      Way nw = n.knownway;
      if (nw.distance>currdist+1){
        nw.distance=currdist+1;
        nw.direction=ANTIDIR[dir];
        queue.addLast(n);
      }
    }
  }

  if (!homefound) {
    dontknowwayhome = true;
    if (!tryIfNotKnown)
      return false;
  }

  queue = new PriorityQueue(ComparePossibleWay.COMPARATOR);
  homecell.possibleway.set(0,-1);
  queue.addFirst(homecell);

  while(!queue.isEmpty()) {
    Cell c = (Cell)queue.removeMin();
    int currdist = c.possibleway.distance;

    if (c==curcell) {
      break;
    }

    for(int dir=MINDIR; dir<=MAXDIR; dir++) {
      //if (!c.isKnown(dir)) continue;
      if (c.isWall(dir) || c.isCliffDown(dir)) continue;
      Cell n = c.getLink(dir);

      if (n==null) {
        logln("findhome: Leerer Nachbarlink bei "+c+", "+DIRSTRING[dir]+"!");
        continue;
      }

      if (n.getState()==NOTSEEN) continue; // ungesehene Felder nicht nutzen

      Way nw = n.possibleway;
      if (nw.distance>currdist+1){
        nw.distance=currdist+1;
        nw.direction=ANTIDIR[dir];
        queue.addLast(n);
      }
    }
  }

  return true;
}

protected void checkifneedtovisit(Cell c) {
  if (c.getState()!=JUSTSEEN) return;
  if (c.getCorn()!=0) return;

  boolean allKnown = true;
  allKnown &= c.isKnown(N);
  allKnown &= c.isKnown(NE);
  allKnown &= c.isKnown(SE);
  allKnown &= c.isKnown(S);
  allKnown &= c.isKnown(SW);
  allKnown &= c.isKnown(NW);

  if (!allKnown) return;

  if (verbose() && DEBUGSTEPS) logln("Need not visit "+c.fullDebug());
  c.setState(ENTERED);
}

protected void lookaround() {
  curcell.setCorn(getavailablecorn());

  boolean allKnown = (curcell.getState()==ENTERED);
  if (allKnown){
  allKnown &= curcell.isKnown(N);
  allKnown &= curcell.isKnown(NE);
  allKnown &= curcell.isKnown(SE);
  allKnown &= curcell.isKnown(S);
  allKnown &= curcell.isKnown(SW);
  allKnown &= curcell.isKnown(NW);
  }

  if (allKnown) return;

  if (verbose() && DEBUGSTEPS) logln("Looking around at "+curcell);

  for(int i=0; i<NUMDIR; i++) {
    int flag = lookForward();

    if ((flag&WALLMASK)==WALL) {
      map.setWall(position, direction);
    }else
    if ((flag&WALLMASK)==CLIFFUP) {
      map.setCliffUp(position, direction);
    }else
    if ((flag&WALLMASK)==CLIFFDOWN) {
      map.setCliffDown(position, direction);
      Cell n = curcell.getLink(direction);
      n.incState(JUSTSEEN);
    }else{
      map.setFree(position, direction);
      Cell n = curcell.getLink(direction);
      n.incState(JUSTSEEN);
    }

    if ((flag&CORNMASK)==CORN) {
      Cell n = curcell.getLink(direction);
      if (n.getCorn()==UNKNOWNCORNSTATE)
        n.setCorn(UNKNOWNCORNAMOUNT);
    }else{
      if ((flag&WALLMASK)!=SLOPEUP
      &&  (flag&WALLMASK)!=CLIFFUP
      &&  (flag&WALLMASK)!=WALL) {
        Cell n = curcell.getLink(direction);
        n.setCorn(0);
      }
    }

    turnleft();
  }

  curcell.incState(ENTERED);
  if (verbose() && DEBUGSTEPS) logln("Looked around at "+curcell);

  if (dontknowwayhome) {
    if (findhome(false)) // nur true liefern, wenn ein Weg bekannt ist
      if (verbose()) logln("Heimweg wiedergefunden!");
  }

//  map.updateWayHome(curcell);
//  if (map.DEBUGWAYHOME && verbose()) logln("Updated wayhome at "+curcell);
//  map.updateWayMin(curcell);
//  if (map.DEBUGWAYMIN && verbose()) logln("Updated waymin at "+curcell);

} // lookaround

protected void goforward() {
  if (curcell.isWall(direction) || curcell.isCliffUp(direction)) {
    logln("Fehler: Habe versucht, gegen eine Wand oder Klippe zu laufen!");
    logln("Position: "+position+", Richtung: "+DIRSTRING[direction]);
    throw new DoStopException("Lauf gegen Wand.");
  }

  boolean res = forward();
  if (!res) {
    logln("Fehler: Bin gegen eine Wand oder Klippe gelaufen, obwohl keine gemerkt war!");
    logln("Position: "+position+", Richtung: "+DIRSTRING[direction]);
    throw new DoStopException("Lauf gegen Wand.");
  }

  Cell oldcurr = curcell;
  curcell = curcell.getLink(direction);
  position = curcell.getPosition();
  if (verbose() && DEBUGSTEPS) logln("Walked from "+oldcurr
    +" in direction "+DIRSTRING[direction]+" to "+curcell+", step "+getSteps());

  if (atHome()) dodrophere();
  lookaround();
}

protected int dotake() {
  if (!atHome()) {
    int r = take(CORNCARRY_MAX);
    curcell.setCorn(getavailablecorn());
    return r;
  }else{
    logln("Versuch, Mais vom Heimatfeld aufzunehmen!");
    return 0;
  }
}

protected int dodrophere() {
  int r = dropHere(CORNCARRY_MAX);
  curcell.setCorn(getavailablecorn());
  return r;
}

protected int dodropforward() {
  if (curcell.isCliffDown(direction)) {
    int res = dropForward(CORNCARRY_MAX);
    curcell.getLink(direction).changeCorn(res);
    return res;
  }else{
    logln("Versuch, Mais abzuwerfen ohne Klippe!");
    logln("Position: "+position+", Richtung: "+DIRSTRING[direction]);
    return 0;
  }
}

protected int getavailablecorn() {
  if (atHome())
    return 0;
  else
    return super.getAvailableCorn();
}

protected void turnleft() {
  turn(TURN_LEFT);
  direction = LEFTDIR[direction];
}

protected void turnleft2() {
  turn(TURN_LEFT);
  turn(TURN_LEFT);
  direction = LEFT2DIR[direction];
}

protected void turnaround() {
  turn(TURN_LEFT);
  turn(TURN_LEFT);
  turn(TURN_LEFT);
  direction = ANTIDIR[direction];
}

protected void turnright2() {
  turn(TURN_RIGHT);
  turn(TURN_RIGHT);
  direction = RIGHT2DIR[direction];
}

protected void turnright() {
  turn(TURN_RIGHT);
  direction = RIGHTDIR[direction];
}

protected void turnto(int newdir) {
  if (newdir==LEFTDIR[direction]) turnleft();
  else
  if (newdir==LEFT2DIR[direction]) turnleft2();
  else
  if (newdir==ANTIDIR[direction]) turnaround();
  else
  if (newdir==RIGHT2DIR[direction]) turnright2();
  else
  if (newdir==RIGHTDIR[direction]) turnright();
}

public String getHamsterAuthor() {
  return "ChS";
}
public String getHamsterVersion() {
  return "0.03";
}
public String getHamsterDate() {
  return "31.12.2001";
}

public boolean verbose() {
  return VERBOSE && super.verbose();
}

public static boolean staticVerbose() {
  if (currenthamster==null) return false;
  return currenthamster.verbose();
}

//-----------------------
//public static void staticSetDeveloperFlag(int hx, int hy, int f, int d) {
//  if (currenthamster==null) return;
//  currenthamster.setDeveloperFlag(hx, hy, f, d);
//}
//-----------------------

}