package h2d2c.util;
import java.util.*;
/**
 * AlgoDat SS00 Aufgabe 49
 * Datentyp Deque
 * implementiert mit einem Array unbegrenzter Groesse
 * 
 * @author Christian Semrau, 29.04.2000
 * <a href="mailto:Christian.Semrau@student.uni-magdeburg.de">
 * Christian.Semrau@student.uni-magdeburg.de</a>
 */
public class ArrayDeque {
  protected Object[] feld;  // das Feld das die Daten enthaelt
  protected int offset, length;  // Beginn und Laenge des Datenbereichs
  protected int loadstep;  // Schrittweite der Vergroesserung
  protected int modCount = 0;  // zaehlt die Veraenderungen der Deque.
  // Das macht den Iterator "sicher", indem der Iterator beim Feststellen
  // einer Veraenderung mit einer ConcurrentModificationException abbricht.
  // !! Die protected-Methoden veraendern modCount nicht, man muss also
  // beim Aufruf solcher Methoden selbst den modCount erhoehen !!

/**
 * Legt eine leere Deque an.
 */
public ArrayDeque() {
    this(10);
}

/**
 * Legt eine leere Deque an.
 * @param loadstep die Anzahl von Eintraegen, fuer die beim Vergroessern
 * der Warteschlange Platz reserviert werden soll.
 */
public ArrayDeque(int loadstep) {
    this.loadstep = loadstep;
    feld=new Object[2*loadstep];
    offset=loadstep; length=0;
}

/**
 * Leert die Deque.
 */
public void clear(){
    modCount++;
    offset = feld.length/2; length = 0;
    for (int i=0; i<feld.length; )
          feld[i++] = null;  // gibt die Objekte frei fuer den GC
}

protected void del(int I){
    if (I<0 || I>=length)
        throw new IllegalArgumentException(I+" ausserhalb des Bereichs");
    if (I==0){ feld[offset++] = null; length--;
    }else
    if (I==length-1){ feld[offset+(--length)] = null;
    }else{
        // dieser Fall tritt in dieser Klasse nicht auf,
        // die Methode stammt in dieser Form aus der Sequenz-Klasse
        System.arraycopy(feld,offset+I+1, feld,offset+I, length-I-1);
        feld[offset+(--length)] = null;
    }
}

/**
 * Liefert eine Enumeration fuer dieses Objekt.
 * Waehrend der Verwendung einer Enumeration darf das Objekt
 * nicht veraendert werden - das Verhalten ist sonst undefiniert.
 */
public Enumeration elements() {
    return new Enumeration(){
        int count = 0;
        public boolean hasMoreElements(){
            return count < length;
        }
        public Object nextElement(){
            if (count < length)
                return feld[offset+count++];
            else
                throw new NoSuchElementException();
        }
    };
}

/**
 * Gibt an, ob die Deque leer ist oder nicht.
 */
public boolean isEmpty() {
    return size()<=0;
}

/**
 * Liefert das erste Objekt der Deque.
 */
public Object getFirst() {
    if (length==0)
        throw new NoSuchElementException();
    return feld[offset];
}

/**
 * Liefert das letzte Objekt der Deque.
 */
public Object getLast() {
    if (length==0)
        throw new NoSuchElementException();
    return feld[offset+length-1];
}

/**
 * Fuegt das Objekt vorn an die Deque an.
 */
public void addFirst(Object x){
    modCount++;
    if (offset<=0) resize();
    feld[--offset] = x; length++;
}

/**
 * Fuegt das Objekt hinten an die Deque an.
 */
public void addLast(Object x){
    modCount++;
    if (offset+length>=feld.length) resize();
    feld[offset+(length++)] = x;
}

/**
 * Liefert einen Iterator fuer dieses Objekt.
 * Wird waehrend der Verwendung eines Iterators das Objekt veraendert,
 * dann bricht der Iterator beim naechsten Aufruf von next mit einer
 * ConcurrentModificationException ab.
 */
public Iterator iterator() {
    return new Iterator(){
        int count = 0;
        int expectedModCount = modCount;
        public boolean hasNext(){
            return count < length;
        }
        public Object next(){
            if (count < length){
                try{
                    Object obj = feld[offset+count++];
                    checkForComodification();
                    return obj;
                } catch(IndexOutOfBoundsException e) {
                    checkForComodification();
                    throw new NoSuchElementException();
                }
            }else{
                checkForComodification();
                throw new NoSuchElementException();
            }
        }
        public void remove(){
            throw new UnsupportedOperationException();
        }
        final void checkForComodification() {
            if (modCount != expectedModCount)
            throw new ConcurrentModificationException();
        }
    };
}

/**
 * Liefert das erste Element und entfernt es aus der Deque.
 */
public Object removeFirst() {
    modCount++;
    Object obj = getFirst();
    del(0);
    return obj;
}

/**
 * Liefert das letzte Element und entfernt es aus der Deque.
 */
public Object removeLast() {
    modCount++;
    Object obj = getLast();
    del(length-1);
    return obj;
}

protected void resize() {
    int offset2 = loadstep>0 ? loadstep : (length+2);
    Object[] feld2 = new Object[length+2*offset2];
    System.arraycopy(feld,offset, feld2,offset2, length);
    feld = feld2; offset=offset2;
}

/**
 * Liefert die Anzahl der Elemente in der Deque.
 */
public int size() {
    return length;
}

/**
 * Liefert eine Stringdarstellung der Deque.
 */
public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("[");
    if (length>0) sb.append(feld[offset]);
    for (int i=offset+1; i<offset+length; i++)
        sb.append(",").append(feld[i]);
    sb.append("]");
    sb.append("-(").append(offset+","+length+","+feld.length).append(")");
    return sb.toString();
}
}
