package h2d2c.util;
import java.util.*;

public class PriorityQueue extends ArrayDeque {

/**
 * Vergleicht die Elemente der Deque.
 */
protected Comparator comp;

// Speichert den letzten modCount, bei dem das Feld sortiert war.
// Wenn man Elemente mit insertFirst oder insertLast einfuegt,
// wird die Sortierung zerstoert und das wird spaeter erkannt.
// Methoden, die die Sortierung erhalten, aktualisieren sortModCount.
protected int sortModCount = 0;

public PriorityQueue(Comparator comp) {
  super();
  if (comp==null) throw new IllegalArgumentException("null comparator");
  this.comp = comp;
}

public PriorityQueue(int loadstep, Comparator comp) {
  super(loadstep);
  if (comp==null) throw new IllegalArgumentException("null comparator");
  this.comp = comp;
}

public Object getMin() {
  if (sortModCount != modCount) {
    modCount++;
    Arrays.sort(feld, offset, offset+length, comp);
    sortModCount = modCount;
  }
  return getFirst();
}

public Object removeMin() {
  if (sortModCount != modCount) {
    modCount++;
    Arrays.sort(feld, offset, offset+length, comp);
  }
  Object obj = removeFirst();
  sortModCount = modCount;
  return obj;
}

/**
 * Fuegt das Objekt sortiert ein, laesst doppelte Elemente zu.
 */
public void add(Object obj) {
  add(obj, true);
}

public static final boolean INSERTDOUBLES = true, INSERTNODOUBLES = false;

/**
 * Fuegt das Objekt sortiert ein, laesst doppelte Elemente zu, wenn
 * doubles das angibt.
 * @param double false: Von Objekten, die per Comparator gleich sind,
 * wird nur das erste eingefuegt, die spaereten nicht mehr
 * true: auch gleiche Objekte werden eingefuegt
 */
public void add(Object obj, boolean doubles) {
  if (sortModCount != modCount) {
    if (doubles) {
      // wenn keine Sortierung vorliegt machen wir uns hier nicht die Muehe,
      // das Feld zu sortieren
      addLast(obj);
      return;
    } else {
      // wenn wir keine Dopplungen zulassen wollen, dann muessen wir
      // leider schauen, obs schon drin ist, und dazu sortieren wir das Feld
      modCount++;
      Arrays.sort(feld, offset, offset+length, comp);
    }
  }

  int i = binarySearch(feld, offset, offset+length, obj, comp);
  if (i<0)
    insertAt(-(i+1)-offset, obj);
  else
  if (doubles)
    insertAt(i-offset, obj);
  sortModCount = modCount;
}

protected void insertAt(int I, Object o) {
  modCount++;
  if (I*2>length) {
    if (offset+length>=feld.length) resize();
    System.arraycopy(feld,offset+I, feld,offset+I+1, length-I);
    feld[offset+I] = o; length++;
  }else{
    if (offset<=0) resize();
    System.arraycopy(feld,offset, feld,offset-1, I);
    feld[--offset+I] = o; length++;
  }
}

    // aus java.lang.Arrays modifiziert
    // Suchgrenzen hinzugefuegt
    /**
     * Searches the specified array for the specified object using the binary
     * search algorithm.  The array must be sorted into ascending order
     * according to the specified comparator (as by the <tt>Sort(Object[],
     * Comparator)</tt> method, above), prior to making this call.  If it is
     * not sorted, the results are undefined.  If the array contains multiple
     * elements equal to the specified object, there is no guarantee which one
     * will be found.
     *
     * @param a the array to be searched.
     * @param fromIndex the index of the first element (inclusive) to be
     *        searched.
     * @param toIndex the index of the last element (exclusive) to be searched.
     * @param key the value to be searched for.
     * @param c the comparator by which the array is ordered.  A
     *        <tt>null</tt> value indicates that the elements' <i>natural
     *        ordering</i> should be used.
     * @return index of the search key, if it is contained in the list;
     *	       otherwise, <tt>(-(<i>insertion point</i>) - 1)</tt>.  The
     *	       <i>insertion point</i> is defined as the point at which the
     *	       key would be inserted into the list: the index of the first
     *	       element greater than the key, or <tt>list.size()</tt>, if all
     *	       elements in the list are less than the specified key.  Note
     *	       that this guarantees that the return value will be &gt;= 0 if
     *	       and only if the key is found.
     * @throws ClassCastException if the array contains elements that are not
     *	       <i>mutually comparable</i> using the specified comparator,
     *	       or the search key in not mutually comparable with the
     *	       elements of the array using this comparator.
     * @see Comparable
     * @see #sort(Object[], Comparator)
     */
    public static int binarySearch(Object[] a,
    int fromIndex, int toIndex,  // Suchgrenzen hinzugefuegt
    Object key, Comparator c) {
        int low = fromIndex;
        int high = toIndex-1;

	while (low <= high) {
	    int mid =(low + high)/2;
	    Object midVal = a[mid];
	    int cmp = c.compare(midVal, key);

	    if (cmp < 0)
		low = mid + 1;
	    else if (cmp > 0)
		high = mid - 1;
	    else
		return mid; // key found
	}
	return -(low + 1);  // key not found.
    }

}
