package jay;
import algds.c2002.Hamster;

/**
 * Hamster Jay (J)
 * Autor: Robert Keller (rk@e-publicity.de)
 * Dateien: Jay.java, Cell.java, Constants2.java
 * Die Vorgehensweise von Jay wird in readme.html dargestellt.
 * Kurze Erklrungen finden sich am Kopf jeder Methode.
 */
 
public class Jay extends Hamster{

	// Richtungskonstanten
	public final static int N  = 0;
	public final static int NE = 1;
	public final static int SE = 2;
	public final static int S  = 3;
	public final static int SW = 4;
	public final static int NW = 5;

	// Labyrinth
	Cell[][] lab;
	
	// aktuelle Rekursionstiefe
	int nowDepth;
	
	// aktuelle Positionsvariablen
	int nowx, nowy, dir, dirVirt; 
	
	// 2 Stze an Sicherungsvariablen
	int saveNowy, saveNowx, saveDir;
	int saveNowyBack, saveNowxBack;

public void run() {
	// Hauptmethode startet Jay
	init();
	go();
	comeHere();
	dropHere(CORNCARRY_MAX);
}

private void init(){
	// initialisiert das Labyrinth (61*61) mit "Cell"s
	lab = new Cell[61][61];
	for (int i=0;i<=60;i++)
		for (int j=0;j<=60;j++)
			lab[i][j] = new Cell();
	
	// Startzelle: 30 * 30
	nowx = 30; 
	nowy = 30;
	saveNowx = 30;
	saveNowy = 30;
	dir = N;
	dirVirt = N;
	nowDepth = -1;
}

 
private void go(){
	// Hauptlaufmethode:
	// durchsucht jedes Labyrinth komplett
	boolean goOn = false;
	nowDepth++;
	
	inspectCell();
	
	// 1. Geradeaus
	if (!blockedWay()){
		forwardMe();
		go();
	}
	
	// 2. Links
	turnMeVirt(TURN_LEFT);
	if (!blockedWay()){
		goHomeOrOn();
		go();
	}
	
	// 3. doppelt Links
	turnMeVirt(TURN_LEFT);
	if (!blockedWay()){
		goHomeOrOn();
		go();
	}

	// 4. Rechts
	turnAroundVirt();
	if (!blockedWay()){
		goHomeOrOn();
		go();
	}

	// 5. doppelt Rechts
	turnMeVirt(TURN_RIGHT);
	if (!blockedWay()){
		goHomeOrOn();
		go();
	}
	
	// 6. Rckweg
	turnMeVirt(TURN_RIGHT);
	if (!atHomeMe()){
		if (enoughCornComeHere()){
			comeHere();
			turnMe();
			if(getAvailableCorn() > 0){
				take(CORNCARRY_MAX);
				while (getAvailableCorn() > 0 && getCarriedCorn() == CORNCARRY_MAX && !goOn){
					if (lab[nowy][nowx].distance[0] < 100){
						goHome();
						if (!atHome()) take(CORNCARRY_MAX);
					} else {
						goOn = true;
					}
				}
			}
		}
		forwardMeVirt();
		turnAroundVirt();
		nowDepth--;
		setDistance(nowy,nowx);
		if (atHome()) dropHere(CORNCARRY_MAX);
	} else {
		if (!blockedWay()){	
			goHomeOrOn();
			go();
		}
	}
}
					
private void goHomeOrOn(){
	comeHere();		// lsst Jay zum aktuellen Feld laufen
	dropHere(CORNCARRY_MAX);
	lab[nowy][nowx].corn = getAvailableCorn();
	turnMe();		// dreht Jay, damit die reale Richtung 
					// mit der viruellen bereinstimmt
	forwardMe();	// ein Schritt vorwrts
}

private void goHome(){
	// Hllmethode zur Initialiserung des Heimwegs
	saveDir = dir;
	saveNowyBack = nowy;
	saveNowxBack = nowx;
	if (verbose()) logln("Bringe Mais nach Hause.");
	bringHome(); // luft nach Hause und zurck
	if (verbose()) logln("Setze Suche fort.");
	
	while (dir != saveDir){
		// dreht sich wieder in Ausgangsrichtung
		if (dir > saveDir) turnMeReal(TURN_LEFT);
		if (dir < saveDir) turnMeReal(TURN_RIGHT);
	}
}

private void bringHome(){
	// luft nach Hause und zurck
	int backDir = (dir + 3) % 6;
	if (nowy == 30 && nowx ==30){
		// wenn zu Hause
		dropHere(getCarriedCorn());
		turnAround();
		forwardMe();
	} else {
		// sonst laufe rekursiv auf krzestem Weg nach Hause
		while (dir != getMinDirection(nowy,nowx,0)){
			if (dir > getMinDirection(nowy,nowx,0)) turnMeReal(TURN_LEFT);
			if (dir < getMinDirection(nowy,nowx,0)) turnMeReal(TURN_RIGHT);
		}
		forwardMe();
		bringHome();
		if (nowy == saveNowyBack && nowx == saveNowxBack){
		} else {
			while (dir != backDir){
				if (dir > backDir) turnMeReal(TURN_LEFT);
				if (dir < backDir) turnMeReal(TURN_RIGHT);
			}
			forwardMe();
		}
	}
}

private void comeHere(){
	// bringt Jay auf dem krzesten Weg vom aktuellen Feld 
	// zum aktuellen virtuellen Feld
	
	if (!atHome()) take(CORNCARRY_MAX);
	lab[saveNowy][saveNowx].corn = getAvailableCorn();
	if (nowy == saveNowy && nowx == saveNowx){
		// wenn angekommen: Stackabbau
	} else {
		// sonst gehe in Minimumrichtung zum Feld mit der aktuellen Rekursionstiefe nowDepth
		// hnlich wie bringHome, welche sich auf das Feld der Rekursionstiefe 0 (Basis) bezog
		while (dir != getMinDirection(saveNowy,saveNowx,nowDepth)){
			if (dir > getMinDirection(saveNowy,saveNowx,nowDepth)) turnMeReal(TURN_LEFT);
			if (dir < getMinDirection(saveNowy,saveNowx,nowDepth)) turnMeReal(TURN_RIGHT);
		}
		forwardMeSave();
		comeHere();
	}
}

private boolean enoughCornComeHere(){
	// lohnt es sich hierher zu kommen
	// d.h: Mais und weniger als 100 Schritt nach Hause
	return (lab[nowy][nowx].corn > 0 && lab[nowy][nowx].distance[0] < 100);
}

private boolean blockedWay(){
	// ist der Weg blockiert
	// d.h: Mauer oder schon besuchtes Feld voraus
	
	if (!isWall(nowy,nowx,dirVirt)) visitNextVirt(nowy,nowx,dirVirt);
	
	return (isWall(nowy,nowx,dirVirt) || nextVisited(nowy,nowx,dirVirt)/* || lab[nowy][nowx].distance[0] > 210*/);
}

private boolean isWall(int ypos, int xpos, int direction){
	// ist die Feldseite voraus vermauert?
	return lab[ypos][xpos].isWall(direction);
}

private boolean nextVisited(int ypos, int xpos, int direction){
	// wurde das nchste Feld schon einmal besucht?
	switch(direction){
		case N:  return lab[ypos-1][xpos-1].visited;
		case NE: return lab[ypos-1][xpos].visited;
		case SE: return lab[ypos][xpos+1].visited;
		case S:  return lab[ypos+1][xpos+1].visited;
		case SW: return lab[ypos+1][xpos].visited;
		case NW: return lab[ypos][xpos-1].visited;
		default: if (verbose()) logln("ERROR in nextVisited!");		
	}
	return false;			
}

private void visitNextVirt(int ypos, int xpos, int direction){
	// besucht das nchste Feld virtuell:
	// wenn schon die 5 abgewandten Seiten bekannt (vermauert oder besucht von der anderen Seite)
	// und kein Mais auf Feld, dann markiere Feld als schon besucht
	
	int oldY = ypos;
	int oldX = xpos;
	boolean setVisited = true;
	switch(direction){
		case N:  ypos--;xpos--;break;
		case NE: ypos--;break;
		case SE: xpos++;break;
		case S:  ypos++;xpos++;break;
		case SW: ypos++;break;
		case NW: xpos--;break;
		default: if (verbose()) logln("ERROR in visitNextVirt!");		
	}
	if (!lab[ypos][xpos].seenCorn){
		for (int i=0;i<6;i++){
			if ((i + 3) % 6 != direction){
				setVisited = setVisited && (lab[ypos][xpos].isWall(i) || nextVisited(ypos,xpos,i));
			}
		}
		if (setVisited){
			lab[ypos][xpos].visited = true;
		}
	}
}

private void turnAround(){
	// 3 zusammengefasste Drehbefehle (siehe turnMeReal)
	turnMeReal(TURN_RIGHT);
	turnMeReal(TURN_RIGHT);
	turnMeReal(TURN_RIGHT);
}

private void turnAroundVirt(){
	// 3 zusammengefasste Drehbefehle (siehe turnMeVirt)
	turnMeVirt(TURN_RIGHT);
	turnMeVirt(TURN_RIGHT);
	turnMeVirt(TURN_RIGHT);
}

private void turnMe(){
	// dreht Jay bis er in die aktuelle virtuelle Richtung schaut
	while (dir != dirVirt){
		if (dir > dirVirt){
			dir--;
			turn(TURN_LEFT);
		} else {
			dir++;
			turn(TURN_RIGHT);
		}			
	}
}

private void turnMeVirt(int to){
	// dreht Jay virtuell
	// wichtig fr Bewegungen auf der Karte, ohne reale Bewegungen
	switch(to){
		case TURN_LEFT: 
			if (dirVirt == N) dirVirt = NW;
			else dirVirt -= 1;
		break;
		case TURN_RIGHT:
			if (dirVirt == NW) dirVirt = N;
			else dirVirt += 1;
		break;
	}
}

private void turnMeReal(int to){
	// dreht Jay wirklich
	switch(to){
		case TURN_LEFT: 
			if (dir == N) dir = NW;
			else dir -= 1;
		break;
		case TURN_RIGHT:
			if (dir == NW) dir = N;
			else dir += 1;
		break;
	}
	turn(to);
}

public void forwardMe(){
	// protokolliert Vorwrtsbewegungen und setzt Jay vorwrts
	switch(dir){
		case N:  nowy--;nowx--;break;
		case NE: nowy--;break;
		case SE: nowx++;break;
		case S:  nowy++;nowx++;break;
		case SW: nowy++;break;
		case NW: nowx--;break;
		default: if (verbose()) logln("ERROR in forwardMe!");
	}
	saveNowy = nowy;
	saveNowx = nowx;
	forward();
}

public void forwardMeVirt(){
	// lsst Jay auf der Karte den Weg virtuell zurcklaufen
	switch(dirVirt){
		case N:  nowy--;nowx--;break;
		case NE: nowy--;break;
		case SE: nowx++;break;
		case S:  nowy++;nowx++;break;
		case SW: nowy++;break;
		case NW: nowx--;break;
		default: if (verbose()) logln("ERROR in forwardMeVirt!");
	}
}

public void forwardMeSave(){
	// setzt Jay vorwrts
	// verndert dabei nicht die aktuellen Positionen,
	// sondern arbeitet mit dem Sicherungssatz der Positionen
	// genutzt von comeHere();
	
	switch(dir){
		case N:  saveNowy--;saveNowx--;break;
		case NE: saveNowy--;break;
		case SE: saveNowx++;break;
		case S:  saveNowy++;saveNowx++;break;
		case SW: saveNowy++;break;
		case NW: saveNowx--;break;
		default: if (verbose()) logln("ERROR in forwardMeSave!");
	}
	forward();
}

public void inspectCell(){
	// inspiziert jedes Feld sobald ein neues Feld betreten wird
	
	lab[nowy][nowx].visited = true;
	lab[nowy][nowx].corn = getAvailableCorn();
	
	for (int i=N; i<=NW; i++){
		if ((lookForward() & WALLMASK) == WALL){
			setWall();
		}
		if ((lookForward() & CORNMASK) == CORN){
			switch(dir){
				case N :lab[nowy-1][nowx-1].seenCorn = true;
				case NE:lab[nowy-1][nowx].seenCorn = true;
				case SE:lab[nowy][nowx+1].seenCorn = true;
				case S :lab[nowy+1][nowx+1].seenCorn = true;
				case SW:lab[nowy+1][nowx].seenCorn = true;
				case NW:lab[nowy][nowx-1].seenCorn = true;
			}
		}
		turn(TURN_LEFT);
		if (dir == 0) dir = 5;
		else dir--;
	}
	setDistance(nowy,nowx);
}

private void setWall(){
	// setzt durch Aufruf von inspectCell auf der Karte im akuellen Feld/Richtung eine Mauer
	// setzt eine Mauer auch auf dem Nachbarfeld (der abgewandten Seite)
	lab[nowy][nowx].setWall(dir);
	switch(dir){
		case N:  lab[nowy-1][nowx-1].setWall(S);break;
		case NE: lab[nowy-1][nowx].setWall(SW);break;
		case SE: lab[nowy][nowx+1].setWall(NW);break;
		case S:  lab[nowy+1][nowx+1].setWall(N);break;
		case SW: lab[nowy+1][nowx].setWall(NE);break;
		case NW: lab[nowy][nowx-1].setWall(SE);break;
		default: if (verbose()) log("ERROR in setWall!\n");
	}
}

private boolean atHomeMe(){
	// ist Jay virtuell zu Hause
	return (nowy == 30 && nowx == 30);
}

private void setDistance(int ypos, int xpos){
	// registriert alle Entfernungen bis zur aktuellen Rekursionstiefe
	for (int i=0; i<=nowDepth;i++){
		lab[ypos][xpos].distance[i] = getMinDistance(ypos,xpos,i);
	}
}

private int getMinDistance(int ypos, int xpos, int depth){
	// sucht das begehbare Nachbarfeld, das bezogen auf z.B. die Basis die krzeste Entfernung hat
	// und weist dem aktuellen Feld den Wert + 1 zu
	int tempDistance = 10000;
	if (nowDepth == depth) tempDistance = 0;
	if (!lab[ypos][xpos].isWall(0) && lab[ypos-1][xpos-1].visited && lab[ypos-1][xpos-1].distance[depth] < tempDistance-1) tempDistance = lab[ypos-1][xpos-1].distance[depth]+1;
	if (!lab[ypos][xpos].isWall(1) && lab[ypos-1][xpos].visited && lab[ypos-1][xpos].distance[depth] < tempDistance-1) tempDistance = lab[ypos-1][xpos].distance[depth]+1;
	if (!lab[ypos][xpos].isWall(2) && lab[ypos][xpos+1].visited && lab[ypos][xpos+1].distance[depth] < tempDistance-1) tempDistance = lab[ypos][xpos+1].distance[depth]+1;
	if (!lab[ypos][xpos].isWall(3) && lab[ypos+1][xpos+1].visited && lab[ypos+1][xpos+1].distance[depth] < tempDistance-1) tempDistance = lab[ypos+1][xpos+1].distance[depth]+1;
	if (!lab[ypos][xpos].isWall(4) && lab[ypos+1][xpos].visited && lab[ypos+1][xpos].distance[depth] < tempDistance-1) tempDistance = lab[ypos+1][xpos].distance[depth]+1;
	if (!lab[ypos][xpos].isWall(5) && lab[ypos][xpos-1].visited && lab[ypos][xpos-1].distance[depth] < tempDistance-1) tempDistance = lab[ypos][xpos-1].distance[depth]+1;
	if (tempDistance == 10000) if (verbose()) logln("Error: getMinDistance");
	return tempDistance;
}

private int getMinDirection(int ypos, int xpos, int depth){
	// gibt Jay die Richtung mit der geringsten Entfernung zu einem bestimmten Feld zurck
	int tempDistance = 10000;
	int minDir = 6;
	
	if (!lab[ypos][xpos].isWall(0) && lab[ypos-1][xpos-1].visited && lab[ypos-1][xpos-1].distance[depth] < tempDistance){
		tempDistance = lab[ypos-1][xpos-1].distance[depth];
		minDir = 0;
	}
	if (!lab[ypos][xpos].isWall(1) && lab[ypos-1][xpos].visited && lab[ypos-1][xpos].distance[depth] < tempDistance){
		tempDistance = lab[ypos-1][xpos].distance[depth];
		minDir = 1;
	}
	if (!lab[ypos][xpos].isWall(2) && lab[ypos][xpos+1].visited && lab[ypos][xpos+1].distance[depth] < tempDistance){
		tempDistance = lab[ypos][xpos+1].distance[depth];
		minDir = 2;
	}
	if (!lab[ypos][xpos].isWall(3) && lab[ypos+1][xpos+1].visited && lab[ypos+1][xpos+1].distance[depth] < tempDistance){
		tempDistance = lab[ypos+1][xpos+1].distance[depth];
		minDir = 3;
	}
	if (!lab[ypos][xpos].isWall(4) && lab[ypos+1][xpos].visited && lab[ypos+1][xpos].distance[depth] < tempDistance){
		tempDistance = lab[ypos+1][xpos].distance[depth];
		minDir = 4;
	}
	if (!lab[ypos][xpos].isWall(5) && lab[ypos][xpos-1].visited && lab[ypos][xpos-1].distance[depth] < tempDistance){
		tempDistance = lab[ypos][xpos-1].distance[depth];
		minDir = 5;
	}
	if (tempDistance == 10000) {
		if (verbose()) logln("Error: getMinDirection");
	}
	return minDir;
}

/** Autor(en) des Hamsters. */
public String getHamsterAuthor() { return "Robert Keller (WIF) OvG Magdeburg"; }

/** Datum der Hamsterversion. */
public String getHamsterDate() { return "24.04.2002"; }

/** Hamsterversion. */
public String getHamsterVersion() { return "0.5"; }

} // Jay
