package mannie;
/**	Hier werden alle Informationen ueber die bekannte Welt
 *	des Hemster gespeichert und verwaltet; sie sind staendig
 *	verfuegbar durch die gewaehlten Datenstruktur: Array
 */
public class Map{
	//private Instanz-Variablen
	private int distcounter;
	private Field[][] map=new Field[70][70];	//map
	private Coord HOME=new Coord(35,35);	//home
	private Coord position=HOME;			//position des hamsters
	private int dir;

	//Konstruktor zur Erzeugung einer
	//Map-Instanz mit gleichzeitiger Belegung
	//aller Felder mit default-Werten
	public Map(int x, int y){
		map=new Field[x][y];
		initMap();
	}
	
	//Initialisierung der internen Karte
	public void initMap(){
		for(int i=0; i<70; i++)
	    	for(int j=0; j<70; j++)
				//Distanzen werden standardmaessig mit dist=200
				map[i][j]=new Field(0,200);
		map[35][35].setDist(0);
	}
	
	//Praedikat
	private boolean isHome(int isx, int isy){
		return (isx==35 && isy==35);
	}
	
	//Accessor-Methoden, kuezer als getX()--> x()
	public int x(){return position.getX();}
	public int y(){return position.getY();}
	public int distance(){return map[x()][y()].getDist();}
	public int getDist(int isx,int isy){return map[isx][isy].getDist();}
	
	
	/**EXCEPTION
	 * steht mannie am rand, dann exceptionmglich!!!!!!!!!!!
	 */
	private int[]getAllDists(int x,int y){
		int[] dist = new int[6];
		dist[0]=map[x][y+1].getDist();//oben
		dist[1]=map[x+1][y].getDist();//obenrechts
		dist[2]=map[x+1][y-1].getDist();//untenrechts
		dist[3]=map[x]  [y-1].getDist();//unten
		dist[4]=map[x-1][y].getDist();//untenlinks
		dist[5]=map[x-1][y+1].getDist();//untenrechts
		return dist;
	}
	
	//nur bestimmt Richtung
	public int getDistAt(int x,int y,int d){
		if(d==0) return map[x][y+1].getDist();//oben
		if(d==1) return map[x+1][y].getDist();//obenrechts
		if(d==2) return map[x+1][y-1].getDist();//untenrechts
		if(d==3) return map[x]  [y-1].getDist();//unten
		if(d==4) return map[x-1][y].getDist();//untenlinks
		if(d==5) return map[x-1][y+1].getDist();//untenrechts	
		return -1;
	}
	
	//speziell
	//holt Wandeigenschaft von aktuellem Feld
	//zu aktueller Richtung
	public boolean getWall(){return map[x()][y()].getWall(dir);}
	public boolean getWallAt(int x,int y,int direction){return map[x][y].getWall(direction);} 
	public int getMinDistWay(int isx,int isy,int distance){
		int[] dist = getAllDists(isx,isy);
		for(int i=0;i<6;i++)
			if(!getWallAt(isx,isy,i)&&dist[i]<distance)return i;
		return 100;
	}
	public int getMaxDistWay(int isx,int isy,int distance){
		int[] dist = getAllDists(isx,isy);
		for(int i=0;i<6;i++)
			if(!getWallAt(isx,isy,i)&&dist[i]>distance)return i;
		return 100;
	}
	
	//Umkehrabbildung zur normalen Abbildung
	//"Hamster_explore" --> Map
	public int getDir(int x, int y){
		if(x==x() 	&& y==y()+1)	return 0;
		if(x==x()+1 && y==y()) 		return 1;
		if(x==x()+1 && y==y()-1)	return 2;
		if(x==x() 	&& y==y()-1)	return 3;
		if(x==x()-1 && y==y())		return 4;
		if(x==x()-1 && y==y()+1)	return 5;
		return 100;
	}
	public char getMarkAt(int x,int y,int dir){
		char c='-';
		if(dir==0&&!getWallAt(x,y,dir)){c=map[x][y+1].getMark();}
		if(dir==1&&!getWallAt(x,y,dir)){c=map[x+1][y].getMark();}
		if(dir==2&&!getWallAt(x,y,dir)){c=map[x+1][y-1].getMark();}
		if(dir==3&&!getWallAt(x,y,dir)){c=map[x][y-1].getMark();}
		if(dir==4&&!getWallAt(x,y,dir)){c=map[x-1][y].getMark();}
		if(dir==5&&!getWallAt(x,y,dir)){c=map[x-1][y+1].getMark();}
		return c;
	}
	public int getWay(int isx,int isy){
		for(int i=0;i<6;i++){
			if(!getWallAt(isx,isy,i) && getMarkAt(isx,isy,i)!='s'&&getMarkAt(isx,isy,i)!='g' && getDistAt(isx,isy,i)!=200)
				return i;
		}
		return 100;
	}
	
	//Setz-Methoden
	public void setDir(int direction){dir=direction;}
	//allgemein
	public void setDist(int x, int y,int d){map[x][y].setDist(d);}
	//...speziell
	public void setDist(int distance){distcounter=distance;}
	public void incDist(){
		distcounter++;
		if(map[x()][y()].getDist()>distcounter)
			map[x()][y()].setDist(distcounter);
		else{
			distcounter = map[x()][y()].getDist(); 
		}
	}
	
	public void setMark(int x,int y,char c){
		map[x][y].setMark(c);
	}
	
	//Abbildungsvorschrift fuer Map
	public void updateCoords(){
		switch(dir){
			case 0: position.setCoords(x(),y()+1); break;
			case 1:	position.setCoords(x()+1,y()); break;
			case 2:	position.setCoords(x()+1,y()-1); break;
			case 3: position.setCoords(x(),y()-1); break;
			case 4: position.setCoords(x()-1,y()); break;
			case 5: position.setCoords(x()-1,y()+1); break;
		}
    }
    //alle Waende
    public void setWalls(boolean[] w){map[x()][y()].setWalls(w);}
	
	//iterator-bewegung
	//Bewegung eines virtuellen Hamsters von Feld (x,y)
	//in Richtung dir
	public int[] move(int x,int y,int dir){
		int[] coord = new int[2];
		if(dir==0){coord[0]=x;		coord[1]=y+1;}
		if(dir==1){coord[0]=x+1;	coord[1]=y;}
		if(dir==2){coord[0]=x+1;	coord[1]=y-1;}
		if(dir==3){coord[0]=x;		coord[1]=y-1;}
		if(dir==4){coord[0]=x-1;	coord[1]=y;}
		if(dir==5){coord[0]=x-1;	coord[1]=y+1;}
	return coord;
	}
	
	public void print(String a){
		System.out.println(a);
		
	}
	//loeschen aller Flags mapweit
	private void delMarks(){
		for(int i=0;i<70;i++)
			for(int j=0;j<70;j++)
				map[i][j].setMark('-');
	}
	
	// s=sackgasse
	// -=keine Markierung
	// g=gegangen
	
	//Praedikat Kreuzung(Feld)
	public boolean cross(int isx,int isy){
		int walls=0;
		for(int i=0;i<6;i++)if(getWallAt(isx,isy,i)==true)walls++;
		return(walls<4);
	}
	
	//"Kruemelspur" c bis zur naechsten Kreuzung folgen
	public int followMarks(char c,int isx,int isy){
		for(int i=0;i<6;i++)
			if(getMarkAt(isx,isy,i)==c && !cross(isx,isy))return i;
		return 100;
	}
	
	//Spur c bis zum Ende verfolgen
	public int followMark(char c,int isx,int isy){
		for(int i=0;i<6;i++)
			if(getMarkAt(isx,isy,i)==c)return i;
		return 100;
	}
	
	//Hilfs-Methode fuer untenstehende Methode
	/*public int[] getWayTo(int isx, int isy, int x, int y){
		int[] temp=new int[200];
		for(int i=0;i<200;i++)temp[i]=-1;
		temp = getWayTo(isx,isy,x,y,temp,0);
		for(int i=0;i<200;i++)System.out.print(temp[i+1]);
	return temp;
	}
	*/
	//rekursive Methode, die die Koordinaten des Wegen von
	//(isx,isy) bis (x,y) zurueckgibt
	public int[] getWayTo(int isx, int isy, int x, int y,int[] dirarray,int i){
    	int distdort = map[x][y].getDist();
    	int disthier = map[isx][isy].getDist();
    	if(!isHome(isx,isy)){
    		int[] help = move(isx,isy,getMinDistWay(isx,isy,disthier));
    		dirarray[i]=(getMinDistWay(isx,isy,disthier)+3)%6;
    		isx = help[0];
    		isy = help[1];
    		getWayTo(isx,isy,x,y,dirarray,++i);
    	}
    	
    	//System.out.print(dirarray[i]);
    	return dirarray;
    	
    }
    
    /*public void getWayTo(int isx,int isy,int x,int y){
		map[isx][isy].setMark('g');
		int walkdir = getWay(isx,isy);
		
		if(walkdir!=100){
			int[] help = move(isx,isy,walkdir);
			isx=help[0];isy=help[1];
			print(isx + " : "+ isy);
			getWayTo(isx,isy,x,y);
		}
		
		if(walkdir==100){
			int follow = followMarks('g',isx,isy);
			if(follow!=100){
				map[isx][isy].setMark('s');
				int[] help = move(isx,isy,follow);
				isx=help[0];isy=help[1];
				print("100");
				getWayTo(isx,isy,x,y);
			}
		}
		
	}
	public int[] getWayTo(int isx,int isy,int x,int y,int[] dirarray,int i){
		int distance=map[x][y].getDist();
		
		//if(x==position.getX()&&y==position.getY())return dirarray; 
		
		//distanz des aktuellesn feldes < iterator-distanz
		if(distance > map[isx][isy].getDist()){
			dirarray[i]=getMaxDistWay(isx,isy,map[isx][isy].getDist());
			int[] help = move(isx,isy,dirarray[i]);
			isx=help[0];
			isy=help[1];
			//System.out.println("< X "+isx+" Y "+isy + "aktuelle Dist :"+ map[isx][isy].getDist());
			//System.out.println("maxDist "+getMaxDistWay(isx,isy,map[isx][isy].getDist()));
			getWayTo(isx,isy,x,y,dirarray,++i);
		}
		if(distance <  map[isx][isy].getDist()){
			dirarray[i]=getMinDistWay(isx,isy,map[isx][isy].getDist());
			int[] help = move(isx,isy,dirarray[i]);		
			isx=help[0];
			isy=help[1];
			//System.out.println("< X "+isx+" Y "+isy + "aktuelle Dist :"+ map[isx][isy].getDist());
			//System.out.println("minDist "+getMinDistWay(isx,isy,map[isx][isy].getDist()));
			getWayTo(isx,isy,x,y,dirarray,++i);
		}
		if(distance == map[isx][isy].getDist()){
			dirarray[i]=getMinDistWay(isx,isy,map[isx][isy].getDist());	
			
		}
		
		return dirarray;
	}*/
}