package mort;

import java.util.*;

public class Node {
  public Pos pos;              // backwards link into MazeMap
  int wall[]={-1,-1,-1};       // -1: unknown, 0: no, 1: yes, directions 0 to 2 are saved only 
  public int cornhere=-1;      // -1: don't know, 0: no corn, 1..100: yes
  public int distance=-1;      // distance to home field
  public int distpass=-1;      // pass no. needed for spawn()

  private int walloffsx[]={0,0,0,0,-1,-1}; // which wall is saved in which node, each node only
  private int walloffsy[]={0,0,0,-1,0,1};  // saves three walls...

  public void setWall(int which, boolean state) { // setzt eine Wand in Richtung which auf true / false
    Pos p=(Pos)pos.clone();
    p.x+=walloffsx[which];
    p.y+=walloffsy[which];
    Node n=Mort.map.getNode(p);
    n.wall[which%3]=state? 1: 0;
  }
  
  public boolean getWall(int which) {
    Pos p=(Pos)pos.clone();
    p.x+=walloffsx[which];
    p.y+=walloffsy[which];
    Node n=Mort.map.getNode(p);
    return (n.wall[which%3]==1)? true: false;
  }
  
  public boolean getWallDefault(int which, boolean d) { // liefert d zurck, wenn unbekannt
    if (!knowWall(which)) return d;
    Pos p=(Pos)pos.clone();
    p.x+=walloffsx[which];
    p.y+=walloffsy[which];
    Node n=Mort.map.getNode(p);
    return (n.wall[which%3]==1)? true: false;
  }
  
  public boolean knowWall(int which) {
    Pos p=(Pos)pos.clone();
    p.x+=walloffsx[which];
    p.y+=walloffsy[which];
    Node n=Mort.map.getNode(p);
    return (n.wall[which%3]!=-1)? true: false;
  }
  
  public boolean knowAllWalls() {
    for (int i=0;i<6;i++) if (!knowWall(i)) return false;
    return true;
  }
  
  public boolean isInteresting() {
    if (exposed()) return false;
    return true;
  }
  
  public boolean isLeaf() {
    if (Mort.m.actpos.equals(pos)) return false;
    if (cornhere==-1) return false;
    if (!exposed()) return false;
    for (int i=0;i<6;i++) if (!getWallDefault(i,false)) if (getNext(i).distance>distance) return false;
    int corn=cornhere;
    if (corn>Mort.m.CORNCARRY_MAX) corn=Mort.m.CORNCARRY_MAX;
    if (corn*5>distance) return false;
    return true;
  }
  
  public int countInteresting() {
    int k=0;
    for (int i=0;i<6;i++) if (!getWall(i)) k++;
    return k;
  }
  
  public boolean exposed() {
    if (cornhere==-1) return false;
    for (int i=0;i<6;i++) if (!knowWall(i) && (getNext(i).cornhere==-1)) return false;
    return true;
  }
  
  public Node getNext(int which) {
    Pos p=(Pos)pos.clone();
    p.setNextPos(which);
    return Mort.m.map.getNode(p);
  }
  
  public int unexposedNeighbors() {
    int n=0;
    for (int i=0;i<6;i++) if (!getNext(i).exposed()) n++;
    return n;
  }
  
  public int anyFurther() {
    int spare=-1;
    Vector l=new Vector();
    for (int k=0;k<6;k++) {
      int i=(k+Mort.m.actdir+1)%6;
      if (!getWall(i)) if (getNext(i).distance==distance+1) {
        if (getNext(i).isInteresting()) l.add(new Integer(i));
        if (spare==-1) spare=i;
      }
    }
    if (l.size()!=0) { // take the unexposed neighbor with the largest number of unexposed neighbors
      int value=-1;
      for (int i=0;i<l.size();i++) if (getNext(((Integer)l.elementAt(i)).intValue()).unexposedNeighbors()>value) {
        spare=((Integer)l.elementAt(i)).intValue(); 
        value=getNext(((Integer)l.elementAt(i)).intValue()).unexposedNeighbors();
      }
    }
    return spare;
  }
  
  public int anySameLevelInteresting() {
    for (int i=0;i<6;i++) if (!getWall(i)) if (getNext(i).distance==distance) if (getNext(i).isInteresting()) return i;
    return -1;
  }
  
  public int homeHeadingNeighbor() {
    int spare=-1;
    for (int k=0;k<6;k++) {
      int i=(k+Mort.m.actdir)%6;
      if (!getWall(i)) if (getNext(i).distance==distance-1) {
        if (getNext(i).isInteresting()) return i;
        if (spare==-1) spare=i;
      }
    }
    return spare;
  }
  
}
