package neo;

import algds.c2002.Hamster;
import java.util.*;

public class Neo extends Hamster {
  
  //die Karte des Hamsters
  LabMap Map;
  //Position und Richtung des Hamsters
  LabPosition Pos;
  LabDirection Dir;
  
  //Daten des letzten Knotens, den der hamster passiert hat
  LabPosition prevPos;
  LabDirection prevDir;
  
  //die Position des Heimatfeldes
  LabPosition homePos;

  //die Kosten, die wir beim Laufen so erzeugen
  int cost;
  
  //den Graph brauchen wir, um die Wege zu finden
  Graph LabGraph;

  //die todo-Liste des Hamsters
  TodoList todoList = new TodoList();
  
  //ein paar Infos ueber die Hamster-Klasse
  public String getHamsterAuthor() {return "Stefan Haun <mail@tuxathome.de>";}
  public String getHamsterDate()  {return "22.04.2002";}
  public String getHamsterVersion() {return "0.0.2";}

  public void run() {
    //Karte anlegen
    Map = new LabMap(100,100); //mit der groesse muss ich mir noch was ausdenken
    //hamster startet in der Mitte
    Pos = new LabPosition(Map.getWidth()/2, Map.getHeight()/2);
    Dir = new LabDirection();
    
    //das Heimatfeld
    homePos = new LabPosition(Pos);
    
    //Graph anlegen
    LabGraph = new Graph();
    
    //das Heimatfeld ist auf jeden Fall ein Knoten!!
    LabGraph.addNode(Pos);
    //der zuletzt besuchte Knoten
    prevPos = new LabPosition(Pos);
    prevDir = new LabDirection(Dir);
    
    //der rest ist einfach: schritt machen, mais checken ;-)
    while (step()) 
      checkCorn();

    logln("Fertig! :-)");
  
  } //void run()

  //die folgende routine bestimmt die verhaltensweise des hamsters und
  //wird nach jedem schritt aufgerufen
  //der rueckgabewert bestimmt, ob der hamster fertig ist
  boolean step() {
    //wenn unerforscht -> erstmal umgucken
    if (!Map.isExplored(Pos)) LookAround();

    //wenn wir in einer sackgasse sind
    if (Map.isEnd(Pos) && (getAvailableCorn() == 0) && !atHome()) {
        
      //feld blockieren - nochmal muessen wir hier nicht rein
      Map.blockField(Pos);
        
      //und den ausgang suchen
      LabDirection dir = new LabDirection();
      while (!Map.canGo(Pos, dir)) 
        dir.turn(TURN_LEFT);
      turnTo(dir);
        
      //Knoten werden nicht mehr geloescht! das ist im Graph nicht
      //implementiert
      //wenn der Hamster auf ein blockiertes Feld will, ist er sowieso nicht
      //mehr zu retten
      
      //dann machen wir einen schritt und schauen, was uns auf dem naechsten 
      //feld erwartet
      //sackgassen sind wirklich nicht weiter interessant
      forward();
      return true;
    }

    //wenn wir das maul (oder die schnauze?) voll haben, gehts nach haus
    if (getCarriedCorn() == CORNCARRY_MAX) {
      //moeglicherweise muessen wir hierher nochmal zurueck, um das feld
      //zu erforschen
      if (bordersUnexplored(Pos)) todoList.add(Pos, false);
      
      //dazu muessen wir unsere position in den baum mit einfuegen
      if (!LabGraph.isNode(Pos)) PosToGraph();
      //dann bringen wir das zeug nach haus
      goHome();
      
      //danach gehts von vorn weiter
      return true;
    }

    //in einer Gasse muessen wir nur den anderen Ausgang suchen
    //aber nur, wenn es kein knotenpunkt ist - ansonsten koennen wir gleich
    //schauen, was als naechstes zu tun ist
    if (Map.isLane(Pos) && !LabGraph.isNode(Pos)) {
      
      //bevorzugt wollen wir erkunden
      TurnToUnexplored();
      
      //wenn wir keinen Ausgang finden, muessen wir einen suchen
      if (!Map.canGo(Pos, Dir)) {
        LabDirection dir = new LabDirection(Dir.getOpposite());
        do {
          dir.turn(TURN_LEFT);
        } while (!Map.canGo(Pos, dir));
        turnTo(dir);
      } //if

      //dann machen wir wieder einen schritt, der rest interessiert uns nicht 
      //wirklich
      forward();
      return true;
    } //isLane
    
    //Kreuzungen werden im Graphen registriert
    //wenn nicht gleich alles erforscht werden kann, ausserdem in der
    //Todo-List      
    if (Map.isNode(Pos)) {
      //verbindung in die liste aufnehmen
      PosToGraph();
          
      //noch eine Sache: wir muessen checken, ob wir vielleicht irgendwie
      //an einer verbindung vorbeilaufen - dann muss die verbindung auch
      //in den graphen
      //dafuer wieder einmal im kreis drehen
      //alte Position merken
      int d = Dir.getDir();
      LabDirection dir = new LabDirection(d);
      //einmal umschauen
      do {
        LabPosition p = new LabPosition(Pos);
        p.move(dir);
        //wenn wir einen Knoten gefunden haben, tragen wir den ein
        if (Map.isExplored(p) && Map.canGo(Pos, dir) && Map.isNode(p)) {
          LabGraph.addConnection( p, new LabDirection(dir.getOpposite()),
                                  Pos, dir, 1);
        }
        dir.turn(TURN_LEFT);
      } while (dir.getDir() != d);     
      //todoList anpassen
      //die liste loescht den eintrag selbstaendig, wenn er leer ist
      todoList.add(Pos, !bordersUnexplored(Pos));
    } //kreuzung
    
    //jetzt muessen wir uns nur noch ueberlegen, wie es weitergehen soll

    //wenn wir uns eingemauert haben (sollte auf dem home-feld passieren), 
    //dann ist schluss
    if (Map.isUnreachable(Pos)) return false;
    
    //wenn es noch etwas zu erforschen gibt, dann laufen wir vorwaerts
    if (TurnToUnexplored()) 
      forward();
    //sonst schauen wir mal in die todo-liste
    else {
   
      //es scheint sich negativ auf das ergebnis auszuwirken, unerforschte
      //felder gegenueber koerner-feldern zu bevorzugen
      //also lassen wir das!
      
      //erstmal ein wenig aufraeumen. evtl haben wir noch unerforschte felder
      //in der todolist, die sich mittlerweile erledigt haben      
      Iterator ll_it = todoList.iterator();
      while (ll_it.hasNext()) {
        TodoItem it = (TodoItem)(ll_it.next());
        //hat sich der eintrag evtl erledigt?
        if ( !bordersUnexplored(it.getPos()) )
          it.setExplored(true);
        //wenn er jetzt wirklich leer ist (also auch kein mais mehr),
        //loeschen wir ihn
        if (it.isEmpty()) ll_it.remove();
      }
      
      //ansonsten suchen wir uns das feld, das unserer aktuellen Position am
      //naechsten liegt - wieder mit Hilfe des Graphen
      //dass sich die aktuelle position in der todoList befindet, ist
      //nicht moeglich, da:
      //1. wir alles Korn auf dem feld schon aufgesammelt/nach hause gebracht
      //   haben
      //2. wir an dieser stelle der routine nicht landen wuerden, wenn 
      //   noch unerforschte felder anliegen
      
      //wir suchen einfach mal den weg nach Haus. das ergebnis ist 
      //uninteressant, aber der graph speichert ausserdem die entfernungen
      //zu allen anderen knoten
      LabGraph.findPath(Pos, homePos);

      ll_it = todoList.iterator();

      //wohin wollen wir als naechstes?
      LabPosition nextPos=null;
      //der bisher kuerzeste gefundene weg
      int shortest=Graph.MAX_WEIGHT;

      while (ll_it.hasNext()) {
        TodoItem it = (TodoItem)(ll_it.next());
        //es gibt ein paar bedingungen dafuer, dass eine verbindung besser 
        //ist
        //erstmal sollte sie kuerzer sein
        if ( LabGraph.getDist(it.getPos()) < shortest ) {
          //hier koennen jetzt nen haufen einschraenkungen kommen,
          //was wir alles nicht mehr suchen...
          //ich nehme mal die einfache regel, dass die maximal moegliche 
          //zahl zu erreichender punkte mit dem weg groesser null sein muss
          int corn_pts = (it.isExplored()?it.getCorn()*10:200);
          //mehr als 200 Punkte sind auf einmal nicht zu holen!
          if (corn_pts > 200) corn_pts = 200;
          corn_pts -= 2*LabGraph.getDist(it.getPos());
          if ( corn_pts > 0 ) {
            nextPos = new LabPosition(it.getPos());
            shortest = LabGraph.getDist(nextPos);
          } 
        } //if
      } //while

      //wenns noch was zu tun gibt, gehen wir einfach zur gefundenen position,
      if (nextPos != null) 
        gotoPos(nextPos);
      //sonst gehts in den entspurt
      else {
        //wenn wir schon zu haus sind, ist wohl wirklich nichts mehr zu
        //holen. sonst gehen wir nach haus und versuchens da nochmal
        if (atHome()) {
          logln("Nix sinnvolles mehr zu tun...");
          checkCorn();
          return false;
        }
        logln("Nix mehr zu tun...aba wir versuchens von zu Haus aus nochmal");
        goHome();
      }
        
    } //else
    
    //wenn wir hier ankommen, gehts auf jeden fall weiter!
    return true;
    
  } //step

  //die folgende routine behandelt rumliegenden mais
  void checkCorn() {
    // Mais vorhanden?
    if ( (getAvailableCorn() > 0) && !atHome() ) {
        
      //so viele koerner aufnehmen, wie moeglich
      take(getAvailableCorn());

      //den rest auf die todoList
      //leere eintraege werden automatisch geloescht
      todoList.add(Pos, getAvailableCorn());

    } //if
    
    //wann immer wir mal zu haus vorbeikommen, wird der Mais da
    //abgeschmissen
    //koennte nach goHome() passieren ;-)
    if (atHome())
      dropHere(getCarriedCorn());
  } //checkCorn
    

  //die Hamsteraktionen muessen auf die interne Karte abgebildet werden
  //deswegen werden die routinen ueberschrieben
  
  //drehen
  public void turn(int direction) {
    //Hamster drehen
    super.turn(direction);
    //und Richtungsiterator drehen
    Dir.turn(direction);
    //wenn wir auf einem Knoten stehen, muessen wir uns die richtung merken,
    //in der wir ihn verlassen
    if (LabGraph.isNode(Pos))
      prevDir = new LabDirection(Dir);
  }

  //den Hamster in eine bestimmte richtung drehen
  public void turnTo(LabDirection dir) {
    //Differenz zwischen den Richtungen:
    int d = Dir.Difference(dir);
    int t = TURN_RIGHT;
    //vielleicht doch besser links herum drehen?
    if (d>3) {
      t = TURN_LEFT;
      d = 6-d;
    } 
    //Drehung ausfuehren
    for (int i=0; i<d; i++)
      turn(t);
  }

  public boolean forward() {
    //erstma gucken, ob frei ist
    if (!Map.canGo(Pos, Dir)) {
      logln("Kollision vermieden...");
      turn(TURN_LEFT);
      return false;
    }
    //schritt machen
    if (!super.forward()) {
      logln("Boah...da hat der Schritt doch nicht geklappt!!");
      return false;
    }
    //normalerweise muss es hier weitergehen...ob frei ist, wurde ja vorher
    //ueberprueft
    //den schritt auf der karte mitfuehren
    Pos.move(Dir);
    //ausserdem erzeugt jeder schritt Kosten
    cost++;
    //mal gucken, ob wir vielleicht an nem Knoten vorbeikommen
    if (LabGraph.isNode(Pos))
      //verbindung in den graph - falls wir die wirklich noch nicht haben
      PosToGraph();
    //und fertig
    return true;
  }
  
  //nach vorn gucken
  public int lookForward() {
    int v = super.lookForward();
    if ( (v & WALLMASK) == WALL )
      Map.markWall(Pos, Dir);
    return v;
  }
   
  //einmal Umsehen  
  void LookAround () {
    //alte Position merken
    int dir = Dir.getDir();
    //einmal umschauen
    do {
      lookForward();
      turn(TURN_LEFT);
    } while (Dir.getDir() != dir);
    //und Feld als erkundet markieren
    Map.explore(Pos);
  }
  
  boolean TurnToUnexplored() {
    //unsere "Arbeitsrichtung"
    LabDirection dir = new LabDirection(Dir);
    
    //ma gucken, ob was unerforscht ist
    do {
      if ( Map.canGo(Pos, dir) && !Map.isExplored(Pos.getNext(dir)) )  {
          //wenn wir was gefunden haben, drehen wir uns dahin und liefern
          //true zurueck
          turnTo(dir);
          return true;
        }
      //ansonsten beim Feld links weitermachen
      dir.turn(TURN_LEFT);
    } while (dir.getDir() != Dir.getDir());
    
    //wenn wir hier angekommen sind, dann gabs nichts neues
    // -> die Richtung wird beibehalten, false zurueckliefern
    return false;
  }
  
  boolean bordersUnexplored(LabPosition pos) {
    //unsere "Arbeitsrichtung"
    LabDirection dir = new LabDirection(LabDirection.UP);
    
    //ma gucken, ob was unerforscht ist
    do {
      if ( Map.canGo(pos, dir) && !Map.isExplored(pos.getNext(dir)) ) {
          //wenn wir was gefunden haben, liefern wir true zurueck
          return true;
        }
      //ansonsten beim Feld links weitermachen
      dir.turn(TURN_LEFT);
    } while (dir.getDir() != LabDirection.UP);
    
    //wenn wir hier angekommen sind, dann gabs nichts neues
    return false;
  }

  //aktuelle Position in den Graph einfuegen
  void PosToGraph() {
    LabGraph.addConnection(prevPos, prevDir, Pos, 
                           new LabDirection(Dir.getOpposite()), cost);
    prevPos = new LabPosition(Pos);
    prevDir = new LabDirection(Dir);
    cost = 0;
  }
  
  //nach Hause gehen - meistens, um Mais abzulegen
  void goHome() {
    //einfach zum home-feld gehen
    gotoPos(homePos);  
    //und jetzt sollten wir zu haus sein
  }
  
  //zu einer angegebenen Position gehen
  void gotoPos(LabPosition dest) {
    //Pos und dest muessen knoten sein!
    if (!LabGraph.isNode(dest)) {
      logln("Ohoh...den Zielknoten gibts im Graph nicht!!");
      return;
    }
    if (!LabGraph.isNode(Pos)) PosToGraph();
    
    //den Weg suchen
    Stack way = LabGraph.findPath(Pos, dest);
    //und dem angegebenen Pfad folgen
    followPath(way);
    //und jetzt sollten wir da sein
  }
  
  //bis zum naechsten knoten gehen
  //das entspricht dem algorithmus, der einer gasse folgt 
  void gotoNextNode() {
    while (!LabGraph.isNode(Pos)) {
      if (!Map.canGo(Pos, Dir)) {
        LabDirection dir = new LabDirection(Dir.getOpposite());
          do {
            dir.turn(TURN_LEFT);
          } while (!Map.canGo(Pos, dir));
          turnTo(dir);
        } //if
      forward();
    } //while
  }
  
  //einem pfad durch den Graphen folgen
  void followPath(Stack path) {
    //solange wir pfad-anweisungen haben
    while (!path.empty()) {
      //zur neuen position drehen
      turnTo((LabDirection)path.pop());
      //Schritt machen
      forward();
      //und bis zum naechsten knoten bewegen
      gotoNextNode();
    } //while
  } //followPath

} //class Neo

//diese Klasse speichert die Richtung
class LabDirection {
  //Statusvariablen fuer die Richtungen
  public static final int LEFT_UP    =   1; 
  public static final int UP         =   2;
  public static final int RIGHT_UP   =   4;
  public static final int RIGHT_DOWN =   8;
  public static final int DOWN       =  16;
  public static final int LEFT_DOWN  =  32;
  
  //damit Iterieren wir durch die Richtungen:
  //die Drehung geht dabei rechts herum
  static final int[] DirArray = {UP, RIGHT_UP, RIGHT_DOWN, 
                                 DOWN, LEFT_DOWN, LEFT_UP};

  //aktuelle Richtung als Index im DirArray
  int Dir;
  
  //Konstruktoren
  LabDirection() {
    setDir(UP);
  }
  
  LabDirection(int dir) {
    setDir(dir);
  }
  
  LabDirection(LabDirection dir) {
    setDir(dir.getDir());
  }
  
  //bei vorgegebener Richtung die Position im Array suchen
  void setDir (int dir) {
    for (int i=0; i < DirArray.length; i++)
      if (DirArray[i] == dir) {
        this.Dir = i;
        return; //schluss, wenn wirs haben
      }
  }
  
  //Richtung zurueckliefern
  public int getDir() { return DirArray[Dir];}
  public int getIdx() { return Dir;}
  
  //drehen
  public void turn(int dir) {
    if (dir == Hamster.TURN_LEFT)
      Dir = ( (Dir==0) ? DirArray.length-1 : Dir-1 );
     else
      Dir = ( (Dir==DirArray.length-1) ? 0 : Dir+1 );
  } //void turn(int)

  //Differenz zwischen zwei Richtungen
  //liefert die Anzahl der Rechtsdrehungen zurueck, die noetig sind, um von
  //einer Richtung (this) zur naechsten (Dir) zu gelangen
  int Difference(LabDirection Dir) {
    int d = (Dir.getIdx() - getIdx()) % 6;
    return (d<0) ? (d+6) : d;
  }
  
  //Entgegengesetzte Richtung zurueckliefern
  int getOpposite() {
    return DirArray[(getIdx()+3) % 6];
  }

  //vergleichsmethode
  boolean equals(LabDirection d) {
    return getIdx() == d.getIdx();
  }

  //der vollstaendigkeit halber und fuer die logs
  public String toString() {
    String s = "";
    if (getDir() == UP)         s = "UP";
    if (getDir() == RIGHT_UP)   s = "RIGHT_UP";
    if (getDir() == RIGHT_DOWN) s = "RIGHT_DOWN";
    if (getDir() == DOWN)       s = "DOWN";
    if (getDir() == LEFT_DOWN)  s = "LEFT_DOWN";
    if (getDir() == LEFT_UP)    s = "LEFT_UP";
    return "("+new Integer(getIdx()).toString()+":"+
           new Integer(getDir()).toString()+") "+s;
  }    
} //class Direction

//diese Klasse speichert die Position auf der Hintergrundkarte
class LabPosition {
  //instanzuvariablen
  int PosX;
  int PosY;
  
  //Konstruktoren
  LabPosition() { PosX = 0; PosY = 0; }
  LabPosition(LabPosition Pos) { PosX = Pos.PosX; PosY = Pos.PosY; }
  LabPosition(int x, int y) { PosX = x; PosY = y; }
  
  //set-Methoden
  void setX(int x) {PosX = x;}
  void setY(int y) {PosY = y;}
  
  //get-Methoden
  int getX() {return PosX;}
  int getY() {return PosY;}
  
  //die beiden folgenden routinen errechnen die koordinatenveraenderung
  //fuer einen schritt in die angegebene richtung
  //  !! dabei ist die neue y-position von der alten x-position !!
  //  !! abhaengig                                              !!
  public int deltaX(LabDirection Dir) {
    if ( (Dir.getDir() == LabDirection.RIGHT_UP) || 
         (Dir.getDir() == LabDirection.RIGHT_DOWN) ) return +1;
    if ( (Dir.getDir() == LabDirection.LEFT_UP ) || 
         (Dir.getDir() == LabDirection.LEFT_DOWN) ) return -1;
    return 0;
  } //deltaX
  
  public int deltaY(LabDirection Dir) {
    //oben,unten sind unabhaengig von X
    if (Dir.getDir() == LabDirection.UP) return -1;
    if (Dir.getDir() == LabDirection.DOWN) return +1;
    
    //der rest bestimmt sich nach der X-position
    if (PosX%2 == 0) {
      if ( (Dir.getDir() == LabDirection.LEFT_UP) || 
           (Dir.getDir() == LabDirection.RIGHT_UP) ) return -1;
    } else {
      if ( (Dir.getDir() == LabDirection.RIGHT_DOWN) || 
           (Dir.getDir() == LabDirection.LEFT_DOWN) ) return +1;
    }
    return 0;
  } //deltaY 

  //Position einen Schritt in die angegebene Richtung setzen
  void move(LabDirection Dir) {
    //WICHTIG: ERST DIE Y-POSITION SETZEN, DA DIESE VON DER X-POS
    //         ABHAENGIG IST!!
    PosY += deltaY(Dir);
    PosX += deltaX(Dir);
  }   
  
  //ein LabPosition-Objekt erzeugen, das bereits einen Schritt in die
  //angegebene Richtung weitergesetzt ist
  LabPosition getNext(LabDirection Dir) {
    //neues Objekt anlegen
    LabPosition Pos = new LabPosition(this);
    //Schritt ausfuehren
    Pos.move(Dir);
    //und das objekt zurueckgeben
    return Pos;
  }

  //Vergleich zweier Positionen
  public boolean equals(LabPosition pos2) {
    return (getX() == pos2.getX()) && (getY() == pos2.getY());
  }

  //der vollstaendigkeit halber und fuer die logs
    public String toString() {
      return "("+new Integer(getX()).toString()+"/"+
                 new Integer(getY()).toString()+")";
    }
  
} //class LabPosition

//die Landkarte, die der Hamster sich merkt
class LabMap {
  //instanzvariablen
  int[][] Fields;
  int Width;
  int Height;

  //noch ein paar Konstanten fuer den Status eines Felds
  public static final int EXPLORED   =  64;
  public static final int BLOCKED    = 128;  

  //Konstruktor
  LabMap(int w, int h) {
    Width=w;
    Height=h;
    //Feld anlegen...
    Fields = new int[w][h];
    //und initialisieren: Feld ist nach allen Richtungen offen, nicht gesperrt
    //und nicht erkundet
    for (int i=0; i<w; i++)
      for (int j=0; j<h; j++) 
        Fields[i][j] = LabDirection.LEFT_UP+LabDirection.UP+
                       LabDirection.RIGHT_UP+LabDirection.RIGHT_DOWN+
                       LabDirection.DOWN+LabDirection.LEFT_DOWN;
  }
  
  //Feldmasze zurueckliefern
  int getWidth() {return Width;}
  int getHeight() {return Height;}
  
  //ein bestimmtes Feld zurueckliefern
  //wenn die Koordinaten ausserhalb der Karte liegen, wird BLOCKED 
  //zurueckgeliefert
  int getField(LabPosition Pos) {
    if ( (Pos.getX() < 0 || Pos.getY() > getWidth() ) || 
         (Pos.getX() < 0 || Pos.getY() > getHeight())    )
      return BLOCKED;
    return Fields[Pos.getX()][Pos.getY()];
  }

  //den Status einer bestimmten Eigenschaft zurueckliefern
  public boolean hasState(LabPosition Pos, int state) {
    return ((Fields[Pos.getX()][Pos.getY()] & state) == state);
  }
  
  //den Status einer bestimmten Eigenschaft setzen
  public void setState(LabPosition Pos, int state) {
    Fields[Pos.getX()][Pos.getY()] = Fields[Pos.getX()][Pos.getY()] | state;
  }
  
  //den Status einer bestimmten Eigenschaft loeschen
  public void clearState(LabPosition Pos, int state) {
    Fields[Pos.getX()][Pos.getY()] = Fields[Pos.getX()][Pos.getY()] & ~state;
  }
  
  //Feld in angegebener Richtung als Wand markieren
  public void markWall(LabPosition Pos, LabDirection Dir) {
    clearState(Pos, Dir.getDir());
    //auf dem angrenzenden Feld markieren wir die wand auch gleich noch
    //(waende sind in beiden richtungen undurchlaessig ;-) )
    clearState(Pos.getNext(Dir), Dir.getOpposite());
  }

  //ganzes Feld blockieren
  public void blockField(LabPosition Pos) { setState(Pos, BLOCKED); }
    
  //Feld als erkundet markieren
  public void explore(LabPosition Pos) { setState(Pos, EXPLORED); }

  //pruefen, ob Feld in gegebener Richtung offen ist (laut karte)
  public boolean canGo(LabPosition Pos, LabDirection Dir) {
    return ( hasState(Pos, Dir.getDir())
             && ( !isBlocked(Pos.getNext(Dir)) ) );
  }

  //Feld wurde erkundet
  public boolean isExplored(LabPosition Pos) {return hasState(Pos, EXPLORED);}
  
  //Feld ist blockiert (Sackgasse etc)
  public boolean isBlocked(LabPosition Pos) {return hasState(Pos, BLOCKED);}  
  
 //nach wie vielen Seiten ist das Feld offen?
  public int openDirs(LabPosition Pos) {
    int c=0; //zaehler;
    LabDirection dir = new LabDirection(LabDirection.UP); //iterator
    do {
      c += (canGo(Pos, dir)?1:0);
      dir.turn(Hamster.TURN_LEFT);
    } while (dir.getDir() != LabDirection.UP);
    return c;
  }

  //Feld ist Knotenpunkt: nach mehr als 2 Seiten offen
  public boolean isNode(LabPosition Pos) {return (openDirs(Pos) > 2);}
  
  //Feld ist Gasse: nur nach 2 Seiten offen
  public boolean isLane(LabPosition Pos) {return (openDirs(Pos) == 2);}
  
  //Feld ist Sackgasse
  //normalerweise wird das feld anschliessend gesperrt, wenn hier true 
  //rauskommt
  public boolean isEnd(LabPosition Pos) {return (openDirs(Pos) == 1);}

  //Field ist nicht erreichbar: Nach keiner Seite offen
  public boolean isUnreachable(LabPosition Pos) {return (openDirs(Pos) == 0);}  

} //class LabMap


//ab hier kommen Klassen fuer den Graph

//Knotenklasse
//  speichert X/Y-Position
//  ist in der Lage, einen hashwert zurueckzuliefern
//  und kann fuer Hashtable als schluessel verwendet werden
class GrNode {
 
  LabPosition Pos; 
    
  public GrNode() { this(0,0); }
  public GrNode(LabPosition pos) { this(pos.getX(), pos.getY()); }
  public GrNode(int x, int y) { Pos = new LabPosition(x, y); }
    
  //Hashfunktion
  public int hashCode() {
    //die 1000 muesste hier noch durch ne Konstante ersetzt werden
    //der wert muss groesser als der kantenlaenge der internen karte
    //sein!
    return Pos.getX()*1000+Pos.getY();
  }
  
  //equals-Funktion
  public boolean equals(Object o) {
    //Zwei Knoten stimmen ueberein, wenn ihre Hash-Codes uebereinstimmen
    return ((GrNode)o).hashCode() == hashCode();
  }
  
  //Ausgabefunktion
  public String toString() {
    return Pos.toString();
  }
  
} //class GrNode

//Kantenklasse
//  speichert Ziel einer Kante und Kosten des Weges
class GrEdge {

  GrNode dest; //Ziel der Kante
  int cost;  //"Kosten" fuer den Durchlauf dieser Kante
  
  //um wegplaene aufzustellen: die richtung, in die die kante zeigt
  LabDirection Dir;
    
  public GrEdge(GrNode d) { this(d, 0, null);}
  public GrEdge(GrNode d, int c, LabDirection dir) {
    dest=d; 
    cost=c; 
    Dir = new LabDirection(dir);
  }

  //Selektoren    
  public int getCost() { return cost; }
  public GrNode getDest() { return dest; }
  public LabDirection getDir() { return Dir; }
  
  //Modifikatoren
  public void setCost(int c) { cost = c; }
  public void setDir(LabDirection dir) { Dir = new LabDirection(dir); }
  
  //Ausgabefunktion
  public String toString() {
    return dest.toString()+":"+cost+":"+Dir.toString();
  }
  
} //class GrEdge


class Graph {
  
  //Hashtable fuer Abbildung Knoten->Index
  private Hashtable nodeHash;

  //Knotenliste, speichert fuer jeden Knoten eine LinkedList
  private Vector nodeList;  
  
  //da stehen knotenentfernungen drin, wenn wir einen weg gesucht haben
  int D[];
  
  //das maximale Gewicht fuer eine Kante - soll ein "unendlich" symbolisieren
  static final int MAX_WEIGHT=100000;

  //Konstruktoren
  public Graph() {
   nodeHash = new Hashtable();
   nodeList = new Vector();
   int D[] = new int[0];
  }

  //Index eines Knotens ermitteln
  //  rueckgabewert -1, wenn nicht vorhanden
  public int getNodeIdx(GrNode node) {
    if (! nodeHash.containsKey(node) ) return -1;
    return ( (Integer)nodeHash.get(node) ).intValue();
  }

  //Knoten zur Knotenliste hinzufuegen
  public void addNode(GrNode node) {
    //nur, wenn nicht schon vorhanden
    if ( getNodeIdx(node) != -1 ) return;
    
    //Knoten zur nodeList hinzufuegen
    nodeList.addElement( new LinkedList() );
    
    //Knoten zum nodeHash hinzufuegen
    nodeHash.put( node, new Integer(nodeList.size()-1) );
    
  }
  
  //addNode noch einmal fuer LabPosition
  public void addNode(LabPosition Pos) {
    addNode( new GrNode(Pos) );
  }

  //pruefen, ob eine angegebene Position ein Knoten ist
  public boolean isNode(LabPosition Pos) {
    return isNode(new GrNode(Pos));
  }
  
  public boolean isNode(GrNode node) {
    return (getNodeIdx(node) != -1);
  }

  //verbindung einfuegen
  //dabei ist src der startknoten, 
  //srcDir die richtung, in der man am startknoten losgehen muss, um zum
  //zielknoten zu gelangen,
  //dest der zielknoten, analog destDir,
  //cost sind die schritte, die fuer den weg noetig sind
  public void addConnection(GrNode src, LabDirection srcDir,
                            GrNode dest, LabDirection destDir,
                            int cost) {
    //keine reflexiven verbindungen
    if (src.equals(dest)) return;
    
    //wenn einer der knoten noch nicht existiert, wird er angelegt
    addNode(src);
    addNode(dest);
    
    //die beiden kanten hinzufuegen
    //hinrichtung
    addEdge(src, dest, cost, srcDir);
    //rueckrichtung
    addEdge(dest, src, cost, destDir);
  }

  //verbindung einfuegen fuer LabPosition
  public void addConnection(LabPosition src, LabDirection srcDir,
                            LabPosition dest, LabDirection destDir,
                            int cost) {
    addConnection(new GrNode(src), srcDir, new GrNode(dest), destDir, cost);
  }

  //eine Kante an einem Knoten in einer bestimmten richtung suchen
  GrEdge getEdge(GrNode src, LabDirection Dir, boolean remove) {
    if (!isNode(src)) return null;
    LinkedList ll = (LinkedList)nodeList.elementAt(getNodeIdx(src));
    Iterator ll_it = ll.iterator();
    while (ll_it.hasNext()) {
      GrEdge edge = (GrEdge)ll_it.next();
      //wenn die Richtung stimmt, sind wir fertig
      if (edge.getDir().equals(Dir)) {
        if (remove) ll_it.remove();      
        return edge;
      }
    } //while
    
    //wenn wir hier ankommen, gibt es die Kante nicht
    return null;
  }
  
  //eine Kante an einem Knoten mit bestimmten Ziel suchen
  GrEdge getEdge(GrNode src, GrNode dest, boolean remove) {
    //wenn einer der knoten nicht existiert, gibt es auch die verbindung nicht
    if ( (!isNode(src)) || (!isNode(dest)) ) return null;
    
    LinkedList ll = (LinkedList)nodeList.elementAt(getNodeIdx(src));
    Iterator ll_it = ll.iterator();
    while (ll_it.hasNext()) {
      GrEdge edge = (GrEdge)ll_it.next();
      //wenn das Ziel stimmt, sind wir fertig
      if (edge.getDest().equals(dest)) {
        if (remove) ll_it.remove();      
        return edge;
      }
    } //while

    //wenn wir hier ankommen, gibt es die Kante nicht
    return null;
  }
  
  //eine Kante zum Graphen hinzufuegen
  public void addEdge(GrNode src, GrNode dest, int cost, LabDirection dir) {
    //reflexive verbindungen sind nicht erlaubt
    if (src.equals(dest)) return;
    
    //nicht existente Knoten werden angelegt
    addNode(src);
    addNode(dest);    
    
    //gibt es die kante schon?
    GrEdge edge = getEdge(src, dest, false);
    if (edge != null) {
      //jetzt ist es interessant, ob wir eine kante mit weniger aufwand 
      //haben - die nehmen wir dann
      if (cost < edge.getCost()) {
        edge.setCost(cost);
        edge.setDir(dir);
      }
      //fertig
      return;
    }

    //sonst legen wir sie an
    LinkedList ll = (LinkedList)nodeList.elementAt(getNodeIdx(src));
    ll.add(new GrEdge(dest, cost, dir));
  } //addEdge

  //zur Kontrolle die toString-methode, die die Adjazenzliste ausgibt
  public String toString() {
    String res="";
    
    //durch die Hash-Keys iterieren
    Enumeration enum = nodeHash.keys();
    while (enum.hasMoreElements()) {
      GrNode key = (GrNode)enum.nextElement();
      res += key.toString();
      
      LinkedList ll = (LinkedList)nodeList.elementAt( 
          ( (Integer)nodeHash.get(key) ).intValue()
        );
      Iterator ll_it = ll.iterator();
      while (ll_it.hasNext())
        res += " - " + ((GrEdge)ll_it.next()).toString();
        
      res += "\n";
    } //while
    
    return res;
  } //String toString(void)

  //kuerzesten Weg suchen
  Stack findPath(LabPosition start, LabPosition dest) {
  
    //das werden die indizes von start/zielknoten
    //hier wird nur noch mit indizes im int-bereich gearbeitet
    int startIdx = getNodeIdx(new GrNode(start));
    int destIdx = getNodeIdx(new GrNode(dest));
    
    //die Anzahl der Knoten
    int nodeCount = nodeList.size();
    
    //entfernungen zu allen knoten
    D = new int[nodeCount];
    //wird mit "unendlich" initialisiert
    for (int i=0; i<nodeCount; i++) D[i] = MAX_WEIGHT;
    D[startIdx] = 0;
    
    //hier drin landet dann der kuerzeste weg
    int[] vorgaenger = new int[nodeCount];
    
    //noch zu besuchende Knoten
    HashSet B = new HashSet();
    //da kommt der startknoten rein
    B.add(new Integer(startIdx));

    //und dann gehts los
    while (!B.isEmpty()) {
      //jetzt waehlen wir das i aus B, mit dem D[i] minimal wird und
      //entfernen es
      Iterator B_it = B.iterator();
      int i = ((Integer)(B_it.next())).intValue();      
      while (B_it.hasNext()) {
        int a = ((Integer)(B_it.next())).intValue();
        if (D[a] < D[i]) i = a;
      }
      B.remove(new Integer(i));
      
      //jetzt schauen wir durch die nachbarn von i
      Iterator l_it = ((LinkedList)(nodeList.get(i))).iterator();
      while (l_it.hasNext()) {
        GrEdge edge = (GrEdge)(l_it.next());
        int j = getNodeIdx(edge.getDest());
        //wenn wir da noch nicht waren, wird er spaeter noch besucht
        if (D[j] == MAX_WEIGHT) B.add(new Integer(j));
        
        //"verkuerzen"
        if (D[i] + edge.getCost() < D[j]) {
          D[j] = D[i] + edge.getCost();
          vorgaenger[j] = i;
        }
      } //while (hasNext)
    } //while (isEmpty)
    //puh, geschafft...das war der algorithmus von Dijkstra (jedenfalls sollte
    //er das werden...)
    
    //jetzt machen wir aus dem ergebnis noch einen laufstack, der dann
    //einfach abgearbeitet werden muss
    //dabei bewegen wir uns durch das vorgaenger-array, bis wir am zielknoten
    //angelangt sind
    //der stack beinhaltet die richtungen, in die sich der hamster an
    //knotenpunkten drehen muss; die muessen wir hier noch ermitteln
    Stack wayPoints = new Stack();
    int n=destIdx;
    
    while (n != startIdx) {
      //zuerst die kante bestimmen, die zum vorgaengerknoten fuert
      Iterator it = ((LinkedList)nodeList.elementAt(vorgaenger[n])).iterator();
      while (it.hasNext()) {
        GrEdge edge = (GrEdge)it.next();
        //wenn die indizes uebereinstimmen
        if (getNodeIdx(edge.getDest()) == n) {
          //richtung auf den stack und naechstes element
          wayPoints.push(new LabDirection(edge.getDir()));
          break;
        }
      }
      //weiter mit naechstem vorgaenger
      n = vorgaenger[n];
    } // while
      
    //den Stack liefern wir als "wegbeschreibung" zurueck
    return wayPoints;
  } //findPath

  //entfernung vom letzten such-startknoten bis pos zurueckgeben
  //-1, wenn nicht vorhanden
  int getDist(LabPosition pos) {
    if (!isNode(pos)) return -1;
    
    int idx = getNodeIdx(new GrNode(pos));
    if (D.length < idx) return -1;
    
    return D[idx];
  }  
  
} //class Graph


//Eintrag in der todoList
class TodoItem {
  
  //Attribute
  LabPosition pos;
  int corn;
  boolean explored;
  
  //Konstruktoren
  TodoItem(LabPosition pos, int corn, boolean explored) {
    this.pos = new LabPosition(pos);
    this.corn = corn;
    this.explored = explored;
  }
  
  //ganz kurzer konstruktor ;-)
  TodoItem(LabPosition pos) {
    this(pos, 0, true);
  }
  
  //Selektoren
  LabPosition getPos() { return pos; }
  int getCorn() { return corn; }
  boolean isExplored() { return explored; }
  
  //Modifikatoren
  //die Position laesst sich nicht veraendern
  void setCorn(int corn) { this.corn = corn; }
  void setExplored(boolean explored) { this.explored = explored; }

  //Praedikate
  
  //ein Eintrag ist "leer", wenn er ein feld nicht mit koernern und nicht
  //als unerforscht registriert ist
  boolean isEmpty() { return (getCorn()==0) && (isExplored()); }
  
  //Ausgabe
  public String toString() {
    return "["+getPos()+":"+(isExplored()?"erforscht":"unbekannt")+"-"+getCorn()+"]";
  }

} //class TodoList


//Todo-List
//erweiterungen: behandelt eintraege mit gleicher position
class TodoList extends LinkedList {

  //element mit angegebener position entfernen und zurueckgeben
  //null, wenn nichts gefunden wurde
  TodoItem remove(LabPosition pos) {
    
    //Iteratorobjekt anlegen
    Iterator li = iterator();

    //die gesamte Liste durchlaufen
    while (li.hasNext()) {
      TodoItem item = (TodoItem)li.next();
    
      //wenn die Positionen uebereinstimmen
      if (item.getPos().equals(pos)) {
        //entfernen 
        li.remove();
        //und zurueckliefern
        return item;
      }
    }
    
    //wenn wir hier ankommen, gibt es das gesuchte objekt nicht
    return null;
  
  } //TodoItem remove(LabPosition)  
  
  //add-Methoden
  
  //eintrag ueber die korn-anzahl hinzufuegen
  boolean add(LabPosition pos, int corn) {
    
    //erstmal gucken, obs fuer das feld schon einen eintrag gibt
    TodoItem item = remove(pos);
    
    //wenn nix gefunden wurde, wird ein neuer eintrag angelegt
    //dann sind wir auch fertig
    if (item == null)
      item = new TodoItem(pos);
      
    //dann wird die kornanzahl gesetzt
    item.setCorn(corn);
    
    //wenn der eintrag jetzt nicht "leer" ist, gehts anschliessend
    //zurueck in die liste
    if (!item.isEmpty())
      return add(item);
      
    //wenn wir hier ankommen, ist auf jeden fall alles gut gegangen
    return true;
  }
  
  //jetzt das ganze fuer einen "forschungseintrag"
  boolean add(LabPosition pos, boolean explored) {
    TodoItem item = remove(pos);
    if (item == null)
      item = new TodoItem(pos);
    
    item.setExplored(explored);
    
    //wenn der eintrag jetzt nicht "leer" ist, gehts anschliessend
    //zurueck in die liste
    if (!item.isEmpty())
      return add(item);

    return true;
  }

} //class TodoList


//Ende des Programms...Fehler, die ich bis hier hin nicht gemacht habe, werden
//auch nicht mehr auftreten ;-)