
package nexus;

import algds.c2002.Hamster;
import algds.c2002.Constants; 

import java.util.Stack;

/**
 * 
 *
 * @author <a href="mailto:jensL81@gmx.de">Jens Lincke</a>
 */
class FieldMap implements Constants{

    /**
     * 
     */
    private int MAP_LENGHT;
    /**
     * Karte aus einem festen 2-Dimenstionalen Array von Feldern
     */
    private Field[][] map; 

    /**
     * Creates a new <code>FieldMap</code> instance.
     */
    FieldMap(){
	MAP_LENGHT=(2*MAZELENGTH_MAX)+1;
	map = new Field[MAP_LENGHT][MAP_LENGHT];
	
	for(int i=0; i<MAP_LENGHT; i++)
	    for(int j=0; j<MAP_LENGHT; j++)
		map[i][j]=null;
    }

    /** 
     *
     *  Position(0,0) ist das Homefeld, 
     *  im Array liegt es aber in der Mitte, 
     *  deshalb wird das Koordinatensystem transformiert
     *
     *  @param p <code>Position</code> des Feldes
     *  @return a <code>Field</code> value
     */
    public Field getField(Position p){
    	return map[p.x+MAZELENGTH_MAX][p.y+MAZELENGTH_MAX];
    }

    /**
     * Describe <code>setField</code> method here.
     *
     * @param field a <code>Field</code> value
     * @param p a <code>Position</code> value
     */
    public void setField(Field field, Position p){
    	map[p.x+MAZELENGTH_MAX][p.y+MAZELENGTH_MAX]=field;
    }
    
        
    /**
     * Verbindungen zu direkten Nachbarn  
     *
     * @param p  einer <code>Position</code> 
     * @return a <code>Stack</code> von Richtungen in denen
     *           es Nachbarn gibt
     */
    public Stack getConnections(Position p){
	Stack s = new Stack();
	Field f = getField(p);
	int op_dir;
	
	if(f == null){ 
	    return s;
	}else if(f.looks == null){
	    //kann man aus den Nachbarfeldern Durchgnge erkennen? 
	    for(int dir=0; dir < Position.DIRECTIONS; dir++){
		f = getField(p.relative(dir));
		if((f != null ) && 
		   f.isReachable(Position.opposite_dir(dir))){
		    s.push(new Integer(dir));
		}
	    } 
	}else{ 
	    //schaue ob du von diesem Feld Durchgnge erkennst
	    for(int dir=0; dir < Position.DIRECTIONS; dir++){
		if(f.isPassable(dir)){
		    s.push(new Integer(dir));
		}
	    }
	}
    	return s;
    }
    

    /**
     *
     * @param p  die fragliche Position
     * @return true, wenn es noch unangesehene Nachbarn hat
     */
    public boolean hasUnknownNeighbors(Position p){
	Stack neighbors = getNeighbors(p);
	Position p_i;
	while(!neighbors.empty()){
	    p_i = (Position) neighbors.pop();

	    if(p_i.equals(1,-3))
	    	if(Nexus.DEBUG)System.out.println("!!" + getField(p_i));

	    if(getField(p_i) == null) return true;
	}
	return false;
    }

    /**
     * 
     * @param p a <code>Position</code> value
     * @return ein Stapel von Nachbarpositionen, zu denen eine
     * Verbindung besteht.
     */
    public Stack getNeighbors(Position p){
	Stack neighbors = new Stack();
	for(int dir=0; dir < Position.DIRECTIONS; dir++){
	    neighbors.push(p.relative(dir));
	}	
	return neighbors;
    }
    

}





