


package nexus;

import java.util.Stack;

/**
 *  Baumstruktur, in der man nur von Kindern zu Eltern gehen kann
 *
 *
 */
class Path{

    Path parent;  
    Position position;
    int move; // Bewegungsrichtung von parent zu this 0..5 , 
    boolean oneway;


    Path(){
	parent = null;
	move = 0;
	position = null;
	oneway = false;
    }
    
    Path(Path parent, int move){
	this.parent = parent;
	this.move = move;
	position = null;
	oneway = true;
    }

    Path(Path parent, int move, Position position){
	this.parent = parent;
	this.move = move;
	this.position = position;
	oneway = true;
    }

    Path(Path parent, int move, Position position, FieldMap map){
	this.parent = parent;
	this.move = move;
	this.position = position;
	if(parent == null)
	    oneway = false;
	else oneway = (parent.oneway || 
		       !map.getField(parent.position).isPassableViceVera(move));
    }

    public boolean isCurrentPosition(){
	return (parent == null);
    }
    
    public Stack getMoves(){
	return getMoves(new Stack());
    }

    public Stack getMoves(Stack s){
	if(parent == null){
	    return s;
	}else {
	    s.push(new Integer(move));
	    return parent.getMoves(s);
	} 
    }


    public int length(){
	if(parent == null)
	    return 0;
	else return( 1 + parent.length()); 
    }

    public String toString(){
	if(parent == null){
	    return("path: " + position);
	}else return(parent.toString() +"-" +  move + "->" + position  ); 
    }

    public boolean isOneWay(FieldMap map){
	if(parent == null){
	    return false;
	}else
	    return(!map.getField(parent.position).isPassableViceVera(move)    
		   || parent.isOneWay(map));
    }


    /**
     *  @param path Hinweg
     *  @return krzester Rckweg
     */
    public Path getPathRetour(Nexus hamster){
	PathBetween back = new PathBetween(hamster, 
					   position, hamster.position);
	return back.findPath();
    }
    
    /**
     *  @param path Hinweg
     *  @return krzester Rckweg
     */
    public boolean hasPathRetour(Nexus hamster){
	if(!oneway)
	    return true;
	else{
	    return getPathRetour(hamster)!=null; 
	}
    }

    public boolean staysOneLevel(Nexus hamster){
	return staysOneLevel(hamster, this);
    }
    
    static public boolean staysOneLevel(Nexus hamster, Path p){
	if(p.parent==null)
	    return true;
	Field f = hamster.map.getField(p.parent.position); 
	//if(!f.isPassableViceVera(p.parent.move))
	//    System.out.println("notpassable:" +  p.parent.position 
	//                     + " in " + p.move);
	return f.isPassableViceVera(p.move) 
	    && staysOneLevel(hamster, p.parent);
    }



}

