
package nexus;


/**
 *  <h4>Relative Postionen im Uhrzeigersinn</h4>
 *
 * <pre>
 *          __
 *       __/0 \__
 *      /5 \__/1 \
 *      \__/p \__/
 *      /4 \__/2 \
 *      \__/3 \__/
 *         \__/
 *
 * </pre>
 */


class Position{



    /**
     *  <code>D_NORTH</code>  = 0
     *
     */
    static final int D_NORTH=0;

    /**
     *  <code>D_NORTHEAST</code> = 1
     *
     */
    static final int D_NORTHEAST=1;

    /**
     *  <code>D_SOUTHEAST</code> = 2
     *
     */
    static final int D_SOUTHEAST=2;

    /**
     *  <code>D_SOUTH</code>  = 3
     *
     */
    static final int D_SOUTH=3;
    /**
     * <code>D_SOUTHWEST</code> = 4
     *
     */
    static final int D_SOUTHWEST=4;
    /**
     * <code>D_NORTHWEST</code> = 5
     *
     */
    static final int D_NORTHWEST=5;

    /**
     * <code>DIRECTIONS</code> = 6
     *
     */
    static final int DIRECTIONS=6;

    /**
     *  Abszisse
     */
    int x;

    /**
     * Ordinate
     */
    int y;

    /**
     * Creates a new <code>Position</code> instance.
     *
     * @param x an <code>int</code> value
     * @param y an <code>int</code> value
     */
    Position(int x, int y){
	this.x = x;
	this.y = y;
    }

    /**
     * Describe <code>relative</code> method here.
     *
     * @param direction an <code>int</code> value
     * @return a <code>Position</code> value
     */
    Position relative(int direction){
	// auf Vertreter von zwischen 0 und 5 beschrnken
	direction = dir_class(direction);
	switch(direction){
	case D_NORTH:
	    return new Position(x, y + 1);
	case D_NORTHEAST: //gerade vs. ungerade
	    return new Position(x+1, y + (x % 2 == 0 ?  0 : 1));
	case D_SOUTHEAST:
	    return new Position(x+1, y + (x % 2 == 0 ? -1 : 0));
	case D_SOUTH:
	    return new Position(x, y - 1);
	case D_SOUTHWEST:
	    return new Position(x-1, y + (x % 2 == 0 ?  -1 : 0));
	case D_NORTHWEST:
	    return new Position(x-1, y + (x % 2 == 0 ? 0 : 1));
	}
	return null;
    }


    /**
     * Describe <code>dir_class</code> method here.
     *
     * @param n an <code>int</code> value
     * @return quivalenzklassenvertreter der Richtung (0..5)
     */
    public static int dir_class(int n){
	int k = n % DIRECTIONS;
	return (k < 0 ? k + DIRECTIONS : k);
    }


    /**
     * Describe <code>opposite_dir</code> method here.
     *
     * @param dir an <code>int</code> value
     * @return entgegengesetzte Richtung
     */
    public static int opposite_dir(int dir){
	return dir_class(dir + 3);
    }

    /**
     * Vergleicht die Positionen nach den Koordinaten
     * 
     * @param p2 a <code>Position</code> value
     * @return a <code>boolean</code> value
     */
    public boolean equals(Position p2){
	return ((x == p2.x) && (y == p2.y));
    }

    /**
     * Vergleicht die Positionen nach den Koordinaten
     *
     * @param x2 an <code>int</code> value
     * @param y2 an <code>int</code> value
     * @return a <code>boolean</code> value
     */
    public boolean equals(int x2, int y2){
	return ((x == x2) && (y == y2));
    }

    /**
     * Describe <code>isHome</code> method here.
     *
     * @return true, wenn das Feld das Heimatfeld ist
     */
    public boolean isHome(){
	return equals(0,0);
    }

    /**
     * Describe <code>toString</code> method here.
     *
     * @return a <code>String</code> value
     */
    public String toString(){
	return "(" + x +  " " + y + ")";
    }



    /**
     * Describe <code>logAllYouKnow</code> method here.
     *
     * @param hamster a <code>Nexus</code> value
     */
    public void logAllYouKnow(Nexus hamster){
	hamster.logln("!" + this + " "  +
		      hamster.map.getField(this) + " " +
		      hamster.map.getConnections(this) +
		      " hasUnknownNeighbors:" +
		      hamster.map.hasUnknownNeighbors(this)
		      );
    }





}



