package nexus;
import java.util.Stack;


/**
 * Stellt ein den Breadth-First Search Algorithmus zur Verfgung
 *
 * @author <a href="mailto:jensL81@gmx.de">Jens Lincke</a>
 */
abstract class ShortestPath{

    Nexus hamster;
    BinMap discovery;
    Position start;
    boolean oneway;

    /**
     * Creates a new <code>ShortestPath</code> instance.
     *
     * @param hamster ein Hamsteruniversum
     */
    ShortestPath(Nexus hamster){
	this.hamster = hamster ;
	 discovery = new BinMap();
	 start = hamster.position;
    }

    /**
     *  Creates a new <code>ShortestPath</code> instance.
     *
     * @param hamster Hamsteruniversum
     * @param start alternative Startpostion
     *
     */
    ShortestPath(Nexus hamster, Position start){
	this.hamster = hamster ;
	discovery = new BinMap();
	oneway = false;
	this.start = start;
    }

    
    /**
     * Describe <code>getNextUnexploredPaths</code> method here.
     *
     * @param path a <code>Path</code> value
     * @return a <code>Stack</code> value
     */
    Stack getNextUnexploredPaths(Path path){
	Stack directions = hamster.map.getConnections(path.position);
	Stack paths = new Stack();
	Position p;
	int i;
	while(!directions.empty()){
	    i = ((Integer) directions.pop()).intValue();
	    p = path.position.relative(i);;

	    if(!discovery.isExplored(p)){
		paths.push(new Path(path,i,p));
		discovery.setExplored(p);
	    }
		
	}
	return paths;
    } 


    /**
     * Bei welchen Feld soll der Pfad enden?
     *
     * @param path a <code>Path</code> value
     * @return true, wenn der Pfad zu diesem Feld fhren soll
     */
    protected abstract boolean compare(Path path);
    
    /**
     * Shortest-Path Suche auf Basis der 
     * "Breadth-First Search" (vgl. Goodrich & Tamassia)
     * 
     * Da die einzelnen verbindungen des Graphen nicht gewichtet sind,
     * ist dieser Algorithmus ausreichend. 
     * 
     *
     * @return a <code>Path</code> value
     */
    public Path findPath(){
	
	Stack level = new Stack();
	
	level.push(new Path(null, 0, start)); //Anfangswerte
	discovery.setExplored(start);

	Stack nextLevel;
	
	Path path;

	while(!level.empty()){
	    nextLevel = new Stack();
	    while(!level.empty()){
		path = (Path) level.pop();
		
		// Hook, bei dem sich die Erben "einhaken"
		if(compare(path)){
		    return path;
		}
		nextLevel.addAll(getNextUnexploredPaths(path));
	    }
	    level = nextLevel;
	}
	// der Graph wurde traversiert ohne etwas zu finden
	return null;
    } 

}

