package recursiverunner;

import algds.c2002.Hamster;
import java.util.Stack;
import java.util.Vector;

public class RecursiveRunner extends Hamster
{
	boolean destroy = false;
	int depth = -1;	// Tiefe der Rekursion
	int direction = 0;	// Kompass
	Position[][] mazeFields = new Position[62][62];	// Reprsentation des Labyrinths
	Stack wayBack = new Stack();
	Position counter = new Position(0,0);	// globale Koordinatenanzeige


	// Klasse zur Darstellung eines einzelnen Feldes
	
	public class Position
	{
		int x = 0;	// Koordinaten
		int y = 0;
		int corn = 0;	// Anzahl der Krner
		int parentDirection;	// Richtung des vorher besuchten Feldes
		int homeDirection;	// Richtung des krzesten Heimweges
		int homeDistance;	// Lnge des krzesten Heimweges
		boolean visited = false;
		// freie und erreichbare Nachbarfelder
		boolean[] roomFree = new boolean[]{true,true,true,true,true,true};
		boolean[] accessible = new boolean[]{true,true,true,true,true,true};

		// Konstruktor
		public Position(int x, int y)
		{
			this.x = x;
			this.y = y;
		}

		// Berechnen der Koordinaten eines angrenzenden Feldes
		Position next(int dir)
		{
			Position newPos = new Position(this.x,this.y);
			int i = (dir<=2 ? +1 : -1);
			if(dir>=3) dir-=3;
			if(dir<=1) newPos.x += i;
			if(dir>=1) newPos.y += i;
			return newPos;
		}
		
		// Setzen und Abfragen der Variablen
		void setWall(int dir)
		{
			Position neighbour = getMazeField(this.next(dir));
			roomFree[dir] = accessible[dir] = false;
			neighbour.roomFree[(dir+3)%6] = neighbour.accessible[(dir+3)%6] = false;
		}
		
		void setVisited()
		{
			visited = true;
			for(int i=0; i<=5; i++)
			{
				Position neighbour = getMazeField(this.next(i));
				neighbour.setRoomFull((i+3)%6);
			}
		}
		
		void resetVisited()
		{
			visited = false;
			for(int i=0; i<=5; i++)
			{
				Position neighbour = getMazeField(this.next(i));
				if(neighbour.accessible[(i+3)%6]) neighbour.roomFree[(i+3)%6] = true;
			}
		}
		
		void setRoomFull(int dir)
		{
			roomFree[dir] = false;
		}

		boolean roomLeft()
		{
			return (roomFree[0] || roomFree[1] || roomFree[2] || roomFree[3] || roomFree[4] || roomFree[5]);
		}

		void setHomeWay(int dir, int length)
		{
			if(length<this.homeDistance-1)
			{
				this.homeDistance = length+1;
				this.homeDirection = dir;
				for(int i=0; i<=5; i++)
				if(accessible[i])
				{
					Position neighbour = getMazeField(this.next(i));
					neighbour.setHomeWay((i+3)%6,this.homeDistance);
				}
			}
		}

		// nach Wnden umschauen
		void lookAround()
		{
			for(int i=0; i<=5; i++)
			{
				turn(TURN_LEFT);
				if(((lookForward() & WALLMASK) == WALL)) setWall(direction);
			}
		}

		// Nachbarfelder berprfen
		void checkNeighbours()
		{
			for(int i=0; i<=5; i++)
			{
				Position neighbour = getMazeField(this.next(i));
				if(accessible[i] && neighbour.visited)
				{
					setRoomFull(i);
					setHomeWay(i,neighbour.homeDistance);
				}
			}
		}

		// Suchen des nchsten freien Feldes
		int nextRoom()
		{
			int prefered = getPriority(this);
			if(prefered!=-1) return prefered;
			if(x>0 && y>=x) prefered = 4;
			if(y>=0 && x>y) prefered = 3;
			if(x>=0 && y<0) prefered = 2;
			if(x<=0 && y>0) prefered = 5;
			if(x<0 && y<=x) prefered = 1;
			if(y<=0 && x<y) prefered = 0;
			int[] row = new int[]{0,+1,+2,-1,-2,3};
			for(int i=0; i<=5; i++)
			{
				int dir = (6+prefered+row[i])%6;
				if(roomFree[dir]) return dir;
			}
			return -1;
		}

		public boolean equals(Object o)
		{
			Position p = (Position)o;
			return (this.x==p.x && this.y==p.y);
		}
	
		public String toString()
		{
			return "Position: "+x+":"+y;
		}
	}

	// Umrechnen der Labyrinth- in Array-Koordinaten
	public Position getMazeField(Position pos)
	{
		return mazeFields[pos.x+30][pos.y+30];
	}

	// Suchen eines angrenzenden Feldes, dass eingeschlossen ist
	public int getPriority(Position position)
	{
		for(int i=0; i<=5; i++)
		{
			if(position.roomFree[i])
			{
				Position neighbour = getMazeField(position.next(i));
				if(!neighbour.roomLeft()) return i;
			}
		}
		return -1;
	}

	// Heimkehren und Abladen der Krner	
	void homeRun()
	{
		while(!atHome())
		{
			Position backPos = getMazeField(counter);
			turnTo(backPos.homeDirection);
			forward();
			wayBack.push(new Integer((backPos.homeDirection+3)%6));
		}
		dropHere(getCarriedCorn());
		if(!destroy) backRun();
	}

	// Zurckgehen des Heimweges, um weiter Korn einzusammeln
	void backRun()
	{
		while(!wayBack.empty())
		{
			int backDir = ((Integer)wayBack.pop()).intValue();
			turnTo(backDir);
			forward();
		}
	}

	// berschriebene Hamster-Methoden	
	public boolean forward()
	{
		counter = counter.next(direction);
		return super.forward();
	}

	public void turn(int dir)
	{
		super.turn(dir);
		if(dir == TURN_LEFT)
			direction = (direction+5)%6;
		if(dir==TURN_RIGHT)
			direction = (direction+1)%6;
	}
	
	// Drehen in eine bestimmte Richtung
	void turnTo(int dest)
	{
		while(dest!=direction)
		if(dest<direction)
			if(direction-dest<=3) turn(TURN_LEFT);
			else turn(TURN_RIGHT);
		else
			if(direction-dest<=-3) turn(TURN_LEFT);
			else turn(TURN_RIGHT);
	}

	// Besuchen eines Feldes
	boolean visit(Position parent)
	{
		depth++;
		Position pos = getMazeField(counter);
		pos.setVisited();
		pos.parentDirection = pos.homeDirection = (direction+3)%6;
		pos.homeDistance = parent.homeDistance+1;
		pos.corn = getAvailableCorn();
		pos.lookAround();
		pos.checkNeighbours();

		// angrenzende Felder werden rekursiv besucht
		while(pos.roomLeft() && !destroy)
		{
			int nextRoom = pos.nextRoom();
			turnTo(nextRoom);
			int back = (direction+3)%6;
			forward();
			boolean result = visit(pos);
			if(result==false) pos.setRoomFull(nextRoom);
			pos.checkNeighbours();
			if(!destroy)
			{
				turnTo(back);
				forward();
				dropHere(getCarriedCorn());
				pos.corn = getAvailableCorn();
			}
			if(pos.roomLeft() && wayBack.size()<depth)
			{
				destroy = false;
				backRun();
			}
			if(pos.equals(new Position(0,0)))
			{
				destroy = false;
				wayBack.clear();
			}
		}

		// Aufsammeln von Korn
		if(!atHome())
		{
			take(CORNCARRY_MAX);
			pos.corn = getAvailableCorn();
			if(getCarriedCorn()==CORNCARRY_MAX)
			{
				destroy = (pos.corn==0 && !pos.roomLeft());
				homeRun();
			}
		}

		if(pos.roomLeft() || pos.corn!=0) pos.resetVisited();
		if(destroy)
		{
			int backDir = ((Integer)wayBack.firstElement()).intValue();
			if(backDir==((pos.parentDirection+3)%6)) wayBack.remove(0);
			else wayBack.add(0,new Integer(pos.parentDirection));
		}
		depth--;
		return (pos.roomLeft() || pos.corn!=0);
	}

	// Initialisieren des Labyrinthes
	void initField()
	{
		for(int i=0; i<=61; i++)
			for(int j=0; j<=61; j++)
				mazeFields[i][j] = new Position(i-30,j-30);
	}

	public void run()
	{
		initField();
		visit(new Position(0,0));
		dropHere(getCarriedCorn());
	}
	
	public String getHamsterAuthor()
	{
		return "Roland Pfisterer & Bjoern Boehm";
	}
	
	public String getHamsterDate()
	{
		return "30.04.02";
	}
	
	public String getHamsterVersion()
	{
		return "42.6";
	}

}