package scrat;

import java.util.*;


/**
 * 
 * @author Sebastian Stober, Florian Heidenreich
 * @since Apr. 2002
 */
public class Scrat extends simpleScrat {
  
  protected int strategy = 4;       // Auswahl der Erkundungsstrategie
  
  public void run() {

    initScrat(); // muss der erste Befehl der run()-Methode sein
    
    setDebugText(verbose()); // Texte anschalten, falls verbose()
    setDebugBreak(false);
    
    if (!configFlags[0]) {// erweiterte Strategie-Parameter setzen:
      /**
       * configFlags[0] ist initial false, dh. die Einstellungen werden im folgenden manuell
       * vorgenommen
       *
       * configFlags[0] muss auf true gesetzt werden, wenn:
       * ein Hamster hoeherer Stufe von Scrat erbt. diese muss nur die configFlags
       * und strategy setzten (ohne dabei Methoden aufrufen zu muessen) und kann damit
       * das Verhalten des Hamsters beeinflussen
       * in der initScrat()-Methode werden dann die Werte automatisch gesetzt 
       *
       * das Erben sollte so erfolgen:
       * public class scrat411111111 extends Scrat.Scrat {
       *   configFlags[0] = true; // wichtig!
       *   configFlags[1] = ... 
       *   configFlags[8] = ...
       *   strategy = 0..4
       * }
       * die Ziffern hinter dem classname heissen: 4 -> strategy = 4, jede 1 steht fuer true
       * damit erkennt man sofort, mit welchen Parametern gestartet wurde und man braucht die
       * Parameter nicht immer in Scrat.Scrat zu setzen
       */

      useAdvancedWayHome(true);
      /**
       * wirkt immer
       *
       * bei Heimweg-Suche wird dropForward() genutzt, es werden dropNodes angelegt als Zwischenlager
       */
      
      setPreferCliffUp(true);    
      /**
       * wirkt nur, wenn useAdvancedWayHome(true)
       *
       * bei Heimweg-Suche werden CliffUps beim Backtracing vorgezogen, damit dropForward()
       * oefter genutzt werden kann
       */

      setPreferCliffUpHome(true); 
      /**
       * wirkt nur, wenn useAdvancedWayHome(true)
       *             und setPreferCliffUp(false) 
       */

      setFindAlternativeDropNodes(true, 3, 3);
      /**
       * wirkt nur, wenn useAdvancedWayHome(true)
       *
       * wenn useAdvancedWayHome(true) kann es vorkommen, das bei einer dropNode das Corn-Maximum
       * erreicht ist
       * in diesem Fall kann nach Alternativer gesucht werden (Nachbarfelder)
       * die Suchtiefe wird ueber den 2. Parameter eingestellt
       * der 3. Parameter gibt den erlaubten Mehraufwand an Schritten an
       */

      setPreferHeight(true);
      /**
       * wirkt immer
       *
       * bei der Breitensuche werden zuerst Knoten mit hohen Levels erweitert
       * das fuehrt dazu, dass der Hamster zunaechst versucht, auf "Gebirgen" zu laufen
       */

      setUseWayHomeHeuristic(true);
      /**
       * wirkt immer
       * sollte kein Weg nach hause bekannt sein, versucht der Hamster den direkten Weg
       */

      setRememberDropNodesAtHome(true);
      /**
       * kann die gesamte Strategie veraendern!
       */

      setGotoCornNeighbour(false);
      /**
       * Hamster geht immer zum Corn, wenn etwas auf einem Nachbarfeld liegt
       */ 

      setGotoCornNeighbourHeightOK(true);
      /**
       * wirkt wenn setGotoCornNeighbour(false)
       * Hamster geht nur dann zum Corn, wenn etwas auf einem Nachbarfeld liegt und das Nachbarfeld
       * midestens die gleiche Hoehe hat
       */

      setGotoCornNeighbourBackOK(false);
      /**
       * wirkt wenn setGotoCornNeighbour(false)
       * Hamster geht nur dann zum Corn, wenn etwas auf einem Nachbarfeld liegt und man vom Nachbarfeld
       * wieder zurueckgehen kann
       */
                               
      setDistanceControl(true);
      /**
       * aktiviert von die Kontrolle, ob sich der Weg noch lohnt
       * diese Erweiterung ist sehr rechenintensiv und daher ist es guenstiger, sie zu einem spaeteren 
       * Zeitpunkt einzuschalten (z.B. wenn alles entdeckt wurde)
       */
    }
  
    
    if ((strategy == 0) || debug) { // debug - modus ?
      setKeyControls(true);
    }
    else {
      setKeyControls(false);
    } 

    // eine volle Drehung zum Orientieren
    for (int i=0; i<7; i++) {
      turnPos();
    }
    
    switch (strategy){
    case 1: runParanoidMode();break;
    case 2: runGreedyMode();break;
    case 3: runExplorer();break;
    case 4: runAdvancedParanoidMode();break;
    default: { //manuelle steuerung
      while (!finished) {
        try {      
          Thread.sleep(200);
        } 
        catch (InterruptedException e) {
        }
      } 
    }
    }
    setKeyControls(false);
    
    
  }
  
  
  /**
   * steuert den Erkundungslauf des Hamster:
   * -sammelt solange Koerner, bis er keine mehr findet
   * -wenn keine Koerner mehr da sind versucht er weiter zu erkunden 
   *  bis wieder Koerner gefunden wurden
   * -wenn's nix mehr zu erkunden gibt und keine Koerner mehr rumliegen is Ende
   */
  private void runGreedyMode() {
    //boolean canGoThere;
    while (somethingToExplore() || checkCornAvailable()) {
      //canGoThere = false;
      if (checkCornAvailable()) {
        //myMap.printCornNodes();
        while (!gotoNextCorn() && somethingToExplore()){
          // cant go there !!!
          gotoNextUnexplored(); // find way
        }
        if (takeAll()) { // voll
          goHome();
          dropAll();
        }
      }
      else { // kein Corn da
        gotoNextUnexplored();
      }                     
    }
    System.out.println("explore = "+somethingToExplore()+"  corn = "+checkCornAvailable());
    goHome();
    dropAll();
    finished = true; 
  } 

  /**
   * steuert den Erkundungslauf des Hamster:
   * -sammelt solange Koerner, bis er keine mehr findet
   * -wenn keine Koerner mehr da sind versucht er weiter zu erkunden 
   *  bis wieder Koerner gefunden wurden
   * -wenn's nix mehr zu erkunden gibt und keine Koerner mehr rumliegen is Ende
   */
  private void runParanoidMode() {

    while (somethingToExplore() || checkCornAvailable()) {
      //System.out.println("explore = "+somethingToExplore()+"  corn = "+checkCornAvailable());
      if (!gotoCornNeighbour() ) { // kein Nachbar hat corn
        gotoNextWP(); // also suche den naechsten WP
      }
      if (debug) {
        while (wait) {
          try {      
            Thread.sleep(200);
          } 
          catch (InterruptedException e) {
          }
        }
        wait = true;
      }
      if (takeAll()) { // nehme alles auf, was geht
        logln("going home");
        goHome();      // und gehe nach Hause falls voll
        logln("drop corn");
        dropAll();
      }
    }
    //System.out.println("explore = "+somethingToExplore()+"  corn = "+checkCornAvailable());
    logln("going home, last time");
    goHome();
    logln("drop corn");
    dropAll();
    finished = true; 
  } 

  private void runExplorer(){
    while (gotoNextUnexplored()) {
      if (takeAll()) { // voll
        logln("going home");
        goHome();
        logln("drop corn");
        dropAll();
      }
    }
    logln("everything explored");
    // alles entdeckt
    while (checkCornAvailable()) {
      gotoNextCorn();
      if (takeAll()) { // voll
        logln("going home, no unexplored");
        goHome();
        logln("drop all, no unexplored");
        dropAll();
      }
    }
    logln("going home, last time");
    goHome();
    dropAll();
    finished = true; 
  }

  private void runAdvancedParanoidMode(){
    boolean somethingToDo = true;
    while (somethingToDo && (somethingToExplore() || checkCornAvailable())) {
      //logln("explore = "+somethingToExplore()+"  corn = "+checkCornAvailable());
      if (!gotoCornNeighbour() ) { // kein Nachbar hat corn
        somethingToDo = gotoNextWP(); // also suche den naechsten WP
        //debugWait();
      }
      //debugWait();
      if (somethingToDo) {
        if (takeAll()) { // nehme alles auf, was geht
          if (transportWayHome()) {
            logln("at home");
            dropAll();
            rememberDropNodes();
            logln("drop corn");
          }         
        }
      } 
    }
    /* collect corn */
    logln("-------------------- ALLES ENTDECKT !!! ----------------------");
    debugWait();
    setRememberDropNodesAtHome(true);
    setDistanceControl(true);
    while (rememberDropNodes()){
      logln("rememeberDropNodes()");
      while (gotoNextCorn()) {
        logln("gotoNextCorn()");
        if (takeAll()) { // nehme alles auf, was geht
          if (transportWayHome()) {
            logln("at home");
            dropAll();
            logln("drop corn");
          }         
        }
      }
    }
    //logln("explore = "+somethingToExplore()+"  corn = "+checkCornAvailable());
    logln("going home, last time");
    goHome();
    logln("drop corn");
    dropAll();
    finished = true; 
  } 






  /* Autor(en) des Hamsters. */
  public String getHamsterAuthor() { return "Sebastian Stober, Florian Heidenreich"; }

  /* Datum der Hamsterversion. */
  public String getHamsterDate() { return "28.04.2002"; }

  /* Hamsterversion. */
  public String getHamsterVersion() { return "1.4 p11010_1"; }

}// scrat