package scrat;

/**
 * 
 * @author Sebastian Stober
 * @since Apr. 2002
 */
public class cell{ //implements algds.c2002.Constants{

  //Hamsterkonstanten: mussten nochmal extra implementiert werden wegen Problem mit switch-parseInt()
  public final static int WALL      = 0x02; //Integer.parseInt("0010",2);
  public final static int CLIFFDOWN = 0x04; //Integer.parseInt("0100",2);
  public final static int CLIFFUP   = 0x06; //Integer.parseInt("0110",2);
  public final static int SLOPEDOWN = 0x08; //Integer.parseInt("1000",2);
  public final static int SLOPEUP   = 0x0a; //Integer.parseInt("1010",2);
  
  // *************************** Konstanten ************************* //
  static final int  UNKNOWN = -2; // Nachbar unbekannt
  
  static final int
    
    BDR_UNKNOWN   = 0x00, // 000,  
    BDR_CLIFFDOWN = 0x02, // 010,
    BDR_SLOPEDOWN = 0x01, // 001,
    BDR_FLAT      = 0x03, // 011,
    BDR_SLOPEUP   = 0x05, // 101,
    BDR_CLIFFUP   = 0x06, // 110,
    BDR_WALL      = 0x07, // 111,

    BDR_MASK_LENGTH = 3,
    
    CELL_VISITED   = 1 << 29,
    CELL_SEEN      = 1 << 28,
    CELL_REACHABLE = 1 << 27;
    
  static final int[] BDR_SHIFT = { 0,
                                   3,
                                   6,
                                   9,
                                   12,
                                   15 };

  static final int BDR_BITS = 0x07;

  static final int[] BDR_MASK = { BDR_BITS << BDR_SHIFT[0],
                                  BDR_BITS << BDR_SHIFT[1], 
                                  BDR_BITS << BDR_SHIFT[2],
                                  BDR_BITS << BDR_SHIFT[3],
                                  BDR_BITS << BDR_SHIFT[4],
                                  BDR_BITS << BDR_SHIFT[5] };
  
  
  // *************************** direkte Info aus der World ************************* //
  private int[] koord; // Koordinaten der Zelle

  private int[] neighbours = {UNKNOWN, 
                              UNKNOWN, 
                              UNKNOWN, 
                              UNKNOWN, 
                              UNKNOWN, 
                              UNKNOWN}; // Von dieser Zelle aus erreichbare Knoten

  private int height = 0; // Hoehe der Zelle in diesem Level

  private int flags = 0;

  // *************************** zusaetzliche Info fuer Algs ****************************//


  // *************************** Konstruktoren ****************************************//
  public cell(){
    this(0,0);
  }

  public cell(int x, int y){
    this.koord = new int [] {x, y};
  }
  
  // *************************** Private Methoden *********************************//
  public static String boolStr(int n){
    String s = "";
    for (int i = 0; i < 32; i++) {
      if (i % 3 == 0) {
        s=" "+s; // Luecken fuer bessere Lesbarkeit
      } 
      
      s = (n%2) + s;
      n >>= 1;
    }
    return s;
  }

  private int getBorderMask(int dir){
    return ((flags >> BDR_SHIFT[dir % 6]) & BDR_MASK[0] );
  }


  private boolean compareBorderMask(int dir, int mask){
    return (getBorderMask(dir) == mask);
  }

  private void setBorderMask(int dir, int mask){
    //System.out.println("setBorderMask("+dir+", "+boolStr(mask)+")");
    dir = (dir + 6) % 6;                        // zur Sicherheit vor Ueberlaeufen
    flags = flags & ~BDR_MASK[dir];             // alten Wert loeschen
    //System.out.println("-alten Wert loeschen: "+boolStr(flags)+")");
    flags = flags | (mask << BDR_SHIFT[dir]);   // Maske entsprechend dir verschieben und setzen
    //System.out.println("-neuer Wert         : "+boolStr(flags)+")");
  }


  // *************************** Methoden zum Auslesen *********************************//
  public int getX(){
    return this.koord[0];
  }

  public int getY(){
    return this.koord[1];
  }
  
  public int getHeight(){
    return this.height;
  }
  
  public int getNeighbour(int dir){
    return neighbours[dir % 6];
  }

  public int[] getNeighbours(){
    return (int[]) this.neighbours.clone();
  }

  // ACHTUNG: liefert keine UNKNOWNs zurueck
  public int[] getChilds(){
    int[] res = new int[6];
    for (int i = 0; i < 6; i++) {
      if (isAccessible(i) && !isUnknown(i)) {
        res[i] = neighbours[i];
      }
      else res[i] = -1;
    }
    return res;
  }

  public boolean isVisited(){
    return ((flags & CELL_VISITED) > 0);
  }

  public boolean isSeen(){
    return ((flags & CELL_SEEN) > 0);
  }

  public boolean isReachable(){
    return ((flags & CELL_REACHABLE) > 0);
  }

  public boolean neighbourUnknown(int dir){
    //System.out.println("neighbourUnknown("+dir+")");
    return (neighbours[(6+dir)%6] == UNKNOWN);
  }

  // *************************** Methoden zum Testen der Border-Infos *********************************//
  public boolean isUnknown(int dir){
    return compareBorderMask(dir, BDR_UNKNOWN);
  }

  public boolean isCliffDown(int dir){
    return compareBorderMask(dir, BDR_CLIFFDOWN);
  }

  public boolean isSlopeDown(int dir){
    return compareBorderMask(dir, BDR_SLOPEDOWN);
  }

  public boolean isFlat(int dir){
    return compareBorderMask(dir, BDR_FLAT);
  }

  public boolean isSlopeUp(int dir){
    return compareBorderMask(dir, BDR_SLOPEUP);
  }

  public boolean isCliffUp(int dir){
    return compareBorderMask(dir, BDR_CLIFFUP);
  }

  public boolean isWall(int dir){
    return compareBorderMask(dir, BDR_WALL);
  }

  // *************************** Erweiterte Methoden zum Testen der Border-Infos ***************************//
  public boolean isAccessible(int dir){
    return !(    compareBorderMask(dir, BDR_WALL)
              || compareBorderMask(dir, BDR_CLIFFUP) ); //Anm.: bei BDR_UNKNOWN wird TRUE angenommen
  }

  public boolean canSeeNeighbour(int dir){
    return !(    compareBorderMask(dir, BDR_WALL)
              || compareBorderMask(dir, BDR_CLIFFUP)
              || compareBorderMask(dir, BDR_SLOPEUP) ); //Anm.: BDR_UNKNOWN kann nicht angenommen werden
  }

  public boolean isExplored(){
    for (int i=0; i<6; i++) {
      if (isUnknown(i)) { // noch unbekannte Border -> noch nicht vollstaendig entdeckt
        return false; 
      }
    }
    return isSeen(); // alle Borders bekannt, dann nur noch pruefen, ob schon draufgeschaut
  }

  /*
  public boolean hasCorn(){
    return ((corn > 0) || (corn == CORN_AVAILABLE));
  }
  */

  // *************************** Methoden zum Speichern *********************************//
  public void setX(int x){
    this.koord[0] = x;
  }

  public void setY(int y){
    this.koord[1] = y;
  }

  public void setHeight(int h){
    this.height = h;
  }

  /*
  public void setCorn(int c){
    this.corn = c;
  }
  */

  /**
   * Prueft zunaechst, ob ein kuerzerer Weg als der bisher gespeicherte vorliegt. Ist dies der Fall
   * wird dir Richtungsmaske geloescht und die neue Entfernung gespeichert.
   * Im Falle eines mindestens gleich kurzem Weg wird die Richtung markiert.
   * TRUE wird nur zurueckgegeben, falls dies ein kuerzerer Weg war, da das unter Umstaenden eine 
   * Aktualisierung der Nachfolger dieser Zelle auf der Map-Ebene noetwendig macht.
   * 
   * @param dir    Richtung, aus der die Zelle erreicht wurde, d.h. in diese Richtung geht's auf dem Weg der
   *               Laenge dist zum Heimatfeld
   * @param dist   Laenge des kuerzesten Weges zum Heimatfeld ueber die Richtung dir
   */
  /*
  public boolean setWay(int dir, int dist){
    boolean ret = false;
    if (dist < distToWP) { // kuerzerer Weg gefunden,
      dirWP = 0; // dann andere gespeicherte Richtungen loeschen
      distToWP = dist; // neue Entfernung
      ret = true; // Karte muss aktualisiert werden
    }
    if (dist == distToWP) { // kuerzer oder gleich -> Richtung hinzufuegen
      dirWP = dirWP | (1 << ((6+dir)%6)); //  Maske setzen
    }
    return ret;
  }

  public void resetWay(){
    dirWP = 0;
    distToWP = CANT_REACH; // weiiiit weg ;-)
  }

  //deprecated
  public void setDistToWP(int d){              
    this.distToWP = d;
  }

  //deprecated
  public void setDirToWP(int dir){
    this.dirWP = (6+dir)%6;
  }
  */

  public void setNeighbour(int dir, int n){
    this.neighbours[(6+dir)%6] = n;
  }
  
  public void setVisited(){
    flags |= CELL_VISITED;
    //System.out.println(boolStr(flags));
  }

  public void setSeen(){
    flags |= CELL_SEEN;
  }

  public void setReachable(){
    flags |= CELL_REACHABLE;
  }

  // *************************** Methoden zum Speichern der Border-Infos *********************************//
  public void setUnknown(int dir){
    setBorderMask(dir, BDR_UNKNOWN);
  }

  public void setCliffDown(int dir){
    setBorderMask(dir, BDR_CLIFFDOWN);
  }

  public void setSlopeDown(int dir){
    setBorderMask(dir, BDR_SLOPEDOWN);
  }

  public void setFlat(int dir){
    setBorderMask(dir, BDR_FLAT);
  }

  public void setSlopeUp(int dir){
    setBorderMask(dir, BDR_SLOPEUP);
  }

  public void setCliffUp(int dir){
    setBorderMask(dir, BDR_CLIFFUP);
  }

  public void setWall(int dir){
    setBorderMask(dir, BDR_WALL);
  }

  //schnittstelle zu Hamsterkonstanten
  public void setBorder(int dir, int view){
    //System.out.println("setBorder("+dir+", "+boolStr(view)+")");
    switch (view) {
    case CLIFFDOWN: 
      setBorderMask(dir, BDR_CLIFFDOWN);
      break;
    case SLOPEDOWN: 
      setBorderMask(dir, BDR_SLOPEDOWN);
      break;
    case SLOPEUP: 
      setBorderMask(dir, BDR_SLOPEUP);
      break;
    case CLIFFUP: 
      setBorderMask(dir, BDR_CLIFFUP);
      break;
    case WALL: 
      setBorderMask(dir, BDR_WALL);
      break;
    default: 
      setBorderMask(dir, BDR_FLAT);
      break;
    }
  }

  /**
   * setzt das Gegenstueck (Border) von dem, was gerade gesehen wird (view)
   * 
   * @param dir    Richtung, der Border, die gesetzt werden soll
   * @param view   direkter Wert aus lookForwrd() & mask
   */
  public void setComplementBorder(int dir, int view){
    switch (view) {
    case CLIFFDOWN: 
      setBorderMask(dir, BDR_CLIFFUP);
      break;
    case SLOPEDOWN: 
      setBorderMask(dir, BDR_SLOPEUP);
      break;
    case SLOPEUP: 
      setBorderMask(dir, BDR_SLOPEDOWN);
      break;
    case CLIFFUP: 
      setBorderMask(dir, BDR_CLIFFDOWN);
      break;
    // ab hier aendert sich nix mehr
    case WALL: 
      setBorderMask(dir, BDR_WALL);
      break;
    default: 
      setBorderMask(dir, BDR_FLAT);
      break;
    }
  }

  

  // *************************** Methoden zum Zeichnen der Zelle *********************************//
  public String toString(){
    String s = "Koord: ["+koord[0]+", "+koord[1]+"], Level: "+height/*+", distToWP: "+distToWP*/
      +",seen: "+isSeen()+", visited: "+isVisited()+", reachable: "+isReachable()+"\n"+boolStr(flags)+"\n";
    //System.out.println(boolStr(flags));
    for (int i=0; i < 6; i++) {
      s=s+"["+neighbours[i]+": ";
      //System.out.println(boolStr(getBorderMask(i)));
      switch (getBorderMask(i)) {
      case BDR_UNKNOWN :
        s=s+"? ] ";//"UNKN] ";
        break;
      case BDR_CLIFFDOWN :
        s=s+"> ] ";//"CLDN] ";
        break;
      case BDR_SLOPEDOWN :
        s=s+"\\ ] ";//"SLDN] ";
        break;
      case BDR_FLAT :
        s=s+"- ] ";
        break;
      case BDR_SLOPEUP :
        s=s+"/ ] ";//"SLUP] ";
        break;
      case BDR_CLIFFUP :
        s=s+"< ] ";//"CLUP] ";
        break;
      case BDR_WALL :
        s=s+"| ] ";//"WALL] ";
        break;           
      }
    }
    return s;
  }

}
