package scrat;

import algds.c2002.Hamster;
import algds.c2002.Hamster2;
import algds.c2002.Constants;
//import algds.c2002.*;

import java.util.*;

/**
 * 
 * @author Sebastian Stober
 * @since Apr. 2002
 */
public class map implements algds.c2002.Constants{

  private cell[] nodes; // dyn. Feld fuer die Knoten
  private int nodesSize = 0;

  private int[][] map;
  private int[][][] heuristics;
  boolean[][] marked;
  
  private int mx,my; // mitte der Karte
  private int currentNode = 0;                         // Startknoten
  private int viewDir = 1;                             // Startblickrichtung -> N


  final static int WALK_LUT[][] = {{-1,0},{0,-1},{+1,-1},{+1,0},{0,+1},{-1,+1}}; // aus MazeConstants
  
  private final static int 
    CANT_REACH      = Integer.MAX_VALUE,
    CORN_UNKNOWN    = -2,
    CORN_AVAILABLE  = -1,
    CORN_MORETHAN   = 1000,
                    
    DIFF_SIZE       = 100,
    VALUE_COUNT     = 4, 
    HEURISTIC_COUNT = 1,
      
    DIST_FROM_WP    = 0, // [0] => Entfernung
    DIR_FROM_WP     = 1, // [1] => Wegrichtung (Kanten im Graph)
    CORN_COUNT      = 2; // [2] => Anzahl Maiskoerner
  
  private int[][] values;
  
  private sortableSet cornNodes;    // speichert ids der Zellen, auf denen Mais liegt
  private sortableSet exploreNodes; // speichert ids der Zellen, die noch erkundet werden muessen

  private sortableSet open;
  private sortableSet closed;

  private sortableSet goals;

  private sortableSet dropNodes; // Felder, auf die Corn geworfen werden kann

  private int FADtolerance = 3;
  private int FADsearchDepth = 3;
  
  // *************************** Konstruktoren ****************************************//
  public map(int corn){
    cornNodes = new sortableSet(DIST_FROM_WP);
    exploreNodes = new sortableSet(DIST_FROM_WP);
    open = new sortableSet(DIST_FROM_WP);
    closed = new sortableSet(DIST_FROM_WP);

    goals = new sortableSet(DIST_FROM_WP);

    values = new int[DIFF_SIZE][VALUE_COUNT];

    nodes = new cell[DIFF_SIZE];
    for (int l = 0; l < nodes.length; l++) {
      nodes[l] = new cell();
    }

    mx = MAZELENGTH_MAX+2;
    my = MAZELENGTH_MAX+2;
    
    int maxsize = mx * 2;
    
    map = new int[maxsize][maxsize]; //worst-case
    for (int i = 0; i < map.length; i++) {
      for (int j = 0; j < map.length; j++) {
        map[i][j] = -1; // Startwerte setzen
      }
    }
    setMapEntry(0,0,0); // Heimatfeld eintragen

    heuristics = new int[maxsize][maxsize][HEURISTIC_COUNT];
    marked = new boolean[maxsize][maxsize];

    cell c = new cell(0,0); // neue Zelle erzeugen
    values[0][CORN_COUNT] = corn;
    c.setHeight(0);
    c.setSeen();
    
    dropNodes = new sortableSet(DIST_FROM_WP);

    add(c);
    
  }


  private void add(cell n){
    //if (debug) System.out.println("== add ==");
    //if (debug) System.out.println("\n "+nodes.length+ " < "+ nodesSize);
    
    if ((nodes.length <= nodesSize)) { // kein platz mehr
      if (debug) System.out.println("!!!!!!!!!!!!    ADD -> resize   !!!!!!!!!!!!!!!");
      resize(DIFF_SIZE); // array vergroessern
    }
    nodes[nodesSize++] = n;
  }

  private void resize(int diff){

    /* zunaechst nodes[] erweitern */
    int j;
    cell[] buffer = new cell[nodes.length]; // Pufferfeld anlegen
    for (j=0; j < buffer.length; j++) {
      buffer[j] = new cell();
    }
    System.arraycopy(nodes, 0, buffer, 0,nodes.length); // alte Werte in buffer sichern
    
    nodes = new cell[buffer.length + diff]; // Array vergroessern
    for (j=0; j < nodes.length; j++) {
      nodes[j] = new cell();
    }
    System.arraycopy(buffer, 0, nodes, 0,Math.min(nodes.length,buffer.length)); 
    // Werte aus buffer zurueckkopieren

    /* jetzt values[] erweitern */

    int[][] buffer2 = (int[][]) values.clone();
    
    values = new int[buffer2.length + diff][VALUE_COUNT]; // Array vergroessern
    
    int l = Math.min(values.length,buffer2.length);
    
    for (j = 0; j < l; j++) {
      System.arraycopy(buffer2[j], 0, values[j], 0, VALUE_COUNT); // alte Werte aus buffer zurueckschreiben
    }
    
    for (j = l; j < values.length; j++) {
       Arrays.fill(values[j], 0); // restliche Werte mit Nullen fuellen
    }
  }

  
  private int getMapEntry(int x, int y){
    return map[mx+x][my+y];
  }

  private void setMapEntry(int x, int y, int id){
    map[mx+x][my+y] = id;
  }

  

  /**
   * aktualiesiert die Karte, falls noetig
   * 
   * @param lookForward
   *               direkter Wert aus lookForward()
   */
  public void updateMap(int corn, int lookForward, int heightDifference){
    //cell c = getCurrentCell();
    cell c = nodes[currentNode];
    cell n;
    if (!c.isVisited()) { // Zelle noch nicht besucht
      c.setSeen();
      c.setVisited();
      c.setReachable();
      exploreNodes.remove(currentNode); //kann jetzt aus Liste gestrichen werden
      updateCorn(corn);
    }
    if (c.neighbourUnknown(viewDir)) { // Nachbarzelle noch unbekannt
     /* Nachbarzelle erzeugen */
      c.setNeighbour(viewDir,addNeighbourCell(heightDifference));
      
     /* Border setzen */
      c.setBorder(viewDir, lookForward & WALLMASK); 

     /* Gegenstueck zur Border setzen */
      n = nodes[c.getNeighbour(viewDir)];
      n.setComplementBorder(viewDir+3, lookForward & WALLMASK);
      
     /* Explore-Status bestimmen */
      if (c.isAccessible(viewDir)) { // keine Wand, kein Cliffup, dann muss das Feld auch erreichbar sein
        //if (debug) System.out.println("1 -> fuege "+c.getNeighbour(viewDir)+" in exploreNodes.");
        n.setReachable(); // kann die Zelle erreichen
        exploreNodes.insert(c.getNeighbour(viewDir)); // hinzufuegen zu unerkundeten Feldern
      }
    }
    else { // Nachbarzelle existiert schon
      n = nodes[c.getNeighbour(viewDir)];
      if (c.isUnknown(viewDir)) { //Border in Blickrichtung noch unbekannt
       /* Border setzen */
        c.setBorder(viewDir, lookForward & WALLMASK); 
        
       /* Gegenstueck zur Border setzen */
        n.setComplementBorder(viewDir+3, lookForward & WALLMASK); 

       /* Explore-Status bestimmen */
        if (!n.isExplored()) { // noch nicht vollstaendig entdeckt
          if (c.isAccessible(viewDir)) { // Feld erreichbar
            //if (debug) System.out.println("1 -> fuege "+c.getNeighbour(viewDir)+" in exploreNodes.");
            n.setReachable(); // kann die Zelle erreichen
            exploreNodes.insert(c.getNeighbour(viewDir)); // hinzufuegen zu unerkundeten Feldern
          }
        }
        else { // Nachbar ist jetzt vollstaendig erkundet
          exploreNodes.remove(c.getNeighbour(viewDir));
        }                            
      }
      
      // else -> interessiert nicht, nix neues
    }
    
    // Versuch, auf Nachbarfeld nach corn zu sehen, falls Corn unbekannt ist
    if (values[c.getNeighbour(viewDir)][CORN_COUNT] == CORN_UNKNOWN){
      // Cornmenge auf Nachbarfeld noch unbestimmt
      if (c.canSeeNeighbour(viewDir)) {
        //if (debug) System.out.println("kann aufs Nachbarfeld sehen");
        // kann aufs Nachbarfeld sehen
        n.setSeen();
        if (n.isExplored()) { // vollstaendig entdeckt
          exploreNodes.remove(c.getNeighbour(viewDir));
        }
        if ((lookForward & CORNMASK) == CORN) {
          cornNodes.insert(c.getNeighbour(viewDir));
          values[c.getNeighbour(viewDir)][CORN_COUNT] = CORN_AVAILABLE;
        }
        else{
          values[c.getNeighbour(viewDir)][CORN_COUNT] = 0;
        }
        // Cornmenge unbekannt bzw. kein corn vorhanden.
      }
    }
  }    
  
  /**
   * erzeugt in Blickrichtung eine neue Zelle, verknuepft diese mit der aktuellen und errechnet die 
   * Koordinaten und die Hoehe
   * 
   * @return id der neuen Zelle
   */
  private int addNeighbourCell(int heightDifference){
    int id = nodesSize; // id der neuen Zelle
    cell c = nodes[currentNode]; // aktuelle Zelle ermitteln
    cell n = new cell(c.getX()+WALK_LUT[(viewDir+6)%6][0], 
                      c.getY()+WALK_LUT[(viewDir+6)%6][1]); // neue Koordinaten berechnen
    setMapEntry(n.getX(),n.getY(),id); // in Karte eintragen
    n.setHeight(c.getHeight()+heightDifference); // Hoehe der Nachbarzelle berechnen
    
    // hier muessen jetzt noch Verbindungen zu anderen existierenden Zellen hergestellt werden:
    int id2;
    for (int i = 0; i < 6; i++) { // alle Richtungen bearbeiten
      id2 = getMapEntry(n.getX()+WALK_LUT[i][0], n.getY()+WALK_LUT[i][1]);
      if (id2 >= 0) { // Nachbarzelle schon eingetragen, dann ist der Link zu dieser noch unbekannt
        n.setNeighbour(i, id2);              // Link von n zum Nachbar
        nodes[id2].setNeighbour(i+3, id);  // Link vom Nachbar zu n
      }
      // else -> Nachbar noch unbekannt -> INIT-Zustand der neuen Zelle
    }
    
    add(n); // neue Zelle zum Nodes-Vektor hinzufuegen
    values[id][CORN_COUNT] = CORN_UNKNOWN; // Koernerzahl initialisieren
    return id; // liefert id der neuen Zelle

  }

  // *************************** Wegplanungsmethoden ****************************************//
  
  private void setHeuristicValue(int x, int y, int index, int value, boolean isMarked){
    heuristics[mx+x][my+y][index] = value;
    marked[mx+x][my+y] = isMarked;
  }

  private int getHeuristicValue(int x, int y, int index){
    return heuristics[mx+x][my+y][index];
  }

  private void resetHeuristic(int index){
    for (int i1=0; i1 < heuristics.length; i1++) {
      for (int i2=0; i2 < heuristics.length; i2++) {
        heuristics[i1][i2][index] = Integer.MAX_VALUE;
        marked[i1][i2] = false;
      }
    }
  }


  private void computeHeuristic(int index, int startNode){
    final int OFFSET = 10000;
    simpleSet expandNodes = new simpleSet();
    simpleSet newNodes = new simpleSet();
    simpleSet oldNodes = new simpleSet();
    //simpleSet oldNodes

    int startID = (nodes[startNode].getX()+mx)*OFFSET + (nodes[startNode].getY()+my);

    //boolean[][] marked = new boolean[this.map.length][this.map.length];
    
    /*
    for (int i1 = 0; i1 < marked.length; i1++) {
      for (int i2 = 0; i2 < marked.length; i2++) {
        heuristics[i1][i2][index] = Integer.MAX_VALUE;
        marked[i1][i2] = false;
      }
    }
    */
    
    extendedBSF(startNode,1,new simpleSet[] {});  // gesamte Karte generieren
    // alle erreichten Knoten stehen jetzt in closed
    for (int i = 0; i < closed.getSize(); i++) {
      setHeuristicValue(nodes[closed.get(i)].getX(),
                        nodes[closed.get(i)].getY(),
                        index,                        
                        values[closed.get(i)][DIST_FROM_WP],
                        true);
    }
    // jetzt sind alle bekannten Entfernungen von StartNode in die HeuristicMap eingetragen
    // als naechstes muessen Luecken gefuellt werden
    int minDist;
    //int depth = Math.max(nodes[currentNode].getX(), nodes[currentNode].getY()) + 2;
    expandNodes.insert(mx*OFFSET + my); // Heimatfeld
    if (debug) System.out.println("Home = "+(mx*OFFSET + my)+" -> x = "+mx+"  y = "+my);
    
    int n,n2,dir,id,dist;
    int xpos,ypos;
    startNode = nodes[startNode].getX()*OFFSET + nodes[startNode].getY();
    while (!oldNodes.contains(startNode)) {
      newNodes.removeAll();
      while (!expandNodes.isEmpty()) {
        n = expandNodes.popLast(); // von hinten wegnehmen geht schneller -> Stack
        xpos = n / OFFSET;
        ypos = n % OFFSET;
        if (debug) System.out.println("n = "+n+" -> x = "+xpos+"  y = "+ypos
                                      +"  expandNodes="+expandNodes.getSize());
        debugWait(500);
        id = map[xpos][ypos];
        

        // die Nachbarn ermitteln
        for (dir = 0; dir < 6; dir++) {
          n2 = (xpos+WALK_LUT[dir][0])*OFFSET + ypos+WALK_LUT[dir][1];
          if (!oldNodes.contains(n2)) {
            debugWait(500);
            if (debug) System.out.println("insert("+n2+")"
                                        +" newNodes = "+newNodes.getSize());
            newNodes.insert(n2);
          }
          else {
            if (debug) System.out.println("oldNode");
          }
          
        }

        if (heuristics[xpos][ypos][index] == CANT_REACH) { // Entfernung unbekannt
          if (id < 0) {  // unbekannt
            // wir kennen dieses Feld noch gar nicht, daher auch die Borders nicht
            minDist = Integer.MAX_VALUE;
            for (dir = 0; dir < 6; dir++) {
              dist = heuristics[ xpos + WALK_LUT[dir][0] ][ ypos + WALK_LUT[dir][1] ][index];
              if (dist < minDist) { 
                minDist = dist;  
              }
            }
            heuristics[xpos][ypos][index] = minDist+1;
          }
          else { // ist schon bekannt, kann also sein, dass man wirklich nicht hinkommnt
            if (!nodes[id].isExplored()) { // also keine "insel"
              minDist = Integer.MAX_VALUE;
              for (dir = 0; dir < 6; dir++) {
                if (!nodes[id].isWall(dir) && !nodes[id].isCliffUp(dir)) {
                  // zusaetzliche Abfrage, damit nicht versucht wird, durch Waende zu gehen
                  dist = heuristics[ xpos + WALK_LUT[dir][0] ][ ypos + WALK_LUT[dir][1] ][index];
                  if (dist < minDist) { 
                    minDist = dist;  
                  }  
                }
              }
              heuristics[xpos][ypos][index] = minDist+1;
            }                                      
          }
        } // end if
        oldNodes.insert(n);
        expandNodes.remove(n); // loeschen
        if (debug) System.out.println("removed("+n+")   size = "+expandNodes.getSize()
                                      +" oldNodes = "+oldNodes.getSize());
        debugWait(500);
      }
      //expandNodes = newNodes.clone;
      for (n = 0; n < newNodes.getSize(); n++) {
        expandNodes.insert(newNodes.get(n));
      }

    }
    
  }

  // alternative Methode
  public int[] getNextWPHome(){
    // berechne fuer alle unentdeckten die Entfernung 
    for (int i = 0; i < exploreNodes.getSize(); i++) {
      // alle unexplored resetten, sonst funktioniert das nicht
      values[exploreNodes.get(i)][DIST_FROM_WP] = CANT_REACH; 
    }
    extendedBSF(currentNode,exploreNodes.getSize(),new simpleSet[] {exploreNodes}); 
    //computeHeuristic(0,currentNode); // HeuristicKarte berechnen
    // durch den Aufruf von extendedBSF(alle) haben wir schon die Entferungen der WPs.
    int minE = -1;
    int minDist = Integer.MAX_VALUE;
    int minDistHome = Integer.MAX_VALUE;
    int dist, distHome;
    for (int i = 0; i < exploreNodes.getSize(); i++) {
      /*distHome = getHeuristicValue(nodes[exploreNodes.get(i)].getX(),
                                   nodes[exploreNodes.get(i)].getY(),
                                   0);
      */
      distHome = Math.abs(nodes[exploreNodes.get(i)].getX())
               + Math.abs(nodes[exploreNodes.get(i)].getY());
        // Schaetzung
      dist = values[exploreNodes.get(i)][DIST_FROM_WP] + distHome;
      if (dist < 0) {
        dist = CANT_REACH; // wegen Uberlauf
      }
      if (debug) System.out.println("n = "+exploreNodes.get(i)+" dist = "
                                  +values[exploreNodes.get(i)][DIST_FROM_WP]+" + "
                                  +distHome+" = "+dist); 

         
      if ((minDist > dist)       // insgesamt kuerzer
          || ((minDist == dist) && (minDistHome > distHome))) { // bei Gleichheit, kleinere Heuristic
        if (debug) System.out.println("neues Minimum");
        minE = exploreNodes.get(i);
        minDist = dist;
        minDistHome = distHome;
      }
    }
    if (debug) System.out.println("minE = "+minE+" dist = "
                                  +values[minE][DIST_FROM_WP]+" + "
                                  +minDistHome+" = "+minDist); 
    extendedBSF(currentNode,1,new simpleSet[] {new simpleSet(new int[] {minE})}); 
    return doBacktracing();
  }


  // *************************** Bewegungsmethoden ****************************************//
  public int getDirection(){
    return viewDir;
  }

  public boolean canGoForward(){
    if (nodes[currentNode].isAccessible(viewDir)) {
      return true;
    }
    else {
      return false;
    }
  }

  // vorwaerts gehen
  public void goForward(){
    currentNode = nodes[currentNode].getNeighbour(viewDir); // naechste Zelle ermitteln
  }
  
  // drehe im Uhrzeigersinn
  public void turnPos(){
    viewDir++;
    viewDir %= 6;
  }

  // drehe gegen den Uhrzeigersinn
  public void turnNeg(){
    viewDir += 5; // entspricht -1 im Modul 6
    viewDir %= 6;
  }
  
  private int[] doBacktracing(){
    //debugWait();
    int n = goals.getMin(); // sortieren
    if (n== -1) {
      if (debug) System.out.println("goals is empty.");
      return new int[] {-1};
    }
                            //if (debug) System.out.println(goals);
    if ((values[n][DIST_FROM_WP] == CANT_REACH) || (values[n][DIST_FROM_WP] < 0)){ //fiese Falle !!!
                            // "<=" , da wenn == schon da ;-)
      if (debug) System.out.println("Cant reach.");
      return new int[] {-1};
    }
    if (values[n][DIST_FROM_WP] == 0) {
      if (debug) System.out.println("Already there.");
      return new int[] {-2};
    }
    int[] ret = new int[values[n][DIST_FROM_WP]+1]; // maximale WegLaenge
    ret[0] = n;  // Ziel wird im 0ten Element gespeichert
    int i = ret.length;     //rueckwaerts gehen!!!
                            //if (debug) System.out.println("size="+goals.getValue(0)+"+1,  i="+i);
                            //if (debug) System.out.println("values["+n+"] -> "+values[n][DIST_FROM_WP]);
    int dir;
    while (n != currentNode) { //noch nicht da
                            //if (debug) System.out.println("dir = "+ cell.boolStr(values[n][DIR_FROM_WP]));
      for (dir = 0; dir < 6; dir++) {
        if ((values[n][DIR_FROM_WP] & (1 << dir)) > 0) {
// <modifitziert>
          //lastOKDir = dir;
         /* if (nodes[nodes[n].getNeighbour(dir)].getHeight() >= nodes[n].getHeight()) {
            if (debug) System.out.println("Prioritaet: "
                               +nodes[nodes[n].getNeighbour(dir)].getHeight() +" <= "
                               +nodes[n].getHeight());
         */   
            break;
         // }
 // </modifitziert>


        //old:  break;            // gibt die erstbeste zurueck, auch wenn es mehrere gibt !!!
        }
      }
      //dir = lastOKDir;                            
                            //if (debug) System.out.println(n+" != "+currentNode+" -> ret["+(i-1)+"] = "+(dir+3)
                            //                   +" dist = "+values[n][DIST_FROM_WP]);
      ret[--i] = dir + 3;   // Richtung eintragen
      n = nodes[n].getNeighbour(dir); // naechste Zelle ermitteln
    }
    return ret;
    // liefert Knoten mit kleinster Entfernung
  }
  
  private boolean distanceControl = true;
  
  public void setDistanceControl(boolean b){
    distanceControl = b;
  }

  private boolean nextWPTooFar(int dest){
    if (debug) System.out.println("check Entferung von "+dest);
    int nutzen = 200;//Math.min(200, values[dest][CORN_COUNT] * 10);
    if (nutzen < 0) {
      nutzen = 200; // bei unbekannter Cornmenge
    }           
    if (!nodes[dest].isExplored()) {
      nutzen = 200;
    }
    if (debug) System.out.print("Nutzen = "+nutzen+" ,");

    // Weg Home -> Corn
    extendedBSF(0,1,new simpleSet[] {new simpleSet(new int[] {dest})}); 
    int distHomeDest = values[dest][DIST_FROM_WP];
     
    // Weg Dest -> Home
    extendedBSF(dest,1,new simpleSet[] {new simpleSet(new int[] {0})}); 
    int distDestHome = values[0][DIST_FROM_WP];
     
    // Weg WP -> Corn
    extendedBSF(currentNode,1,new simpleSet[] {new simpleSet(new int[] {dest})}); 
    int distWPDest = values[dest][DIST_FROM_WP];
    
    int aufwand = distDestHome + Math.min(distHomeDest, // Transportweg home<->corn
                                          distWPDest);
    if (debug) System.out.println("aufwand = "+aufwand );
    if (nutzen > aufwand) {
      return false;
    }
    /*
     * hier entsteht ein Problem, wenn auf dem Feld mehr Corn liegt und man auf dem Rueckweg dropForward()
     * nutzen koennte, weil dann die Hin-Weg Kosten u.U. nicht mehr so stark ins Gewicht fallen
     * (mit jeder Ladung nimmt der Anteil des Hin-Weges ab) -> ist aber ein Extremfall
     **/                                                  
    return true;
  }


  public int[] getNextUnexplored(){
    if (exploreNodes.isEmpty()) {
      return new int[] {-1};
    }
    if (!distanceControl) { // egal, wie weit weg
      extendedBSF(currentNode,1,new simpleSet[] {exploreNodes});
      return doBacktracing();
    }
    else { // pruefen, ob es sich ueberhaupt noch lohnt
      extendedBSF(currentNode,exploreNodes.getSize(),new simpleSet[] {exploreNodes}); // alle besuchen
      // alle exploreNodes, die erreicht wurden, stehen jetzt in goals
      // nextWPTooFar veraendert goals, daher sichern:
      sortableSet myGoals = (sortableSet) goals.clone();
      int dest = -1;
      while (!myGoals.isEmpty()) {
        dest = myGoals.popMin(); // probieren wir's mal mit dem Minimum
        if (!nextWPTooFar(dest)) { // nicht zu weit weg
          break; // while
        }
        else {
          dest = -1;
        }
      }
      if (dest < 0) { // keinen Erfolg gehabt?
        return new int[] {-1};
      }
      // haben einen gefunden:
      extendedBSF(currentNode,1,new simpleSet[] {new simpleSet(new int[] {dest})}); 
      return doBacktracing();
    }  
  }

  public int[] getNextCorn(){
    if (cornNodes.isEmpty()) {
      return new int[] {-1};
    }
    if (!distanceControl) { // egal, wie weit weg
      extendedBSF(currentNode,1,new simpleSet[] {cornNodes}); 
      //printCornNodes();
      //if (debug) System.out.println("goals: "+goals);
      return doBacktracing();
    }
    else { // pruefen, ob es sich ueberhaupt noch lohnt
      extendedBSF(currentNode,cornNodes.getSize(),new simpleSet[] {cornNodes}); // alle besuchen
      // alle cornNodes, die erreicht wurden, stehen jetzt in goals
      // nextWPTooFar veraendert goals, daher sichern:
      sortableSet myGoals = (sortableSet) goals.clone();
      int dest = -1;
      while (!myGoals.isEmpty()) {
        dest = myGoals.popMin(); // probieren wir's mal mit dem Minimum
        if (!nextWPTooFar(dest)) { // nicht zu weit weg
          break; // while
        }
        else {
          dest = -1;
        }
      }
      if (dest < 0) { // keinen Erfolg gehabt?
        // na gut, dann versuchen wir mal die dropNodes:
        rememberDropNodes();
        extendedBSF(currentNode,cornNodes.getSize(),new simpleSet[] {cornNodes}); // alle besuchen
        // alle cornNodes, die erreicht wurden, stehen jetzt in goals
        // nextWPTooFar veraendert goals, daher sichern:
        myGoals = (sortableSet) goals.clone();
        dest = -1;
        while (!myGoals.isEmpty()) {
          dest = myGoals.popMin(); // probieren wir's mal mit dem Minimum
          if (!nextWPTooFar(dest)) { // nicht zu weit weg
            break; // while
          }
          else {
            dest = -1;
          }
        }
        if (dest < 0) { // wieder keinen Erfolg gehabt?
          return new int[] {-1};
        }
        // nochmal Schwein gehabt, was?
        extendedBSF(currentNode,1,new simpleSet[] {new simpleSet(new int[] {dest})}); 
        return doBacktracing();
      }
      // haben einen gefunden:
      extendedBSF(currentNode,1,new simpleSet[] {new simpleSet(new int[] {dest})}); 
      return doBacktracing();
    }  
  }

  public int[] getWayHome(){
    extendedBSF(currentNode,1,new simpleSet[] {new simpleSet(new int[] {0})}); 
    return doBacktracing();
  }

  public int[] getNextWP(){
    //if (debug) System.out.println("getNextWP()");
    if (!distanceControl) { // egal, wie weit weg
      extendedBSF(currentNode,1,new simpleSet[] {cornNodes, exploreNodes});
      return doBacktracing();
    }
    else { // pruefen, ob es sich ueberhaupt noch lohnt
      extendedBSF(currentNode,
                  cornNodes.getSize()+exploreNodes.getSize(),
                  new simpleSet[] {cornNodes, exploreNodes}); // alle besuchen
      // alle nodes, die erreicht wurden, stehen jetzt in goals
      // nextWPTooFar veraendert goals, daher sichern:
      sortableSet myGoals = (sortableSet) goals.clone();
      int dest = -1;
      while (!myGoals.isEmpty()) {
        dest = myGoals.popMin(); // probieren wir's mal mit dem Minimum
        //if (debug) System.out.println("Min = "+dest);
        //if (debug) System.out.println(myGoals);
        //if (debug) System.out.println("myGoals.isEmpty() -> "+myGoals.isEmpty());
        if (!nextWPTooFar(dest)) { // nicht zu weit weg
          break; // while
        }
        else {
          dest = -1;
        }
      }
      if (dest < 0) { // keinen Erfolg gehabt?
        // na gut, dann versuchen wir mal die dropNodes:
        rememberDropNodes();
        extendedBSF(currentNode,cornNodes.getSize(),new simpleSet[] {cornNodes}); // alle besuchen
        // alle cornNodes, die erreicht wurden, stehen jetzt in goals
        // nextWPTooFar veraendert goals, daher sichern:
        myGoals = (sortableSet) goals.clone();
        dest = -1;
        while (!myGoals.isEmpty()) {
          dest = myGoals.popMin(); // probieren wir's mal mit dem Minimum
          if (!nextWPTooFar(dest)) { // nicht zu weit weg
            break; // while
          }
          else {
            dest = -1;
          }
        }
        if (dest < 0) { // wieder keinen Erfolg gehabt?
          return new int[] {-1};
        }
        // nochmal Schwein gehabt, was?
        extendedBSF(currentNode,1,new simpleSet[] {new simpleSet(new int[] {dest})}); 
        return doBacktracing();
      }
      // haben einen gefunden:
      extendedBSF(currentNode,1,new simpleSet[] {new simpleSet(new int[] {dest})}); 
      return doBacktracing();
    }          
  }

  /**
   * da das Ziel im 0tn Element gespeichert wird und beim Heimweg immer 0 sein muss,
   * wird hier ein spezieller Wert gespeichert, der signalisiert, dass das Corn nach 
   * vorn abgeworfen werden kann
   * 
   * @return 
   */
  private int[] doTransportBacktracing(){
    //debugWait();

    int n = goals.getMin(); // sortieren
    if (n== -1) {
      if (debug) System.out.println("goals is empty.");
      return new int[] {-1};
    }
                            //if (debug) System.out.println(goals);
    if (values[n][DIST_FROM_WP] == CANT_REACH) { //fiese Falle !!!
                            // "<=" , da wenn == schon da ;-)
      if (debug) System.out.println("Cant reach.");
      return new int[] {-2};
    }
    if (values[n][DIST_FROM_WP] <= 0){ //fiese Falle !!!
                            // "<=" , da wenn == schon da ;-)
      if (debug) System.out.println("Already there.");
      return new int[] {-3};
    }
    int[] normalWay = new int[values[n][DIST_FROM_WP]+1]; // maximale WegLaenge
    boolean[] dropPossible = new boolean[normalWay.length]; // markiert Stellen, an denen dropForward() geht
// das kann noch geaendert werden //    
    normalWay[0] = 0;  // Ziel ist Heimatfeld

    int wayNodes[] = new int[normalWay.length]; // speichert die durchlaufenen Knoten
    wayNodes[0] = currentNode; // Startknoten eintragen
    
    int i = normalWay.length;     //rueckwaerts gehen!!!
                            //if (debug) System.out.println("size="+goals.getValue(0)+"+1,  i="+i);
                            //if (debug) System.out.println("values["+n+"] -> "+values[n][DIST_FROM_WP]);
    int dir;
    int lastDir = 0;
    
    while (n != currentNode) { //noch nicht da
      i--;
      wayNodes[i] = n;      // Konten in Abfolge speichern
      //if (debug) System.out.println("n = "+n);
                            //if (debug) System.out.println("dir = "+ cell.boolStr(values[n][DIR_FROM_WP]));
      for (dir = 0; dir < 6; dir++) {
        if ((values[n][DIR_FROM_WP] & (1 << dir)) > 0) {
          lastDir = dir;    // letzte Richtung, die okay war
          if (preferCliffUp) {
            if (nodes[n].isCliffUp(dir)) {
              break;          // optimal
            }
          }
          else {
            if (preferCliffUpHome && (n == 0)) { // Sonderfall: Heimatfeld
              if (nodes[n].isCliffUp(dir)) {
                break;          // optimal
              }
            }
            else {
              // CliffUp egal
              break;
            }       
          }
        }
      }
      dir = lastDir;
      
      if (nodes[n].isCliffUp(dir)) { // es geht nach oben
        if (debug) System.out.println("hier kann dropForward() genutzt werden");
        dropPossible[i] = true;
      }
      else {
        dropPossible[i] = false;
      }
                            //if (debug) System.out.println(n+" != "+currentNode+" -> ret["+(i-1)+"] = "+(dir+3)
                            //                   +" dist = "+values[n][DIST_FROM_WP]);
      normalWay[i] = (dir + 3) % 6;   // Richtung eintragen
      n = nodes[n].getNeighbour(dir); // naechste Zelle ermitteln
    }
    /**
     * an dieser Stelle ist in normalWay der Heimweg gespeichert
     * falls wir irgendwo dropForward nutzen koennen, dann zeigt uns
     * dropPossible[i] an, dass wir statt dem i-ten Schritt einen 
     * dropForward() machen koennen
     * in diesem Fall wird geprueft, ob das Feld noch Platz hat.
     * wenn ja -> okay
     * wenn nein -> findAlternativeDropNode()
     *        -> schlaegt das fehl, wird auf dem normalen Weg weitergegangen
     *           und eventuell die naechste Moeglichkeit genutzt
     */
    i = 1;
    int[] altWay; // speichert einen alternativen Weg
    debugWait();
    while (i < normalWay.length) { // solange suchen, bis Weg zuende
      if (dropPossible[i]) { // auf das i-te Feld kann dropForward() angewendet werden
        if ((wayNodes[i] > 0) // nicht das Heimatfeld
            && (values[wayNodes[i]][CORN_COUNT] == CORNFIELD_MAX)) { // Feld ist schon voll :-o
          // suche nach alternativer Loesung 
          if (debug) System.out.println("findAlternativeDropNode() muss ausgefuehrt werden");
          debugWait();
          altWay = findAlternativeDropNode(currentNode,     // der Startknoten
                                           wayNodes[i-1],   // der Vorgaenger, dh. von hier drop
                                           normalWay[i],    // die Richtung, aus der, der Knoten erreicht wurde
                                           FADsearchDepth); // Suchtiefe nach links und rechts
          if (debug) {
            String s = "";
            for (int j = 0; j < altWay.length; j++) {
              s+=altWay[j]+" ,";
            }
            System.out.println("findAlternativeDropNode() returned "+s);
          } 
          if (altWay[0] >= 0) { // andernfalls haben wir keine Alternative und muessen ganz nach hause
            if (debug) System.out.println("Alternative ist okay");
            return altWay;                
          }
          // else : die if-Teile werden verlassen, dh. dropForward hier nicht moeglich -> gehen weiter
        }
        else { // entweder Heimatfeld -> unbegrenzt Platz, oder noch Platz
          // -> wir koennen statt des i-ten Schittes dropForward() ausfuehren
          altWay = new int[i+1];
          altWay[0] = 1; // Markierung, dass dropForward() genutzt wird
          for (int j = 1; j <= i; j++) {
            altWay[j] = normalWay[j]; // kopieren
          }
          return altWay;
        }
      }     
      i++;
    }
    //debugWait();
    return normalWay; // keine bessere Moeglichkeit gefunden
  }

  public int[] getTransportWayHome(){
    extendedBSF(currentNode,1,new simpleSet[] {new simpleSet(new int[] {0})}); 
    if (advancedWayHome) { // soll dropForward genutzt werden?
      return doTransportBacktracing();   
    }
    return doBacktracing();
  }

  public boolean rememberDropNodes(){
    if (dropNodes.getSize() == 0) {
      return false; // keine dropNodes mehr
    }
    int e;
    for (int i = 0; i < dropNodes.getSize(); i++) {
      e = dropNodes.get(dropNodes.getSize()-1); // letztes Element
      dropNodes.remove(e); // als Sammelplatz loeschen
  /* TODO: der Wert sollte vielleicht auch gespeichert werden */
      cornNodes.insert(e); // in die corn-Liste eintragen
    }
    return true;

  }

  public void cornDroppedForward(int carriedBefore, int carryNow){
    if (debug) System.out.println("cornDroppedForward()");
    int id = nodes[currentNode].getNeighbour(viewDir);
    if (id > 0) { // nicht das Heimatfeld
      if (debug) System.out.println("nicht das Heimatfeld");
      if (values[id][CORN_COUNT] >= 0) { // die Cornmenge ist bekannt
        values[id][CORN_COUNT] += (carriedBefore - carryNow); // neue Menge berechnen    
      }
      else { // Cornmenge ist unbekannt
        if (carryNow == 0) { // alles abgelegt
          values[id][CORN_COUNT] = CORN_MORETHAN + (carriedBefore - carryNow); // neue Menge berechnen    
          // CORN_MORETHAN gilt als Markierung, dass vorher schon eine unbekannte Menge da lag
          if (values[id][CORN_COUNT] == (CORNFIELD_MAX+CORN_MORETHAN-1)) { // voll
            values[id][CORN_COUNT] = CORNFIELD_MAX;
          }
        }
        else { // nicht alles abgelegt -> das sagt uns, dass das Feld jetzt voll ist
          values[id][CORN_COUNT] = CORNFIELD_MAX; // maximale Kapazitaet erreicht
        }                                                                            
      }
      if (debug) System.out.println("loesche aus cornNodes");

      dropNodes.insert(id); // als Sammelplatz abspeichern
      cornNodes.remove(id); // aus der corn-Liste entfernen
    } 
    else if (debug) System.out.println("aufs Heimatfeld abgeworfen"); // der optimalfall
  }

  private int[] findAlternativeDropNode(int startNode, int dropFrom, int dropDir, int searchDepth){
    if (!useAlternativeDropNodes) { // abgestellt
      return new int[] {-1};
    }
    if (debug) System.out.println("findAlternativeDropNode()");
    //final int TOLERANCE = searchDepth; // maximal TOLERANCE Schritte mehr zulassen
    // folge der klippe
    int nodeL = dropFrom, 
        nodeR = dropFrom,
        depthL = 0,
        depthR = 0,
        dirL = (dropDir+5)%6,
        dirR = (dropDir+1)%6;
        //dropNode;
    
    sortableSet altDropNodes = new sortableSet(DIST_FROM_WP);
    // nach rechts suchen
    if (debug) System.out.println("suche rechts");
    while (depthR < searchDepth) {
      if (debug) System.out.println("dirR = "+dirR);
      if (nodes[nodeR].isCliffDown(dirR)){
        if (debug) System.out.println("cliff");
        // es kann hier dropForward() genutzt werden
        if ((values[ nodes[nodeR].getNeighbour(dirR) ][CORN_COUNT] != CORNFIELD_MAX)
            && nodes[nodeR].isReachable()) {
          // es ist noch Platz
          if (debug) System.out.println("frei und dropNode ist erreichbar");
          altDropNodes.insert( nodes[nodeR].getNeighbour(dirR) );
          if (debug) System.out.println(""+altDropNodes);
        } 
        dirR = (dirR+1)%6;
      }
      else { 
        // verfolge das Ganze beim Nachbarn weiter
        if (debug) System.out.println("untersuche rechten nachbar: "+ nodes[nodeR].getNeighbour(dirR));
        nodeR = nodes[nodeR].getNeighbour(dirR);
        dirR = (dirR+4) % 6;
        depthR++;
      }
    }
    debugWait();
    
    // nach links suchen
    if (debug) System.out.println("suche links");
    while (depthL < searchDepth) {
      if (debug) System.out.println("dirL = "+dirL);
      if (nodes[nodeL].isCliffDown(dirL)) {
        if (debug) System.out.println("cliff");
        // es kann hier dropForward() genutzt werden
        if ((values[ nodes[nodeL].getNeighbour(dirL) ][CORN_COUNT] != CORNFIELD_MAX)
            && nodes[nodeL].isReachable()) {
          // es ist noch Platz
          if (debug) System.out.println("frei");
          altDropNodes.insert( nodes[nodeL].getNeighbour(dirL) );
          if (debug) System.out.println(""+altDropNodes);
        } 
        dirL = (dirL+5)%6;
      }
      else { 
        // verfolge das Ganze beim Nachbarn weiter
        if (debug) System.out.println("untersuche linken nachbar: "+ nodes[nodeL].getNeighbour(dirL));
        nodeL = nodes[nodeL].getNeighbour(dirL);
        dirL = (dirL+2) % 6;
        depthL++;
      }
      //debugWait();
    }
    if (debug) System.out.println(""+altDropNodes);
    //debugWait();
    if (altDropNodes.isEmpty()) {
      return new int[] {-1};
    }

    // jetzt die kuerzeste Moeglichkeit suchen    
    // -> zuerst die Entfernung vom Startpunkt berechnen
    extendedBSF(startNode, altDropNodes.getSize()+1, new simpleSet[] {altDropNodes,
                                                                      new simpleSet(new int[] {0})}); 
    int normalWay = values[0][DIST_FROM_WP];
    if (debug) System.out.println("normalWay = "+normalWay);
    // die Werte sichern
    int[] distTo = new int[altDropNodes.getSize()];
    for (int i = 0; i < distTo.length; i++) {
      distTo[i] = values[altDropNodes.get(i)][DIST_FROM_WP];
      if (debug) System.out.println("distTo["+i+"] = "+distTo[i]);
    }
    // Werte gesichert
    // -> jetzt fuer jeden Knoten die Entfernung zum Heimatfeld berechnen
    int minWay = CANT_REACH;
    int minE = -1;
    int[] distFrom = new int[altDropNodes.getSize()];
    for (int i = 0; i < distTo.length; i++) {
      extendedBSF(altDropNodes.get(i), 1, new simpleSet[] {new simpleSet(new int[] {0})}); 
      distFrom[i] = values[0][DIST_FROM_WP];
      if (debug) System.out.println("distFrom["+i+"] = "+distFrom[i]);
      if ((distTo[i] + distFrom[i]) < minWay) {
 // TODO hier kann fuer == noch if corn < corn(minE) -> kann mehr ablegen
        minE = i;
        minWay = distTo[i] + distFrom[i];
      }
    }
    if (debug) System.out.println("minWay = "+minWay+", minE = "+minE);
    // minE enthaelt jetzt den guenstigsten Alternativknoten
    if (minWay > (normalWay + FADtolerance)) {
      if (debug) System.out.println("zu hohe kosten");
      return new int[] {-1}; // kosten zu hoch !
    }
    // ermittlen, von wo dropForward() aufgerufen werden soll
    int newDropDir;
    normalWay = distTo[minE];
    minE = altDropNodes.get(minE); 
    int dropNode = -1;
    for (newDropDir = 0; newDropDir < 6; newDropDir++) {
      if (nodes[minE].isCliffUp(newDropDir)) {
        dropNode = nodes[minE].getNeighbour(newDropDir);
        if (nodes[dropNode].isReachable()) {
          extendedBSF(startNode, 1, new simpleSet[] {new simpleSet(new int[] {dropNode})}); 
          if (values[dropNode][DIST_FROM_WP] < normalWay) {
            break; // hab ihn!
          }
        }
      }
    }

    if (debug) System.out.println("dropNode = "+dropNode);

    if (dropNode < 0) {      // sollte nicht passieren, wenn alles funktioniert ;-)
      if (debug) System.out.println("Fehler in findAlternativeDropNode() -> kein Cliff gefunden");
      return new int[] {-1}; // kosten zu hoch !
    }

    //extendedBSF(startNode, 1, new simpleSet[] {new simpleSet(new int[] {dropNode})}); 
    // -> haben wir schon in der Schleife gemacht
    //int[] way = doTransportBacktracing(); // alternativen Weg zurueckliefern
    extendedBSF(startNode, 1, new simpleSet[] {new simpleSet(new int[] {dropNode})}); 
    int[] way = doBacktracing(); // die einfache Variante tut's auch

    // jetzt haben wir aber nur den Weg bis zum Droppoint
    int[] way2 = new int[way.length+1];
    for (int i = 0; i < way.length; i++) {
      if (debug) System.out.println("way["+i+"] = "+way[i]);
      way2[i] = way[i];
    }
    way2[0] = 1; // markierung fuer drop//way.length;
    way2[way.length] = (newDropDir + 3) % 6;
    if (debug) System.out.println("way["+way.length+"] = "+way2[way.length]);
    if (debug) System.out.println("weg ermittelt");
    return way2;
  }

  public void bfsFullMap(){
    //doBSF(currentNode,1,false,false,new int[] {}); // sucht die ganze Karte ab
    extendedBSF(currentNode,1,new simpleSet[] {}); 
    // liefert aber keine goals zurueck !!!
  }

  public void updateCorn(int corn){
    if (corn > 0) {
      values[currentNode][CORN_COUNT] = corn;
      if (currentNode > 0){ // nicht zuhause
        cornNodes.insert(currentNode);
      }
    }
    else {
      values[currentNode][CORN_COUNT] = 0;
      cornNodes.remove(currentNode);
    }
  }

  public boolean currentCellExplored(){
    return nodes[currentNode].isExplored();
  }

  public boolean checkCornAvailable(){
    //if (debug) System.out.println("cornNodes.size = "+cornNodes.getSize());
    return (cornNodes.getSize() > 0);
  }

  public boolean somethingToExplore(){
    //if (debug) System.out.println("exploreNodes.size = "+exploreNodes.getSize());
    return (exploreNodes.getSize() > 0);
  }

  private boolean allowGotoCornNeighbour         = true;
  private boolean allowGotoCornNeighbourHeightOK = true;
  private boolean allowGotoCornNeighbourBackOK   = false;

  public void setGotoCornNeighbour(boolean b){
    allowGotoCornNeighbour = b;
  }

  public void setGotoCornNeighbourHeightOK(boolean b){
    allowGotoCornNeighbourHeightOK = b;
  }

  public void setGotoCornNeighbourBackOK(boolean b){
    allowGotoCornNeighbourBackOK = b;                 
  }

  public int getDirCornNeighbour(){
    int[] childs = nodes[currentNode].getChilds();
    for (int i = 0; i < 6; i++) {
      if (cornNodes.contains(childs[i])) {
        if (allowGotoCornNeighbour) {
          return i;
        }
        else { // abgestellt
          // sonderfaelle
          if (allowGotoCornNeighbourHeightOK
              && (nodes[childs[i]].getHeight() >= nodes[currentNode].getHeight())) {
            return i;
          }
          if (allowGotoCornNeighbourBackOK
              && nodes[childs[i]].isAccessible(i+3)) {
            return i;
          }
        }            
      }
    }
    return -1;
  }

  public boolean atHome(){
    return (currentNode == 0);
  }

  private void extendedBSF(int start, int goalsMax, simpleSet[] searchGoals){
  
    int i;
    int n;
    int z;
    cell node;

    open.removeAll();
    open.insert(start); // Startknoten zu open hinzufuegen

    closed.removeAll();
    goals.removeAll();

    values[start][DIST_FROM_WP] = 0;
    
    while (!open.isEmpty()) { // kein Knoten mehr zum erweitern, dann aufhoeren
      n = open.extendedPopMin(); // den ersten Knoten aus open holen
      open.remove(n);    // ... und aus open entfernen 
      node = nodes[n];
      
      for (z = 0; z < searchGoals.length; z++) {
        if (searchGoals[z].contains(n)) {
          goals.insert(n);  
          break; // for -schleife
        }
      }

      if (goals.getSize() >= goalsMax) { // Tiefenbegrenzung erreicht
        closed.insert(n); // n zu closed hinzufuegen
        break; // -> na dann Feierabend
      }
  
      // wir sind also noch nicht fertig....
      for (i=0; i < 6; i++) { // dann wollen wir uns doch mal die Nachfolger ansehen
        z = node.getChilds()[i];
        if (z >= 0) { //kann erreicht werden
          if (!(open.contains(z) || closed.contains(z))) { // bisher nicht erreichter Knoten
            open.insert(z);             // zu open hinzufuegen
            values[z][DIST_FROM_WP] = CANT_REACH; // damit setWay() auch Erfolg hat
            //if (debug) System.out.println("voher  values["+z+"] -> "+values[z][DIST_FROM_WP]);
            setWay(z, i+3, values[n][DIST_FROM_WP]+1);
            //if (debug) System.out.println("nacher values["+z+"] -> "+values[z][DIST_FROM_WP]);
          
          }
          else{
            if (setWay(z, i+3, values[n][DIST_FROM_WP]+1)) { // das war ein kuerzerer Weg!
              if (closed.contains(z)) {
                // Nachfolger, die in open u closed sind anpassen
                bfsModifyChilds(z);
              }
              else { // open.contain(z)
              }
            }                              
          }                                
        }
      }            
      //if (debug) System.out.println("ende values["+n+"] -> "+values[n][DIST_FROM_WP]);
      closed.insert(n); // n zu closed hinzufuegen
    }
    //if (open.isEmpty()) if (debug) System.out.println("open is empty");
    // entweder die maximale Tiefe wurde erreicht und es wurde abgebrochen
    // oder alle erreichbaren Knoten wurden erweitert
    
  

  }  

  private void bfsModifyChilds(int node){
    int z;
    for (int i = 0; i < 6; i++) {
      if (nodes[node].isAccessible(i) && !nodes[node].isUnknown(i)) {
        z = nodes[node].getChilds()[i];
        if (open.contains(z) || closed.contains(z)) {
          if (setWay(z, i+3, values[node][DIST_FROM_WP]+1)) { //kuerzerer Weg gefunden
            if (closed.contains(z)) { // Knoten ist schon erweitert
              bfsModifyChilds(z); // rekursiver Aufruf, bis Weg laenger
            }
            else { 
            }
          }
        }
      }
    }
  }

  /**
   * Prueft zunaechst, ob ein kuerzerer Weg als der bisher gespeicherte vorliegt. Ist dies der Fall
   * wird dir Richtungsmaske geloescht und die neue Entfernung gespeichert.
   * Im Falle eines mindestens gleich kurzem Weg wird die Richtung markiert.
   * TRUE wird nur zurueckgegeben, falls dies ein kuerzerer Weg war, da das unter Umstaenden eine 
   * Aktualisierung der Nachfolger dieser Zelle auf der Map-Ebene noetwendig macht.
   * 
   * @param dir    Richtung, aus der die Zelle erreicht wurde, d.h. in diese Richtung geht's auf dem Weg der
   *               Laenge dist zum Heimatfeld
   * @param dist   Laenge des kuerzesten Weges zum Heimatfeld ueber die Richtung dir
   */
  private boolean setWay(int index, int dir, int dist){
    boolean ret = false;
    if (dist < values[index][DIST_FROM_WP]) { // kuerzerer Weg gefunden,
      values[index][DIR_FROM_WP]  = 0;    // dann andere gespeicherte Richtungen loeschen
      values[index][DIST_FROM_WP] = dist; // neue Entfernung
      ret = true;                         // Karte muss aktualisiert werden
    }
    if (dist == values[index][DIST_FROM_WP]) {  // kuerzer oder gleich -> Richtung hinzufuegen
      values[index][DIR_FROM_WP] 
        = values[index][DIR_FROM_WP] | (1 << ((6+dir)%6));       //  Maske setzen
    }
    return ret;
  }

/**
 * 
 * @author Sebastian Stober
 * @since Apr. 2002
 */
  class sortableSet extends simpleSet{
    
    private int assignedValue;  // Zugewiesene Spalte in values[][]

    sortableSet(){
      this(0, 50);
    }

    sortableSet(int valueID){
      this(valueID, 50);
    }   

    sortableSet(int valueID, int diff){
      super(diff);
      this.assignedValue = valueID;
    }

    public Object clone(){
      sortableSet o = (sortableSet) super.clone();
      o.assignedValue = this.assignedValue;
      return o;
    }

    public int getMin() throws NoSuchElementException { 
      if (this.isEmpty()) {
        return -1;
      }
      int minVal = Integer.MAX_VALUE;
      int minE = -1;
      //if (debug) System.out.println("suche neues minE in: "+this);
      for (int i = 0; i < this.size; i++) {
        if (values[content[i]][assignedValue] < minVal) {
          minE = content[i];
          minVal = values[minE][assignedValue];
          //if (debug) System.out.println("neues minE : content["+i+"] = "+minE+" value = "+minVal);
        }
      }
      return minE;
    }

    public int popMin() throws NoSuchElementException { 
      if (this.isEmpty()) {
        return -1;
      }
      int minVal = Integer.MAX_VALUE;
      int minE = -1;
      //int i;
      //if (debug) System.out.println("suche neues minE in: "+this);
      for (int i = 0; i < this.size; i++) {
        if (values[content[i]][assignedValue] < minVal) {
          minE = i;
          minVal = values[content[i]][assignedValue];
          //if (debug) System.out.println("neues minE : content["+i+"] = "+minE+" value = "+minVal);
        }
      }
      minVal = content[minE]; // Element sichern, minVal brauchen wir nicht mehr
      
      // jetzt das Element entfernen
      size--;
      while (minE < size) {
        //if (debug) System.out.println("kopiere ["+(i+1)+"] nach ["+i+"]");
        content[minE] = content[minE+1];
        minE++;
      }
      //if (debug) System.out.println("check resize.");
      // ueberpruefen, ob resize notwendig
      if (size < (content.length - diffsize)) {
        this.resize(-diffsize); // feld verkleinern
      }        

      return minVal; // gesichertes Element zurueckgeben
    }

    //damit der Hamster lieber oben langlaeuft
    public int extendedPopMin() throws NoSuchElementException { 
      if (this.isEmpty()) {
        return -1;
      }
      int minVal = Integer.MAX_VALUE;
      int minE = -1;
      int h = 0;
      //int i;
      //if (debug) System.out.println("suche neues minE in: "+this);
      for (int i = 0; i < this.size; i++) {
        if (values[content[i]][assignedValue] < minVal) {
          minE = i;
          minVal = values[content[i]][assignedValue];
          h = nodes[content[i]].getHeight();
          //if (debug) System.out.println("neues minE : content["+i+"] = "+minE+" value = "+minVal);
        }
        if (preferHeight) { // im Falle von Gleichheit wird Hoehe verglichen
          if (values[content[i]][assignedValue] == minVal) {
            // im Fall von Gleichheit entscheidet die Hoehe
            if (h < nodes[content[i]].getHeight()) {
              minE = i;
              minVal = values[content[i]][assignedValue];
              h = nodes[content[i]].getHeight();
            }
          }
        }
      }
      minVal = content[minE]; // Element sichern, minVal brauchen wir nicht mehr
      
      // jetzt das Element entfernen
      size--;
      while (minE < size) {
        //if (debug) System.out.println("kopiere ["+(i+1)+"] nach ["+i+"]");
        content[minE] = content[minE+1];
        minE++;
      }
      //if (debug) System.out.println("check resize.");
      // ueberpruefen, ob resize notwendig
      if (size < (content.length - diffsize)) {
        this.resize(-diffsize); // feld verkleinern
      }        

      return minVal; // gesichertes Element zurueckgeben
    }


    public int getMax() throws NoSuchElementException { 
      if (this.isEmpty()) {
        return -1;
      }
      int maxVal = -1;
      int maxE = -1;
      for (int i = 0; i < this.size; i++) {
        if (values[content[i]][assignedValue] > maxVal) {
          maxE = content[i];
          maxVal = values[maxE][assignedValue];
        }
      }
      return maxE;
    }

    public String printValues(){
      String s = "";
      for (int i=0; i< this.size; i++) {
        s=s+values[content[i]][this.assignedValue]+", ";
      }
      return s;
    }

  } // Ende sortableSet


  /* for debug only */
  private boolean wait            = true;
  private boolean debug           = false;
  private boolean waitForKey      = false;

  private boolean preferCliffUp           = true;
  private boolean preferCliffUpHome       = true;
  private boolean preferHeight            = true;
  private boolean advancedWayHome         = true;
  private boolean useAlternativeDropNodes = true;


  public void setPreferCliffUp(boolean b){
    preferCliffUp = b;
  }

  public void setPreferHeight(boolean b){
    preferHeight = b;
  }

  public void useAdvancedWayHome(boolean b){
    advancedWayHome = b;
  }

  public void setFindAlternativeDropNodes(boolean b, int depth, int dWay){
    useAlternativeDropNodes = b;
    FADtolerance = dWay;
    FADsearchDepth = depth;
  }

  public void setPreferCliffUpHome(boolean b){
    preferCliffUpHome = b;
  }
  
  public void continueDebug(){
    System.out.println("continue");
    wait = false;
  }                           

  private void debugWait(){
    if (waitForKey) {
      while (wait) {
        try {      
          Thread.sleep(200);
        } 
        catch (InterruptedException e) {
        }
      }
    }
  }

  private void debugWait(int time){
    if (waitForKey) {
      try {      
        Thread.sleep(time);
      } 
      catch (InterruptedException e) {
      }
    }
  }


  public void disableDebug(){
    debug = false;
  }
  
  public void enableDebug(){
    debug = true;
  }

  public void setDebugMode(boolean b){
    debug = b;
  }

  public void setDebugBreak(boolean b){
    waitForKey = b;
  }

  private void debugOutln(String s){
    //if (debug) System.out.println();
    if (debug) {
      System.out.println(s);
    }
  }

  public void printCellDebugInfo(){
    System.out.println("aktuelle Zelle: "+ currentNode);
    System.out.println(""+nodes[currentNode]);
    System.out.println("Zelle in Blickrichtung "+viewDir);
    if (nodes[currentNode].getNeighbour(viewDir) >= 0) { // Eintrag vorhanden
      System.out.println(""+nodes[nodes[currentNode].getNeighbour(viewDir)]);
    }
    else System.out.println("unknown");
  }

  public void printCellDebugInfo(int id){
    System.out.println("Zelle: "+ id 
                       +", Entfernung: "+values[id][DIST_FROM_WP]
                       +"\n Rinchtungen "+cell.boolStr(values[id][DIR_FROM_WP]));
    System.out.println(""+nodes[id]);
  }
    
  public void printCornNodes(){
    System.out.println("Gesamtzahl Felder mit Mais: "+cornNodes.getSize());
    for (int i=0; i <cornNodes.getSize(); i++) {
      printCellDebugInfo( cornNodes.get(i));
    }
  }
  
  public void printExploreNodes(){
    System.out.println("Gesamtzahl unentdeckter Felder: "+exploreNodes.getSize());
    for (int i=0; i <exploreNodes.getSize(); i++) {
      printCellDebugInfo( exploreNodes.get(i) );
    }
  }


  
  //import java.util.Vector;

// ************************* A-Stern Algorithmus fuer Transportlogistic ************************ //
  //private 
  /*
   Zustandsbeschreibung: cornNodes+currentNode
   Operationen: 
    - transportiere n Koerner von a nach b
    - gehe nach a
   Startzustand: na logo
   Endzustand: alle cornNodes=0, currentNode=0 (home)
  
   Kosten: Bewegung
   
   gegeben:
    Startzustand: ok
    Menge O: Operationen
    costs - liefert zu jedem o e O die Kosten (kann ueber O erledigt werden)
    heuristic: erstmal 0
    goal - Test auf Zielzustand - trivial
  
  
   **/


 /*
  class astarStruct{

  Vector open;
  Vector closed;



  astarStruct(){
    this.open = new Vector();
    this.closed = new Vector();
  }



  class node{
    public int id;
    public int corn;

    node(){
      this.id = -1;
      this.corn = 0;
    }

    node(int i, int c){
      this.id = i;
      this.corn = c;
    }               
  }

  // zustand:
  class state{
    Vector nodes;
    int costs;
    int hamsterNode; // nummer der aktuellen Zelle in nodes()

    state(){
      nodes = new Vector();
    }                      
  }
  

  private state transport(state s, int dest){
    extendedBSF(s.hamsterNode, 1, 
                new simpleSet[] { new simpleSet(((node) s.nodes.elementAt(dest)).id) });
    state s2 = new state();
    s2.nodes = (Vector) s.nodes.clone();
    ((node) s2.nodes.elementAt(s.hamsterNode)).corn = Math.max(0,1);

    return s2;
  }

  /*

  private void aStar(int start, int goalsMax, simpleSet[] searchGoals){
  
  /*

    if (s.hamsterNode > 0) {
      for (i = 0; i < s.nodes.size(); i++) {
        if (i != s.hamsterNode){
          s2 = transport(s,i);
          
          open.add...  
        }
      }              
    }
    
   **/  
    
   

} // Ende map
