package scrat;

import java.awt.*;       // fuer Matrix-Frame
import java.awt.event.*; // fuer Matrix-Frame
import algds.c2002.Hamster;
import algds.c2002.Hamster2;
import algds.c2002.Constants;
import java.util.*;

/**
 * 
 * @author Sebastian Stober, Florian Heidenreich
 * @since Apr. 2002
 */                                             
abstract class simpleScrat extends Hamster2 {

  protected boolean turnModeOn = true;
  
  protected map myMap;
  protected boolean keyControlActive = false;
  protected boolean finished = false;
  protected boolean wait = true;
  protected boolean debug = false;

  private boolean allowRememberDropNodesAtHome;

  private boolean allowGotoCornNeighbour;
  //private boolean allowGotoCornNeighbourHeightOK;
  //private boolean allowGotoBackNeighbourBackOK

  private boolean useWayHomeHeuristic;

  protected boolean[] configFlags = {false, true, true, true, true,
                                      true, true, true, true, true };

  private outputFrame theMatrix2; 

  // *************************** Konstruktoren ****************************************//
  protected simpleScrat(){
    // Konstruktor muss leer sein, alles in run()-Methode
  }

  /**
   * Muss als erster Befehl in run() ausgefuehrt werden, um die Karte zu initialisieren
   */
  protected void initScrat(){
    myMap = new map(getAvailableCorn());                    // Karte initialisieren
    updateMap();                                            // Karte aktualisieren

    useAdvancedWayHome(         configFlags[1]);
    setPreferCliffUp(           configFlags[2]);    
    setPreferCliffUpHome(       configFlags[3]); 
    setFindAlternativeDropNodes(configFlags[4], 3, 3);
    
    setPreferHeight(            configFlags[5]);
    
    setUseWayHomeHeuristic(     configFlags[6]);
    
    setRememberDropNodesAtHome( configFlags[7]);
  }

  protected void setPreferCliffUp(boolean b){
    myMap.setPreferCliffUp(b);
  }

  protected void setPreferCliffUpHome(boolean b){
    myMap.setPreferCliffUpHome(b);
  }

  protected void setPreferHeight(boolean b){
    myMap.setPreferHeight(b);
  }

  protected void useAdvancedWayHome(boolean b){
    myMap.useAdvancedWayHome(b);
  }

  protected void setFindAlternativeDropNodes(boolean b, int depth, int dWay){
    myMap.setFindAlternativeDropNodes(b, depth, dWay);
  }

  protected void setUseWayHomeHeuristic(boolean b){
    useWayHomeHeuristic = b;
  }

  protected void setRememberDropNodesAtHome(boolean b){
    allowRememberDropNodesAtHome = b;    
  }

  protected void setGotoCornNeighbour(boolean b){
    // allowGotoCornNeighbour = b;
    myMap.setGotoCornNeighbour(b);
  }

  protected void setGotoCornNeighbourHeightOK(boolean b){
    //allowGotoCornNeighbourHeightOK = b;
    myMap.setGotoCornNeighbourHeightOK(b);
  }

  protected void setGotoCornNeighbourBackOK(boolean b){
    //allowGotoCornNeighbourBackOK = b;
    myMap.setGotoCornNeighbourBackOK(b);
  }

  protected void setDebugText(boolean b){
    myMap.setDebugMode(b);
    
  }

  protected void setDebugBreak(boolean b){
    myMap.setDebugBreak(b);
    this.debug = b;
  }

  protected void setDistanceControl(boolean b){
    myMap.setDistanceControl(b);
  }
    
  /**
   * zum Aktualiesieren der Karte (Aufruf nur verkuerzt->uebersichtlicher)
   */
  protected void updateMap(){
    myMap.updateMap(getAvailableCorn(), lookForward(), getHeightDifference()); // Karte aktualisieren
  }

  protected boolean gotoNextCorn(){
    int[] way = myMap.getNextCorn();
    logln("gotoNextCorn() returned "+way[0]);

    // zuatz wegen dropNodes:
    if (way[0] == -2) {
      return true;
    }

    if (way[0] > 0){ // Karte auf Ziel einstellen
      goThere(way);
    }    
    else {
      return false; //komme nicht hin !!!
    } 
    return true;
  }

  protected boolean gotoNextUnexplored(){
    int[] way = myMap.getNextUnexplored();
    logln("gotoNextUnexplored() returned "+way[0]);
    if (way[0] == -2) {
      return true;
    } 
    if (way[0] > 0){ // Karte auf Ziel einstellen
       goThere(way);
       return true;
    }
    return false;
    // es gibt immer (!) einen Weg
    /*else {
      nothingToExplore = true;
    } */
  }

  private void goThere(int[] way){
    for (int i = 1; i < way.length; i++) {
      goDirection(way[i]);
      //logln(""+myMap.getDirection());
    }
  }

  protected boolean gotoCornNeighbour(){
    int dir = myMap.getDirCornNeighbour();
    //log("gotoCornNeighbour() -> "+dir);
    if (dir > 0) {
      goDirection(dir);
      //logln(" true");
      return true;
    }
    //logln(" false");
    return false;
  }

  protected boolean checkCornAvailable(){
    return myMap.checkCornAvailable();
  }

  protected boolean somethingToExplore(){
    return myMap.somethingToExplore();
  }

  protected void goHome(){
    if (!atHome()) {
      int[] way = myMap.getWayHome();
      //int i = 0;
      while (way[0] < 0){
        logln(" wir haben ein Problem: Es ist kein Heimweg bekannt");
        //finished = true; if (i == 0) break;
        gotoNextUnexplored(); // erstmal suchen
        way = myMap.getWayHome();
        //i++;if (i > 1) {finished = true;break;}
      }
      goThere(way); 
    } 
  }
  
  
  protected void debugWait(){
    wait = true;
    if (debug) {
      while (wait) {
        try {      
          Thread.sleep(200);
        } 
        catch (InterruptedException e) {
        }
      }
      wait = true;
    }
  }

  protected boolean transportWayHome(){
    if (!atHome()) {
      //debugWait();
      int carryCorn = getCarriedCorn();
      int[] way = myMap.getTransportWayHome();
      while (way[0] < 0){
        logln("wir haben ein Problem: Es ist kein Heimweg bekannt");
        if (useWayHomeHeuristic) {
          logln("verwende Heuristic");
          debugWait();
          exploreWPnexttoHome();
        }
        else {
          // Heuristic abgestellt
          gotoNextUnexplored(); // erstmal suchen
        }                  
        way = myMap.getTransportWayHome();
      }
      /* okay, jetzt kennen wir einen Weg nach Hause */
      if (way[0] == 0) { // wir sollen nach hause gehen
        goThere(way); 
        return true; // wir sind wirklich nach hause gegangen
      }
      else {
        /* jetzt wirds kompliziert, unterwegs soll der Mais abgeladen werden... */
        logln(" way.lenght = "+way.length +" way[0] = "+way[0]);
        for (int i = 1; i < way.length-1; i++) {
          logln(i+". Schritt: goDirection("+way[i]+")");
          goDirection(way[i]);
          //logln(""+myMap.getDirection());
        }
        logln("angekommen am DropPoint");
        turnDirection(way[way.length-1]);
        dropForward(CORNCARRY_MAX);
        //debugWait();
        myMap.cornDroppedForward(carryCorn, getCarriedCorn());
        /* Abfrage, ob leer!!! */
        if (getCarriedCorn() > 0) { // noch nicht alles abgelegt
          return transportWayHome();
  // fuehrte zu Stackueberlauf in 30_v_h_3d.maze -> behoben
        }
        return false; // wir sind nicht zuhause
      }
      
    }
    return true;
  }

  protected boolean rememberDropNodes(){
    if (!atHome()){
      return myMap.rememberDropNodes();
    }
    else {
      if (allowRememberDropNodesAtHome)  {
        return myMap.rememberDropNodes();
      }
      return false;
    }
    
    
  }

  protected boolean gotoNextWP(){
    int[] way = myMap.getNextWP();
    logln("gotoNextWP() returned "+way[0]);
    if (way[0] == -2) {
      return true; // bin schon da ;-)
    }
    if (way[0] > 0){ 
       goThere(way);
       return true;
    }
    return false;
  }

  protected boolean exploreWPnexttoHome(){
    int[] way = myMap.getNextWPHome();
    logln("gotoNextWPHome() returned "+way[0]);
    if (way[0] >= 0){ 
       goThere(way);
       return true;
    }
    return false;
  }

  protected boolean takeAll(){
    take(CORNCARRY_MAX);
    myMap.updateCorn(getAvailableCorn());
    return (getCarriedCorn() == CORNCARRY_MAX);
  }

  protected void dropAll(){
    logln("dropping corn...");
    if (!atHome()) {
      logln("NOT AT HOME !!!");
    }
    dropHere(CORNCARRY_MAX);
    myMap.updateCorn(getAvailableCorn());
  }

  // *************************** Bewegungsmethoden ****************************************//
  /**
   * gehe vorwaerts und aktualisiere Koordinaten und Karte
   */
  public void goForward(){
    if (myMap.canGoForward()) { // Bedingung abtesten
      //logln("goForward() okay");
      forward();                              // Hamster-methode
      myMap.goForward();
      updateMap();                            // Karte aktualisieren
      if (turnModeOn) {                       // kann ueber globale Variable abgeschaltet werden
        if (!myMap.currentCellExplored()) {
          //logln("unexplored -> turn");
          for (int i = 0; i < 6; i++) {
            turnPos();                        // drehen, um andere Felder zu entdecken
          }
        }     
      }
    } 
    //myMap.printCellDebugInfo();
  }
  
  /**
   * gehe in Richtung dir
   */
  public void goDirection(int dir){
    turnDirection(dir);
    goForward();
  }

  /**
   * drehe im Richtung dir
   */
  public void turnDirection(int dir){          
    //logln("turnDirection("+dir+")");
    
    dir %= 6; // sicherheitshalber
    if (((myMap.getDirection()+3)%6) <= dir) 
        while (myMap.getDirection() != dir) turnNeg();
    else while (myMap.getDirection() != dir) turnPos();
  }
  
  /**
   * drehe im Uhrzeigersinn und aktualisiere Karte
   */
  public void turnPos(){
    turn(TURN_RIGHT);
    myMap.turnPos();
    updateMap();
  }

  /**
   * drehe gegen den Uhrzeigersinn und aktualisiere Karte
   */
  public void turnNeg(){
    turn(TURN_LEFT);
    myMap.turnNeg();
    updateMap();   
  }

  /** --------------------------- outputFrame ------------------------------ **/

  /**
   * schaltet ZusatzFrame mit KeyControls ein/aus
   * 
   * @param b      true - manuelle Steuerung, false - Automatik
   */
  public void setKeyControls(boolean b){
    if(b != keyControlActive){              // hat sich was geaendert ?
      keyControlActive = b;
      if (b) {
        theMatrix2 = new outputFrame();  // anschalten
      }
      else {
        theMatrix2.exit();               // ausschalten
      }
    }                              
  }




  /**
   * dient in erster Linie zum Debuggun
   *
   * @author Florian Heidenreich
   * @since Apr. 2002
   */

  class outputFrame extends Frame{ 

      //private Image dbImage;
      //private Graphics gr;

      outputFrame(){
          setTitle("-=[ThE MaTRiX 2]=-");
          setSize(200,200);
          setBackground(Color.lightGray);
          setLayout(new BorderLayout());
          setResizable(false);

          addWindowListener(new WindowsL());

          keys al=new keys();
          this.addKeyListener(al);
          
          setVisible(true);

          //dbImage = createImage(300,300);          
          //gr = dbImage.getGraphics();
      }
      public void exit(){
          this.dispose();
      }
      class WindowsL extends WindowAdapter {
          public void windowClosing(WindowEvent event) {
              exit();
          }
      }
      class keys extends KeyAdapter {  
        public void keyPressed(KeyEvent e) {
          int n=e.getKeyCode();
          switch(n) {   
          case KeyEvent.VK_NUMPAD0: {myMap.printCellDebugInfo();break;}
          case KeyEvent.VK_NUMPAD8: {goDirection(1);break;} 
          case KeyEvent.VK_NUMPAD6: {goDirection(2);break;}
          case KeyEvent.VK_NUMPAD3: {goDirection(3);break;}
          case KeyEvent.VK_NUMPAD2: {goDirection(4);break;}
          case KeyEvent.VK_NUMPAD1: {goDirection(5);break;}
          case KeyEvent.VK_NUMPAD4: {goDirection(6);break;}
          case KeyEvent.VK_LEFT: {turnNeg();break;}
          case KeyEvent.VK_RIGHT: {turnPos();break;}
          case KeyEvent.VK_UP: {goForward();break;}
          case KeyEvent.VK_C: {myMap.printCornNodes();break;}
          case KeyEvent.VK_E: {myMap.printExploreNodes();break;}
          case KeyEvent.VK_B: {myMap.bfsFullMap(); break;}
          case KeyEvent.VK_U: {gotoNextUnexplored(); break;}
          case KeyEvent.VK_I: {gotoNextCorn();break;}
          case KeyEvent.VK_H: {goHome(); break;}
          case KeyEvent.VK_F: {finished = true; break;}
          case KeyEvent.VK_T: {takeAll(); break;}
          case KeyEvent.VK_D: {dropAll(); break;}
          case KeyEvent.VK_N: {log("n");wait=false; myMap.continueDebug();break;}
          case KeyEvent.VK_Q: {debug=false; myMap.disableDebug(); break;}
          case KeyEvent.VK_Y: {transportWayHome(); break;}
          case KeyEvent.VK_L: {exploreWPnexttoHome(); break;}
          }
        }
      } // ENDE class keys
  }/** --------------------------- ENDE outputFrame ------------------------------ **/
  /** ---------------------- ENDE Visualisierung und Debug ---------------------- **/
}

