//****************************************************************************
//
//					Hamster TICO (urspr. TICONDEROGA)
//
//	 ein Beitrag zum Hamsterwettbewerb 2002 der Fakultt fr Informatik
//		       der Otto-von-Guericke-Universitt Magdeburg
//
//			 entwickelt und programmiert von:  Andreas Strehl
//
//****************************************************************************

//****************************************************************************
// zugehrige Klassen: Tico.class, Position.class, Feld.class 
//
// Funktionsaufbau dieser Klasse und das umgesetztes Suchprinzip:
// 
// Die hier vorliegende Hamsterklasse steuert den Hamster autonom durch ein
// vorher nicht bekanntes flaches Labyrinth.
// Der Hamster beginnt seine Suche mit dem Anlegen einer internen Karte,
// auf der Informationen, wie z.B. der Rckweg, das auf dem jeweiligen Feld
// verfgbare Korn usw. gespeichert werden. Diese dienen der Information, z.B. 
// ob das Feld schon einmal betreten wurde, in welche Richtung die Rckkehr
// zu erfolgen hat, wie weit das Feld vom Heimatfeld entfernt liegt und in 
// welche Richtung eine verkrzte Heimkehr erfolgen kann.
// Der Hamster beginnt seine Suche mit einem Schritt vorwrts. Sollte dies 
// nicht mglich sein (z.B. aufgrund einer Mauer) dreht er sich im Uhrzeiger-
// sinn und prft bei jedem vorausliegendem Feld, ob sich dieses betreten lt
// und ob dieses Feld schon einmal betreten wurde.
// Sollte dies nicht der Fall sein, beginnt der Hamster rekursiv mit einer Er-
// kundung dieses Feldes und kehrt zum Ausgangsfeld zurck, sobald alle
// anliegenden und erreichbaren Felder erkundet sind.
// 
//
// Prinzip des Kornsammelns:
//
// in der internen Karte wird fr jedes bereits betretende Feld die Entfernung
// (ber das Feld, nicht Luftlinie!) zum Heimatfeld gespeichert. Entfernt sich
// der Hamster bei seiner Suche von dem Heimatfeld legt er zunchst alle im
// Maul tragende Maiskrner ab und nimmt diese auf dem Rckweg mit. Sollten
// sich auf dem Feld schon mehr als 20 Maiskrner befinden, trgt er vor der 
// weiteren Suche zunchst die maximale tragbare Anzahl an Maiskrnern auf
// einem ber die interne Karte berechneten verkrzten Weg zurck zum Heimat-
// feld. 
//
// Navigation:
// Die Navigation erfolg auf einer internen zweidimensionalen Karte, die
// schrg auf das Labyrinth gelegt wird. hnlich dem Prinzip einer Mauer
// ergibt sich dabei, dass eine Feldreihe zur darunterliegenden um ein halbes
// Feld nach rechts verschoben ist. Durch die folgenden Berechnungen lt sich
// jedem Feld eine eindeutige Identifikation ber die Koordinaten zuordnen:
//
// interne
// Richtung	 			    		    X-Koordinate     Y-Koordinate
// 	 (0)	 Schritt nach oben:              +1				  +1
// 	 (1)	 Schritt nach links oben:        +2           unverndert
// 	 (2)	 Schritt nach links unten:       +1               -1
// 	 (3)	 Schritt nach unten:			 -1 			  -1
// 	 (4)	 Schritt nach rechts unten:		 -2			  unverndert
// 	 (5)	 Schritt nach rechts oben:		 -1				  +1
//****************************************************************************

package tico;

import algds.c2002.Hamster;
import java.util.*; 

public class Tico extends Hamster{
	
	//Deklaration der Klassenvariablen
	private final int  heimatX=60;  //definiert die xKoordinate des HeimatFeldes in der internen Karte
  	private final int  heimatY=60;	//definiert die yKoordinate des HeimatFeldes in der internen Karte
  	private Position aktuellePosition=new Position(heimatX,heimatY); //definiert die Navigationsvariable
	private Feld karte[][]=new Feld[(heimatX*2)][(heimatY*2)]; //Array fr die interne Karte
	private int blickrichtung=0;	//Variable fr die interne Richtungsbestimmung
  	private int entfernungHeimat=0; //Variable fr die Entfernung des Hamsters vom Heimatfeld
  	
  	
  	//Hauptmethode "run"
  	public void run(){
  		try{
  		karteAnlegen(); 	//Vorbereitung der internen Karte
  		feldErkunden();		//Erkundet das Heimatfeld
  		kornSuchen(true);   //beginnt mit der rekursiven Suche
  		}
  		catch(Exception e){ //hoffentlich nicht auftretende Exception auffangen
  		logln("Es ist ein mysteriser Fehler aufgetreten...");
  		logln("Beende Suche und kehre zum Heimatfeld zurck.");
  		}
  		heimkehren(); 		//kehrt zum Heimatfeld zurck
  		logln("fertig!");
  	}
  	
  	//Methoden fr Aktionsfolgen des Hamsters
  	
  	//************************************************************************
  	// Methode "kornSuchen"
  	// steuert die Kornsuche. Dabei wird zwecks Felderkundung die unten
  	// beschriebene Methode "feldErkunden()" aufgerufen. 
  	// Desweiteren werden im Uhrzeigersinn - sofern mglich - die voraus-
  	// liegenden Felder betreten, sollte das zu betretende Feld weiter von
  	// dem Heimatfeld entfernt liegen, werden alle im Maul getragenden Mais-
  	// krner abgelegt und auf dem Rckweg mitgenommen.	
  	//************************************************************************
  	private void kornSuchen(boolean vonStartfeld){
  		if (vonStartfeld==false) 
  			vorwaerts();
  		feldErkunden();
  		int eintrittsrichtung=blickrichtung;
  		for (int richtung=0;richtung<=5;richtung++){
  			inRichtungDrehen(eintrittsrichtung+richtung);
  			if (blickrichtung!=karte[aktuellePosition.x][aktuellePosition.y].getWegHeim()){
  				if (lookForward()!=2){
  					if (karte[xKoordinateVorausliegendesFeld()][yKoordinateVorausliegendesFeld()].getBereitsBetreten()==false){	
  						kornAblegen();
  						kornSuchen(false);
		  			}
  				}
  			}
  		}
  		kornHeimschaffen();
  		inRichtungDrehen(karte[aktuellePosition.x][aktuellePosition.y].getWegHeim());
  		vorwaerts();
  	}
  	
  	//************************************************************************
  	// Methode "kornHeimschaffen"
  	// prft zunchst, ob sich der Hamster auf dem Heimatfeld befindet. Sollte
  	// dies nicht der Fall sein, wird geprft, ob sich mehr als 20 Krner auf
  	// dem Feld befinden. Sollte dies der Fall sein, werden die Maiskrner bis 
  	// auf 20 verbleibende auf dem krzesten errechneten Weg zum Heimatfeld
  	// geschafft, ansonsten werden die Maiskrner in das Maul aufgenommen.
  	//************************************************************************
  	private void kornHeimschaffen(){
  		dropHere(getCarriedCorn());
  		while ((lookHere()==1) && (atHome()==false)){
  			if (getAvailableCorn()>20){
  				take(20-getCarriedCorn());
  				aufKuerzestemWegHeimschaffen();
  			}
  			else{
  				take(getAvailableCorn());
  			}
  		}
  	}	
  	
  	//************************************************************************
  	// Methode "kornAblegen"
  	// prft zunchst, ob sich der Hamster auf dem Heimatfeld befindet. Sollte
  	// dies der Fall sein, wird das Mais abgelegt und die Methode beendet.
  	// Anderenfalls wird geprft, ob sich mit dem im Maul getragenen und nun
  	// abzulegendem Mais mehr als hundert Maiskrner auf dem Feld befinden 
  	// wrden. Um zu verhindern, dass sich der Hamster dann in einer Endlos-
  	// schleife verfangen wrde, wird in diesem Fall die Methode
  	// "aufKuerzestemWegHeimschaffen()" aufgerufen, andernfalls das Korn
  	// abgelegt.
  	//************************************************************************
  	private void kornAblegen(){
  		if (atHome()){
  			dropHere(getCarriedCorn());
  		}
  		else if (getAvailableCorn()+getCarriedCorn()>100)
  				aufKuerzestemWegHeimschaffen();
  			 else
  			 	dropHere(getCarriedCorn());
  	}
  	
  	//************************************************************************  	
  	// Methode "aufKuerzestemWegHeimschaffen()"
  	// steuert den Hamster mit den in der internen Karte gespeicherten Werten
  	// fr eine verkrzte Heimkehr zum Heimatfeld, legt das Korn dort ab und
  	// steuert den Hamster anschlieend zum Ausgangsfeld zurck
  	//************************************************************************
  	private void aufKuerzestemWegHeimschaffen(){
		logln("Schaffe Korn auf kuerzestem bekannten Weg heim...");
		if (atHome()!=true){
			int anfaenglicheBlickrichtung=blickrichtung;
  			int anfaenglichesX=aktuellePosition.x;
  			int anfaenglichesY=aktuellePosition.y;
  			while ((aktuellePosition.x!=heimatX) || (aktuellePosition.y!=heimatY)){
  				inRichtungDrehen(karte[aktuellePosition.x][aktuellePosition.y].getKuerzesterKorntransportRichtung());
  				vorwaerts();
  				karte[aktuellePosition.x][aktuellePosition.y].setRueckweg(umgekehrteRichtungErmitteln());
  			}
  			kornAblegen();
  			log("Kehre zum Feld zurck...");
  			while ((aktuellePosition.x!=anfaenglichesX) || (aktuellePosition.y!=anfaenglichesY)){
  				inRichtungDrehen(karte[aktuellePosition.x][aktuellePosition.y].getRueckweg());
  				vorwaerts();
  			}
  			logln("angekommen!");
  			inRichtungDrehen(anfaenglicheBlickrichtung);
  			logln("Setze Suche fort...");
  		}
  		else logln("Mhh, lege das einfach mal hier ab...");
  	}
  	
  	//************************************************************************
  	// Methode "heimkehren()"
  	// steuert den Hamster mit den in der internen Karte gespeicherten
  	// Informationen zurck zum Heimatfeld
  	//************************************************************************
  	private void heimkehren(){
  		while (atHome()!=true){
  			inRichtungDrehen(karte[aktuellePosition.x][aktuellePosition.y].getWegHeim());
  			vorwaerts();
  		}
  		kornAblegen();
  	}
  	
	//Methoden fr Aktionen des Hamsters
	
	//************************************************************************
	// Methode "vorwaerts()"
	// verhindert, da der Hamster gegen eine Wand luft und dadurch Kollisionen
	// verursacht und setzt ferner die Werte fr das Feld in der internen Karte
  	//************************************************************************
	private void vorwaerts(){
  		if ((lookForward()==0) || (lookForward()==1)){
  			forward();
  			aktuellePosition.vorwaerts(blickrichtung);
  			if (karte[aktuellePosition.x][aktuellePosition.y].getBereitsBetreten()==false){
  				karte[aktuellePosition.x][aktuellePosition.y].setWegHeim(umgekehrteRichtungErmitteln());
				karte[aktuellePosition.x][aktuellePosition.y].setEntfernungHeimatfeld(++entfernungHeimat);
				karte[aktuellePosition.x][aktuellePosition.y].setBereitsBetreten(true);  			
  			}
  			else{
  				entfernungHeimat=karte[aktuellePosition.x][aktuellePosition.y].getEntfernungHeimatfeld();
  			}
  		}
  		else warnung("Hamster wollte vor die Wand laufen");
  	}
  	
  	//************************************************************************
  	// Methode "drehen(int drehwert)"
  	// dreht den Hamster unter Beachtung des o.a. Richtungsprinzips
  	//************************************************************************
  	private void drehen(int drehwert){
  		blickrichtung=blickrichtung+drehwert;
  		if (blickrichtung>5){
  			blickrichtung=blickrichtung-6;
  		}
  		if (blickrichtung<0){
  			blickrichtung=blickrichtung+6;
  		}
  		turn(drehwert);
  	}
  	
  	//************************************************************************
  	// Methode "feldErkunden()"
  	// setzt in der internen Karte die Werte fr die Entfernung vom Heimatfeld,
  	// die Richtung, von der das Feld betreten wurde, ferner fr die krzeste
  	// Entfernung vom Heimatfeld und die einzuschlagende Richtung, sowie ob das
  	// Feld bereits erkundet wurde.
  	//************************************************************************
  	private void feldErkunden(){
  		int i=0;
  		int betreteneRichtung=umgekehrteRichtungErmitteln();
  		karte[aktuellePosition.x][aktuellePosition.y].setKuerzesterKorntransportEntfernung(entfernungHeimat);
  		karte[aktuellePosition.x][aktuellePosition.y].setKuerzesterKorntransportRichtung(betreteneRichtung);
  		for (int a=0; a<6; a++){
  			if (lookForward()!=2){
  				if (karte[xKoordinateVorausliegendesFeld()][yKoordinateVorausliegendesFeld()].getAllesErkundet()==true){
  					i=karte[xKoordinateVorausliegendesFeld()][yKoordinateVorausliegendesFeld()].getKuerzesterKorntransportEntfernung();
  					if (entfernungHeimat>(i+1)){
  						karte[aktuellePosition.x][aktuellePosition.y].setKuerzesterKorntransportEntfernung(i+1);
  						karte[aktuellePosition.x][aktuellePosition.y].setKuerzesterKorntransportRichtung(blickrichtung);
  					}
  				}
  			}
  			blickrichtung=blickrichtung+1;
  			if (blickrichtung>5){
  				blickrichtung=blickrichtung-6;
  				}
  			turn(1);
  		}
  		karte[aktuellePosition.x][aktuellePosition.y].setEntfernungHeimatfeld(entfernungHeimat);
  		karte[aktuellePosition.x][aktuellePosition.y].setAllesErkundet(true);	
  	}
	
	//************************************************************************  	
  	// Methode "inRichtungDrehen(int wert)"
  	// dreht den Hamster in die angegebene Richtung - codiert nach dem o.a.
  	// Prinzip
  	//************************************************************************
  	private void inRichtungDrehen(int wert){
  		while (blickrichtung!=wert){
  			if (wert>5){
  				wert=wert-6;
  			}
  			if (wert<0){
  				wert=wert+6;
  			}
  			drehen(1);
  		}
  	}
  		
  	//************************************************************************
  	// Methode "umgekehrteRichtungErmitteln()"
  	// errechnet nach dem o.a. Prinzip die umgekehrte Richtung
  	//************************************************************************
  	private int umgekehrteRichtungErmitteln(){
  		int wert=0;
  		wert=blickrichtung-3;
  		if (wert<0){
  			wert=wert+6;
  		}
  		return wert;
  	}
  	
  	//************************************************************************
  	// Methode "inUmgekehrteRichtungDrehen()"
  	// dreht den Hamster um 180 Grad
  	//************************************************************************
  	private void inUmgekehrteRichtungDrehen(){
  		inRichtungDrehen(umgekehrteRichtungErmitteln());
  	}
  	
  	//************************************************************************
  	// Navigationsmethoden
  	// berechnen nach dem o.a. Navigationsprinzip die Koordinaten des voraus-
  	// liegenden Feldes
  	//************************************************************************
  	private int xKoordinateVorausliegendesFeld(){
  		if (blickrichtung==0){
  	   		return (aktuellePosition.x+1);
  	   	}
  		else if (blickrichtung==1){
  	   			return (aktuellePosition.x+2);
  	   		}
  	   		else if (blickrichtung==2){
  	   				return (aktuellePosition.x+1);
  	   			}
  	   			else if (blickrichtung==3){
  	   					return (aktuellePosition.x-1);
  	   				}
  	   				else if (blickrichtung==4){
  	   						return (aktuellePosition.x-2);
  	   					}
  	   					else if (blickrichtung==5){
  	   							return (aktuellePosition.x-1);
  	   						}
  	   						else System.out.println("Keine Positionsnderung!");
  	   return -1;
  	}
  	
  	private int yKoordinateVorausliegendesFeld(){
  		if (blickrichtung==0){
  	   		return (aktuellePosition.y+1);
  	   	}
  		else if (blickrichtung==1){
  				return (aktuellePosition.y);
  	   		}
  	   		else if (blickrichtung==2){
  	   			return (aktuellePosition.y-1);	
  	   			}
  	   			else if (blickrichtung==3){
  	   					return (aktuellePosition.y-1);
  	   				}
  	   				else if (blickrichtung==4){
  	   						return (aktuellePosition.y);
  	   					}
  	   					else if (blickrichtung==5){
  	   							return (aktuellePosition.y+1);
  	   						}
  	   						else System.out.println("Keine Positionsnderung!");
  	   return -1;
  	}
  	
  	//Methode zum Anlegen der internen Karte
  	private void karteAnlegen(){
  		for (int laufx=0;laufx<(heimatX*2);laufx++){
  			for (int laufy=0;laufy<(heimatY*2);laufy++){
  				karte[laufx][laufy]=new Feld(laufx, laufy);
  			}
  		}
  		//Setzen der Werte fr das Startfeld
  		karte[30][30].setBereitsBetreten(true);
  	}
  	
  	//********************************************
  	//Hilfsmethoden zum Entwickeln des Hamsters
  	//********************************************
  	
  	private void warnung(String text){
  		System.out.println(" !!!  Warnung: "+text+" !!!");
  		logln("Warnung: "+text+"!!!");
  	}
  	
	//Informtationsmethoden - ich denke, die mssen nicht kommentiert werden...
  	public String getHamsterAuthor() {
  		return "Andreas Strehl - Otto-von-Guericke-Universitt Magdeburg";
  	}
  	
  	public String getHamsterVersion() {
  		return "1.0";
  	}
  	
  	public String getHamsterDate() {
    	return "26.04.2002";
	}
}

