package warhamster;
import algds.c2002.*;
import java.awt.Point;
/**
Kapselung eines hexfelds. Jedes hexfeld besitzt eindeutige koordinaten. Das heimatfeld hat die koordinaten
(0,0) waehrend alle anderen koordinaten sich relativ dazu berechnen. Felder die in richtung Norden/Sueden liegen
unterscheiden sich um +/-2 von den direkt darunter/darueberliegenden. Eine schraeger Schritt bedeutet eine
abweichung von +/-1. Norden=Oben, Osten=Links, Sueden=Unten, Westen=Rechts. Die felder werden mit ihren
koordinaten als schluessel in einem hash gespeichert, und dieser hash wird statisch fuer die klasse
Hexfield gespeichert. Somit hat man ueber jedes hexfield zugriff auf die gesamte Karte. Hexfield stellt
diverse routinen zur erkundung und anderen aufgaben zur verfuegung, und wird damit zu einer art
"intelligenter karte" die von mehrerern hamstern benutzt werden kann.<br>
Der sich auf ihr bewegende hamster wird statisch fuer die klasse Hexfield gespeichert, wodurch jedes feld
weiss, welcher hamster sich auf ihm befindet, und ihm "sagen kann, was er zu tun hat" (seine methoden
benutzen kann).
*/
public class Hexfield
implements Constants, Hexfield_Constants {
    //**********************Constructors**********************
    /**Constructor fuer das heimatfeld. Setzt Koordinaten auf (0,0) setzt relative hoehe auf null,
    setzt das feld auf "bekannt", setzt heimpfadlaenge auf 0 und ordnet Heimatfeld in Plan - hash ein.
    <br>Unteraufrufe:<ul>
    <li>Hashtable.put() - methode die HexfieldHash von Hashtable geerbt hat</li>
    </ul>
    */
    Hexfield(){
        //Home = true;
        Position = new Hexpoint(0,0);
        //System.out.println("Constructing : "+this);
        Height=0;
        Known=true;
        Pathlength=0;
        Map.put(Position,this);
        //System.out.println("Constructed  : "+this);
    }
    /**
    Constructor fuer alle nicht-Heimatfelder. Berechnet aus koordinaten von Origin und der gegangen
    richtung Direction die neuen koordinaten und ordnet das konstruierte feld in Plan-hash ein.
    Aufgrund der symmetrie des labyrinths kann davon ausgegangen werden, das ebenfalls alle felder, die
    im selben umkreis um das heimatfeld liegen, existieren. Diese werden rekursiv ebenfalls kontruiert.
    Gestoppt wird, wenn ein feld kontruiert wurde, das bereits im hash existiert. Dieses wird verworfen.
    <br>Unteraufrufe:<ul>
    <li>Hashtable.put() - methode die HexfieldHash von Hashtable geerbt hat</li>
    <li>{@link #Hexfield(Hexfield Origin, byte Direction)} - rekursive konstruktion des naechsten
        Felds auf demselben umkreis</li>
    </ul>
    */
    Hexfield(Hexfield Origin,byte Direction){
        Position = new Hexpoint(Origin.Position,Direction);
        //System.out.println("Constructing : "+this);
        if (Map.get(Position)!=null) return;
        Map.put(Position,this);
        for (byte i=0;i<6;i++)
            Border[i]=(get(i)==null)?(0):(get(i).invertBorder(Border[invert(i)]));
        if (Position.x == 0)
            if (Position.y<0) new Hexfield(this,NWEST);
            else new Hexfield(this,SEAST);
        else if ((Position.x==Position.y)&&(Position.x>0))
            new Hexfield(this,SOUTH);
        else if ((Position.x==Position.y)&&(Position.x<0))
            new Hexfield(this,NORTH);
        else if ((Position.x==-Position.y)&&(Position.x>0))
            new Hexfield(this,SWEST);
        else if ((Position.x==-Position.y)&&(Position.x<0))
            new Hexfield(this,NEAST);
        else if (Position.x>0)
            if (Position.y>Position.x)
                new Hexfield(this,SEAST);
            else if (-Position.y<Position.x) new Hexfield(this,SOUTH);
            else new Hexfield(this, SWEST);
        else if (Position.y<Position.x) new Hexfield(this,NWEST);
        else if (-Position.y>Position.x) new Hexfield(this,NORTH);
        else new Hexfield(this,NEAST);
        //System.out.println("Constructed  : "+this);
    }    
    //********************MemberFunctions*********************
    /**Unbenutzt.
    Gibt den umkreis zurueck, auf dem sich this befindet.
    <br>Unteraufrufe:<ul>
    <li>{@link #distance(Hexfield End)} - umkreis des hexfelds ist definiert durch entfernung zum Heimatfeld</li>
    </ul>
    */
    public int      circle(){
        return (Position.equals(new Point()))?(0):(distance(Home));
    }
    /**Gibt zurueck ob sich Korn auf dem feld befindet.*/
    public boolean  corn(){
        return !(Corn==0);
    }
    /**Unbenutzt. Berechnet distanz zu End in luftlinienschritten.*/
    public int      distance(Hexfield End){
        return (Math.abs(Position.x-End.Position.x)>Math.abs(Position.y-End.Position.y))?
                (Math.abs(Position.x-End.Position.x)):
                ((Math.abs(Position.x-End.Position.x)+Math.abs(Position.y-End.Position.y))/2);
    }
    /**Unbenutzt. Berechnet distanz zu feld, das zu punkt End gehoert, in luftlinienschritten.*/
    public int      distance(Point End){
        return (Math.abs(Position.x-End.x)>Math.abs(Position.y-End.y))?
                (Math.abs(Position.x-End.x)):
                ((Math.abs(Position.x-End.x)+Math.abs(Position.y-End.y))/2);
    }
    
    /**
    Erkundet feld mit hilfe des sich auf ihm bewegenden hamsters. Werte die gesetzt werden sind:
    Hits, Borders, Corn, angrenzende Sichtbare fleder, Visited,  Value und Homepath. Ist this das Heimatfeld,
    wird alles korn abgeworfen, und Hamster.goHome auf false gesetzt. Andernfalls wird soviel korn wie
    moeglich aufgehoben, und kornanzahl auf dem feld geupdatet. Hat hamster die backen voll wird er nach
    hause geschickt. Wurde das feld bereits einmal besucht wird der Homepath neu berechnet, das feld
    bewerted, und geschaut ob der Hamster stoppen sollte. Dann wird returned.<br>
    Wurde das feld noch nie besucht, wird die Visited-flagge gesetzt, alle angrenzenden felder und grenzen
    angeschaut und die sichtbaren auf known gesetzt und bewerted, sowie deren relative hoehendifferenz gespeichert.
    Anschließend wird das aktuelle feld bewerted und sein heimatpfad neu berechnet, sowie gecheckt,
    ob gestoppt werden sollte.<br>
    returns - this
    <br>Unteraufrufe:<ul>
    <li>{@link #setHomepath()} - berechnet Heimatpfad neu</li>
    <li>{@link #value()} - bewerted aktuelles feld</li>
    <li>{@link warhamster.Warhamster#turn(int Direction)} - dreht Hamster zum erkunden aller felder</li>
    </ul>
    */
    public Hexfield explore(){
        //if (Soldier.verbose()) Soldier.logln("Exploring: "+this);
        Hits++;
        if (Soldier.atHome()){
            Soldier.dropHere(Soldier.getCarriedCorn());
            Soldier.GoHome=false;
            Value=HOMEVALUE;
        }
        else{
            if (Soldier.take(CORNCARRY_MAX)!=0) Soldier.Steps=0;
            Corn = Soldier.getAvailableCorn();
        }
        if (Soldier.getCarriedCorn()==CORNCARRY_MAX) Soldier.GoHome=true;
        if (Visited){
            setHomepath();
            value();
            if (Soldier.Steps+Pathlength>199) Soldier.Stop=true;
            return this;
        }
        Visited=true;
        if (Soldier.getAvailableCorn()==0) Value=NULLVALUE;
        for (byte i=0;i<6;i++){
            Soldier.turn(TURN_RIGHT);
            Border[Soldier.Looking]=(Soldier.lookForward()&WALLMASK);
            if (Border[Soldier.Looking]==WALL) continue;
            if (get(Soldier.Looking)==null) new Hexfield(this,Soldier.Looking);
            if (get(Soldier.Looking).Known) continue;
            if ((Border[Soldier.Looking]==SLOPEUP)||
                (Border[Soldier.Looking]==CLIFFUP))
                get(Soldier.Looking).Height=Height+Soldier.getHeightDifference();
            else{
                get(Soldier.Looking).Height=Height+Soldier.getHeightDifference();
                get(Soldier.Looking).Corn= -(Soldier.lookForward() & CORNMASK);
                get(Soldier.Looking).Known=true;
                if (get(Soldier.Looking).Corn==-CORN) Value=CORNVALUE; 
            }
        }
        value();
        setHomepath();
        if (Soldier.Steps+Pathlength>199) Soldier.Stop=true;
        //if (Soldier.verbose()) Soldier.logln("Explored : "+this);
        return this;
    }
    /**Routine um Hexfeld mit argument als Koordinaten zu suchen.
    <br>Unteraufrufe:<ul>
    <li>{@link warhamster.HexfieldHash#get(Point Key)} - liefert Feld, dass Point als koordinaten und
         damit als schluessel hat</li>
    </ul>*/
    public Hexfield get(Point ToFind){
        return Map.get(ToFind);
    }
    /**Routine, um Hexfeld mit Koordinaten (x,y) zu suchen.
    <br>Unteraufrufe:<ul>
    <li>{@link warhamster.HexfieldHash#get(int x,int y)} - liefert feld, welches koordinaten (x,y) hat</li>
    </ul>*/
    public Hexfield get(int x,int y){
        return Map.get(x,y);
    }
    /**Routine um hexfeld zu suchen, welches vom aktuellen aus einen schritt in richtung Direction liegt.
    <br>Unteraufrufe:<ul>
    <li>{@link #get(Point)} - neu konstruierter Hexpoint wird benutzt um feld zu erhalten</li>
    </ul>
    */
    public Hexfield get(byte Direction){
        return get(new Hexpoint(Position,Direction));
    }
    /**Invertiert Direction. Aus NORTH wird SOUTH, etc.*/
    public byte    invert(byte Direction){
        switch(Direction){
        case 0: return 3;
        case 1: return 4;
        case 2: return 5;
        case 3: return 0;
        case 4: return 1;
        case 5: return 2;
        }
        return -1;
    }
    /**Invertiert Walltype. Aus CLIFFUP wird CLIFFDOWN, etc. WALL bleibt WALL.*/
    public int     invertBorder(int Walltype){
        if (Walltype==WALL) return WALL;
        if (Walltype==CLIFFUP) return CLIFFDOWN;
        if (Walltype==CLIFFDOWN) return CLIFFUP;
        if (Walltype==SLOPEUP) return SLOPEDOWN;
        if (Walltype==SLOPEDOWN) return SLOPEUP;
        return 0;
    }
    /**Returned Hexfeld in Nordostrichtung.*/
    public Hexfield NEast(){
        return Map.get(new Hexpoint(Position,NEAST));
    }

    /**Returned Hexfeld in Nordrichtumg.*/
    public Hexfield North(){
        return Map.get(new Hexpoint(Position,NORTH));
    }
    /**Returned Hexfeld in Nordwestrichtung.*/
    public Hexfield NWest(){
        return Map.get(new Hexpoint(Position,NWEST));
    }
    
    /**Returned Hexfeld in Suedostrichtung.*/
    public Hexfield SEast(){
        return Map.get(new Hexpoint(Position,SEAST));
    }
    /**Returned Hexfeld in Suedrichtung.*/
    public Hexfield South(){
        return Map.get(new Hexpoint(Position,SOUTH));
    }
    /**Returned Hexfeld in Suedwestrichtung.*/
    public Hexfield SWest(){
        return Map.get(new Hexpoint(Position,SWEST));
    }

    /**Returned Stringrepraesentation des this-Hexfeldes.*/
    public String   toString(){
        return "Hexfield("+Position.x+","+Position.y+",C"+circle()+","+Value+","+Visited+")";
    }

    /**Validiert heimatpfad des aktuellen feldes und setzt ihn wenn noetig neu.
    Es werden alle heimatpfadlaengen der angrenzenden besuchten felder gecheckt und die kuerzeste
    gespeichert (w). Ist pfadlaenge des aktuellen feldes groesser als w+1, so wird sie neu berechnet
    und {@link #Homepath} auf ein besseres, angrenzendes feld gesetzt, und die heimatpfade aller
    angrezenden felder werden neu berechnet. Ist die aktuelle Pfadlange gleich w+1 passiert nichts.
    Das ist auch die abbruchbedingung fuer die rekursion.
    <br>Unteraufrufe:<ul>
    <li>{@link #get(byte Direction)} - wird benutzt um angrenzendes feld zu erhalten</li>
    <li>{@link #setHomepath()} - rekursive neuberechnung der heimatpfade angrenzender felder</li>
    </ul>
    */
    public int      setHomepath(){
        if (this==Home) return 0;
        int Prepath=Pathlength;
        for (byte i=0;i<6;i++){
            if ((get(i)==null)||
                (Border[i]==WALL)||
                (Border[i]==CLIFFUP)||
                (!get(i).Visited)) continue;
            if (get(i).Pathlength+1<Pathlength){
                Homepath=get(i);Pathlength=get(i).Pathlength+1;
            }
        }
        if (Prepath<Pathlength) throw new Error("Homepath set wrong!");
        if (Prepath!=Pathlength)
            for (byte i=0;i<6;i++)
                if ((get(i)==null)||
                    (Border[i]==WALL)||
                    (Border[i]==CLIFFUP)||
                    (!get(i).Visited)) continue;
                else get(i).setHomepath();
        return Pathlength;
    }
    /**Bewerted das aktuelle feld.
    Die bewertungsfunktion ist kernstueck des hamsterverhaltens, und kann beliebig gestaltet werde,
    um das hamsterverhalten zu aendern. Alles was beachtet werden muss, ist das je groesser der wert,
    desto besser das feld.<br>
    Diese routine bewerted das aktuelle feld mit dem durchschnitt der werte der angrenzenden felder,
    addiert mit dem 10fachen des auf dem aktuellen feld liegenden korns. Nicht existente felder werden
    in die durchschnittsberechnung nicht mit aufgenommen, felder hinter einer mauer oder klippe erhalten
    einen spezialwert (WALLVALUE). Der wert eines feldes kann nicht unter eins sinken.<br>
    Anmerkung: da die wertaenderung angrenzende felder ebenfalls betrifft, muesste sich diese funktion
    aehnlich {@link #setHomepath()} rekursiv durch den gesamtem map ziehen, was aber aufgrund von zeitmangel
    nicht mehr implementiert werden konnte.
    */
    public double   value(){
        if (Home==this) return Value=HOMEVALUE;
        //if ((Corn>20)?(200-2*Pathlength<0):(Corn*10-Pathlength<0)) return Value=BADFIELD;
        byte Valuecount=0;
        double Valuesum= 0d;
        for (byte i=0;i<6;i++){
            if ((Border[i]==WALL)||
                (Border[i]==CLIFFUP)){
                    Valuecount++;
                    Valuesum+=WALLVALUE;
                }
            else if (get(i)==null) continue;
            else { Valuecount++;Valuesum+=get(i).Value;}
        }
        Valuesum=(Valuesum/Valuecount);
        Value=Valuesum+Corn*10;
        Value=(Value<1)?(1):(Value);
        return Value;
    };
    //*************************Fields*************************
    /**Anzhal der koerner auf dem aktuellen feld.*/
    public int      Corn;
    /**Unbenutzt. Relative hoehe dieses feldes zu home.*/
    public int      Height      = 10;
    /**Laenge des optimalen, bekannten heimatweges in schritten.
    Wird absichtlich mit einem viel zu grossen wert initialisiert.*/
    public int      Pathlength  = 100000;
    /**Array, in dem fuer jede Himmelsrichtung die beschaffenheit der grenze gespeichert
    wird (WALL, CLIFFUP, etc).*/
    public int []   Border      = new int [6];
    /**Unbenutzt. Zaehlt wie oft ein feld besucht wurde.*/
    public long     Hits        = 0;
    /**Wert des feldes. Je hoeher, desto besser ist das feld.*/
    public double   Value       = UNKNOWNVALUE;
    /**Flagge die signalisiert, dass feld bekannt ist, also schonmal gesehen wurde.*/
    public boolean  Known       = false;
    /**Unbenutzt. Sollte fuer rekursion der bewertungsfunktion helfen.*/
    public boolean  Valueswitch = false;
    /**Signalisiert, ob feld besucht wurde.*/
    public boolean  Visited     = false;
    /**Punkt, in dem koordinaten des Hexfelds gespeichert werden.*/
    public Hexpoint Position    =null;
    /**Angrenzendes hexfeld, von dem aus der weg nacuh hause am geringsten ist.*/
    public Hexfield Homepath    =null; 
    //**********************StaticFields**********************
    /**Unbenutzt. Zaehlt bekannte felder.*/
    public static int           KnownFields = 1;
    /**Hash, in dem alle hexfelder gespeichert werden. Koordinaten sind schluessel.*/
    public static HexfieldHash  Map         =null;
    /**Pointer auf das heimatfeld.*/
    public static Hexfield      Home        =null;
    /**Pointer auf den hamster der sich auf diesem labyrinth bewegt.*/
    public static Warhamster    Soldier     =null;
}