package warhamster40K;
import algds.c2002.*;
import java.awt.Point;
import java.util.*;
/**
Kapselung eines hexfelds. Jedes hexfeld besitzt eindeutige koordinaten. Das heimatfeld hat die koordinaten
(0,0) waehrend alle anderen koordinaten sich relativ dazu berechnen. Felder die in richtung Norden/Sueden liegen
unterscheiden sich um +/-2 von den direkt darunter/darueberliegenden. Eine schraeger Schritt bedeutet eine
abweichung von +/-1. Norden=Oben, Osten=Links, Sueden=Unten, Westen=Rechts. Die felder werden mit ihren
koordinaten als schluessel in einem hash gespeichert, und dieser hash wird statisch fuer die klasse
Hexfield gespeichert. Somit hat man ueber jedes hexfield zugriff auf die gesamte Karte. Hexfield stellt
diverse routinen zur erkundung und anderen aufgaben zur verfuegung, und wird damit zu einer art
"intelligenter karte" die von mehrerern hamstern benutzt werden kann.<br>
Der sich auf ihr bewegende hamster wird statisch fuer die klasse Hexfield gespeichert, wodurch jedes feld
weiss, welcher hamster sich auf ihm befindet, und ihm "sagen kann, was er zu tun hat" (seine methoden
benutzen kann).
*/
public class Hexfield
implements Constants, Hexfield_Constants {
    //**********************Constructors**********************
    /**Constructor fuer das heimatfeld. Setzt Koordinaten auf (0,0) setzt relative hoehe auf null,
    setzt das feld auf "bekannt", setzt heimpfadlaenge auf 0 und ordnet Heimatfeld in Plan - hash ein.
    <br>Unteraufrufe:<ul>
    <li>Hashtable.put() - methode die HexfieldHash von Hashtable geerbt hat</li>
    </ul>
    */
    Hexfield(){
        //Home = true;
        Position = new Hexpoint(0,0);
        //System.out.println("Constructing : "+this);
        Height=0;
        Known=true;
        Homepathlength=0;
        Map.put(Position,this);
        //System.out.println("Constructed  : "+this);
    }
    /**
    Constructor fuer alle nicht-Heimatfelder. Berechnet aus koordinaten von Origin und der gegangen
    richtung Direction die neuen koordinaten und ordnet das konstruierte feld in Plan-hash ein.
    Aufgrund der symmetrie des labyrinths kann davon ausgegangen werden, das ebenfalls alle felder, die
    im selben umkreis um das heimatfeld liegen, existieren. Diese werden rekursiv ebenfalls kontruiert.
    Gestoppt wird, wenn ein feld kontruiert wurde, das bereits im hash existiert. Dieses wird verworfen.
    <br>Unteraufrufe:<ul>
    <li>Hashtable.put() - methode die HexfieldHash von Hashtable geerbt hat</li>
    <li>{@link #Hexfield(Hexfield Origin, byte Direction)} - rekursive konstruktion des naechsten
        Felds auf demselben umkreis</li>
    </ul>
    */
    Hexfield(Hexfield Origin,byte Direction){
        Position = new Hexpoint(Origin.Position,Direction);
        //System.out.println("Constructing : "+this);
        if (Map.get(Position)!=null) return;
        //Homepath = new RedChain(this);
        Map.put(Position,this);
        for (byte i=0;i<6;i++)
            Border[i]=(get(i)==null)?(0):(get(i).invertBorder(Border[invert(i)]));
        if (Position.x == 0)
            if (Position.y<0) new Hexfield(this,NWEST);
            else new Hexfield(this,SEAST);
        else if ((Position.x==Position.y)&&(Position.x>0))
            new Hexfield(this,SOUTH);
        else if ((Position.x==Position.y)&&(Position.x<0))
            new Hexfield(this,NORTH);
        else if ((Position.x==-Position.y)&&(Position.x>0))
            new Hexfield(this,SWEST);
        else if ((Position.x==-Position.y)&&(Position.x<0))
            new Hexfield(this,NEAST);
        else if (Position.x>0)
            if (Position.y>Position.x)
                new Hexfield(this,SEAST);
            else if (-Position.y<Position.x) new Hexfield(this,SOUTH);
            else new Hexfield(this, SWEST);
        else if (Position.y<Position.x) new Hexfield(this,NWEST);
        else if (-Position.y>Position.x) new Hexfield(this,NORTH);
        else new Hexfield(this,NEAST);
        //System.out.println("Constructed  : "+this);
    }    
    //********************MemberFunctions*********************
    /**Unbenutzt.
    Gibt den umkreis zurueck, auf dem sich this befindet.
    <br>Unteraufrufe:<ul>
    <li>{@link #distance(Hexfield End)} - umkreis des hexfelds ist definiert durch entfernung zum Heimatfeld</li>
    </ul>
    */
    public int      circle(){
        return (Position.equals(new Point()))?(0):(distance(Home));
    }
    /**Bringt felder die bei breitensuche veraendert wurden, aber nicht auf den Pfad stack gekommen sind
    wieder in ihren ausgangszustand.
    */
    public void     cleanUp(Stack toClean){
        Hexfield Current=null;
        while(!toClean.empty()){
            Current=(Hexfield)toClean.pop();
            Current.Searchswitch=false;
            Current.Pathlength=LONGWAY;
        }
        this.Pathlength=LONGWAY;
        this.Searchswitch=false;
    }
    /**Gibt zurueck ob sich Korn auf dem feld befindet.*/
    public boolean  corn(){
        return !(Corn==0);
    }
    /**Unbenutzt. Berechnet distanz zu End in luftlinienschritten.*/
    public int      distance(Hexfield End){
        return (Math.abs(Position.x-End.Position.x)>Math.abs(Position.y-End.Position.y))?
                (Math.abs(Position.x-End.Position.x)):
                ((Math.abs(Position.x-End.Position.x)+Math.abs(Position.y-End.Position.y))/2);
    }
    /**Unbenutzt. Berechnet distanz zu feld, das zu punkt End gehoert, in luftlinienschritten.*/
    public int      distance(Point End){
        return (Math.abs(Position.x-End.x)>Math.abs(Position.y-End.y))?
                (Math.abs(Position.x-End.x)):
                ((Math.abs(Position.x-End.x)+Math.abs(Position.y-End.y))/2);
    }
    /**
    Erkundet feld mit hilfe des sich auf ihm bewegenden hamsters. Werte die gesetzt werden sind:
    Hits, Borders, Corn, angrenzende Sichtbare fleder, Visited,  Value und Homepath. Ist this das Heimatfeld,
    wird alles korn abgeworfen, und Hamster.goHome auf false gesetzt. Andernfalls wird soviel korn wie
    moeglich aufgehoben, und kornanzahl auf dem feld geupdatet. Hat hamster die backen voll wird er nach
    hause geschickt. Wurde das feld bereits einmal besucht wird der Homepath neu berechnet, das feld
    bewerted, und geschaut ob der Hamster stoppen sollte. Dann wird returned.<br>
    Wurde das feld noch nie besucht, wird die Visited-flagge gesetzt, alle angrenzenden felder und grenzen
    angeschaut und die sichtbaren auf known gesetzt und bewerted, sowie deren relative hoehendifferenz gespeichert.
    Anschließend wird das aktuelle feld bewerted und sein heimatpfad neu berechnet, sowie gecheckt,
    ob gestoppt werden sollte.<br>
    returns - this
    <br>Unteraufrufe:<ul>
    <li>{@link #setHomepath()} - berechnet Heimatpfad neu</li>
    <li>{@link warhamster40K.Warhamster40K#turn(int Direction)} - dreht Hamster zum erkunden aller felder</li>
    </ul>
    */
    public Hexfield explore(){
        if (Soldier.verbose()) Soldier.logln("Exploring: "+this);
        Hits++;
        if (Soldier.atHome()){
            Soldier.dropHere(Soldier.getCarriedCorn());
            Soldier.GoHome=false;
            Soldier.Steps=0;
        }
        else{
            if (Soldier.take(CORNCARRY_MAX)!=0) Soldier.Steps=0;
            Corn = Soldier.getAvailableCorn();
        }
        if (Soldier.getCarriedCorn()==CORNCARRY_MAX) Soldier.GoHome=true;
        if (Visited){
            setHomepath();
            if (Soldier.Steps+Homepathlength>199+10*Soldier.getCarriedCorn())
                Soldier.Stop=true;
            return this;
        }
        Visited=true;
        for (byte i=0;i<6;i++){
            Soldier.turn(TURN_RIGHT);
            Border[Soldier.Looking]=(Soldier.lookForward()&WALLMASK);
            if (Border[Soldier.Looking]==WALL) continue;
            if (get(Soldier.Looking)==null) new Hexfield(this,Soldier.Looking);
            if (get(Soldier.Looking).Visited) continue;
            if ((Border[Soldier.Looking]==SLOPEUP)||
                (Border[Soldier.Looking]==CLIFFUP))
                get(Soldier.Looking).Height=Height+Soldier.getHeightDifference();
            else{
                get(Soldier.Looking).Height=Height+Soldier.getHeightDifference();
                get(Soldier.Looking).Corn=((Soldier.lookForward() & CORNMASK)==CORN)?
                                            (20):(0);
                get(Soldier.Looking).Known=true;
            }
        }
        setHomepath();
        if (Soldier.Steps+Homepathlength>199+10*Soldier.getCarriedCorn())
            Soldier.Stop=true;
        if (Soldier.verbose()) Soldier.logln("Explored : "+this);
        return this;
    }
    /**Routine um Hexfeld mit argument als Koordinaten zu suchen.*/
    public Hexfield get(Point ToFind){
        return Map.get(ToFind);
    }
    /**Routine, um Hexfeld mit Koordinaten (x,y) zu suchen.*/
    public Hexfield get(int x,int y){
        return Map.get(x,y);
    }
    /**Routine um hexfeld zu suchen, welches vom aktuellen aus einen schritt in richtung Direction liegt.
    <br>Unteraufrufe:<ul>
    <li>{@link #get(Point)} - neu konstruierter Hexpoint wird benutzt um feld zu erhalten</li>
    </ul>
    */
    public Hexfield get(byte Direction){
        return get(new Hexpoint(Position,Direction));
    }
    /**Invertiert Direction. Aus NORTH wird SOUTH, etc.*/
    public byte     invert(byte Direction){
        switch(Direction){
        case 0: return 3;
        case 1: return 4;
        case 2: return 5;
        case 3: return 0;
        case 4: return 1;
        case 5: return 2;
        }
        return -1;
    }
    /**Invertiert Walltype. Aus CLIFFUP wird CLIFFDOWN, etc. WALL bleibt WALL.*/
    public int      invertBorder(int Walltype){
        if (Walltype==WALL) return WALL;
        if (Walltype==CLIFFUP) return CLIFFDOWN;
        if (Walltype==CLIFFDOWN) return CLIFFUP;
        if (Walltype==SLOPEUP) return SLOPEDOWN;
        if (Walltype==SLOPEDOWN) return SLOPEUP;
        return 0;
    }
    /**Returned Hexfeld in Nordostrichtung.*/
    public Hexfield NEast(){
        return Map.get(new Hexpoint(Position,NEAST));
    }
    /**Returned Hexfeld in Nordrichtumg.*/
    public Hexfield North(){
        return Map.get(new Hexpoint(Position,NORTH));
    }
    /**Returned Hexfeld in Nordwestrichtumg.*/
    public Hexfield NWest(){
        return Map.get(new Hexpoint(Position,NWEST));
    }
    /**Returned Hexfeld in Suedostrichtung.*/
    public Hexfield SEast(){
        return Map.get(new Hexpoint(Position,SEAST));
    }
    /**Returned Hexfeld in Suedrichtung.*/
    public Hexfield South(){
        return Map.get(new Hexpoint(Position,SOUTH));
    }
    /**Returned Hexfeld in Suedwestrichtung.*/
    public Hexfield SWest(){
        return Map.get(new Hexpoint(Position,SWEST));
    }
    /**Sucht per breitensuche das am kuerzesten entfernte, relevante feld.
    Relevante felder sind:<ul>
    <li>Unbesuchte felder</li>
    <li>Felder die genug korn enthalten, um weg zu rechtfertigen</li>
    <li>Heimatfeld, wenn getragenes korn weg rechtfertigt.</li>
    </ul>
    Felder hinter klippe werden nur bei Risk==true bereucksichtigt.
    @return Stack, auf dem Pfad zum gesuchten feld liegt, oder ein leerer stack, wenn kein feld gefunden wurde.
    */
    public Stack    SearchNext(){
        if (Soldier.verbose())
            Soldier.logln("Searching next field, starting from: "+this);
        LinkedList  Queue   =new LinkedList();
        Stack       Cleanup =new Stack();
        Stack       Path    =new Stack();
        Hexfield    Current =this;
        Current.Pathlength=0;
        Current.Searchswitch=true;
        do{
            for(byte i=0;i<6;i++){
                if ((Current.get(i)==null)||
                    Current.get(i).Searchswitch||
                    (Current.Border[i]==WALL)||
                    (Current.Border[i]==CLIFFUP)||
                    ((Current.Border[i]==CLIFFDOWN)&&!Current.get(i).Visited&&!Soldier.Risk)
                ) continue;
                Current.get(i).Searchswitch=true;
                Current.get(i).Pathlength=Current.Pathlength+1;
                Queue.add(Current.get(i));
                //if (Soldier.verbose())
                //    Soldier.logln("Added to Queue: "+Current.get(i)+":"+Current.get(i).Pathlength);
            }
            if (Queue.size()==0){
                cleanUp(Cleanup);
                return Path;
            }
            Current=(Hexfield)Cleanup.push(Queue.removeFirst());
        }while ((Current.Visited)&& //besuchte Feldr kommen nicht in frage
                ((Current!=Home)||((Current==Home)&&(Soldier.getCarriedCorn()*10<Homepathlength)))&&
                //Heimatfeld kommt nur in frage, wenn getragenes Korn weg rechtfertigt
                (((Current.Corn>20)?(200):(Current.Corn*10)) 
                 <=Current.Pathlength+Current.Homepathlength)
                //Feld kommt nur in frage, wenn enthaltenes korn weg rechtfertigt.
                );
        do{
            byte i=0;
            for(i=0;i<6;i++)
                if (!((Current.get(i)==null)||
                      (Current.Border[i]==WALL)||
                      ((Current.Border[i]==CLIFFDOWN))||
                      (Current.get(i).Border[invert(i)]==WALL)||
                      ((Current.get(i).Border[invert(i)]==CLIFFUP))
                     )&&
                    (Current.get(i).Pathlength+1==Current.Pathlength))
                    break;
            Path.push(Current);
            //if (Soldier.verbose())
            //    Soldier.logln("Pushed onto path: "+Current+":"+Current.Pathlength+", i="+i);
            Current=Current.get(i);
        }while(Current.Pathlength>0);
        while(Queue.size()!=0) Cleanup.push(Queue.removeFirst());
        cleanUp(Cleanup);
        this.Pathlength=LONGWAY;
        this.Searchswitch=false;
        return Path;
    }
    /**Returned Stringrepraesentation des this-Hexfeldes.*/
    public String   toString(){
        return "Hexfield("+Position.x+","+Position.y+",C"+circle()+","+Visited+")";
    }
    /**Validiert heimatpfad des aktuellen feldes und setzt ihn wenn noetig neu.
    Es werden alle heimatpfadlaengen der angrenzenden besuchten felder gecheckt und die kuerzeste
    gespeichert (w). Ist pfadlaenge des aktuellen feldes groesser als w+1, so wird sie neu berechnet
    und {@link #Homepath} auf ein besseres, angrenzendes feld gesetzt, und die heimatpfade aller
    angrezenden felder werden neu berechnet. Ist die aktuelle Pfadlange gleich w+1 passiert nichts.
    Das ist auch die abbruchbedingung fuer die rekursion.
    <br>Unteraufrufe:<ul>
    <li>{@link #get(byte Direction)} - wird benutzt um angrenzendes feld zu erhalten</li>
    <li>{@link #setHomepath()} - rekursive neuberechnung der heimatpfade angrenzender felder</li>
    </ul>
         */
    public int      setHomepath(){
        if (this==Home) return 0;
        int Prepath=Homepathlength;
        for (byte i=0;i<6;i++){
            if ((get(i)==null)||
                (Border[i]==WALL)||
                (Border[i]==CLIFFUP)||
                (!get(i).Visited)) continue;
            if (get(i).Homepathlength+1<Homepathlength){
                Homepath=get(i);Homepathlength=get(i).Homepathlength+1;
            }
        }
        if (Prepath<Homepathlength) throw new Error("Homepath set wrong!");
        if (Prepath!=Homepathlength)
            for (byte i=0;i<6;i++)
                if ((get(i)==null)||
                    (Border[i]==WALL)||
                    (Border[i]==CLIFFUP)||
                    (!get(i).Visited)) continue;
                else {get(i).setHomepath();Soldier.Risk=false;}
        return Homepathlength;
    }
    //*************************Fields*************************
    //boolean Home = false;
    /**Anzhal der koerner auf dem aktuellen feld.*/
    public int      Corn;
    /**Unbenutzt. Relative hoehe dieses feldes zu home.*/
    public int      Height          = 10;
    /**Laenge des optimalen, bekannten heimatweges in schritten.
    Wird absichtlich mit einem viel zu grossen wert initialisiert.*/
    public int      Homepathlength  = LONGWAY;
    /**Laenge des pfades von diesem feld zum zu erreichenden feld.*/
    public int      Pathlength      = LONGWAY;
    /**Array, in dem fuer jede Himmelsrichtung die beschaffenheit der grenze gespeichert
    wird (WALL, CLIFFUP, etc).*/
    public int []   Border          = {0,0,0,0,0,0};
    /**Unbenutzt. Zaehlt wie oft ein feld besucht wurde.*/
    public long     Hits            = 0;
    /**Flagge die signalisiert, dass feld bekannt ist, also schonmal gesehen wurde.*/
    public boolean  Known           = false;
    /**Signalisiert, dass feld bereits bei der suche SearchNext() beruecksichtigt wurde,
    und nicht abgearbeitet werden muss.*/
    public boolean  Searchswitch    = false;
    /**Signalisiert, ob feld besucht wurde.*/
    public boolean  Visited         = false;
    /**Punkt, in dem koordinaten des Hexfelds gespeichert werden.*/
    public Hexpoint Position        = null;
    /**Angrenzendes hexfeld, von dem aus der weg nach hause am geringsten ist.*/
    public Hexfield Homepath        = null; 
    //**********************StaticFields**********************
    /**Unbenutzt. Zaehlt bekannte felder.*/
    public static int           KnownFields = 1;
    /**Hash, in dem alle hexfelder gespeichert werden. Koordinaten sind schluessel.*/
    public static HexfieldHash  Map         = null;
    /**Pointer auf das heimatfeld.*/
    public static Hexfield      Home        = null;
    /**Pointer auf den hamster der sich auf diesem labyrinth bewegt.*/
    public static Warhamster40K Soldier     = null;
}