package warhamster40K;
import algds.c2002.*;
import java.awt.*;
import java.util.*;
/**
Strategie<br>
Warhamster40K benutzt fast identische Klassen wie {@link warhamster.Warhamster Warhamster} und
{@link carson.Carson Carson} allerdings eine voellig andere strategie. Um zu entscheiden wo er sich
hinbewegen soll, sucht er, ausgehend von seinem derzeitigen feld, per breitensuche das naechste feld,
auf welchem korn liegt, welches gesehen aber noch nicht besucht wurde und welches nicht hinter einer
klippe abwaerts liegt.<br>
Hierzu wurde die klasse {@link warhamster40K.Hexfield Hexfield} extra um die
Methode {@link warhamster40K.Hexfield#SearchNext() SearchNext()} erweitert, die nach solch einem Feld
sucht, einen Stack, auf dem der weg dorthin liegt zurueckgibt.<br>
Das so gefundene feld ist aufgrund des charakteristikums der breitensuche automatisch das am
kuerzesten entfernteste. Ist dies das heimatfeld, wird der weg dorthin nur angetreten wenn das zur
zeit getragene korn ihn rechtfertigt. Es werden auch nur felder berucksichtigt, die unbesucht sind,
oder deren Korninhalt den weg rechtfertigt.
Wird kein solches feld gefunden wird Risk eingeschaltetm, was bedeutet, das der Hamster auch felder
beruecksichtigen darf, die hinter einer Klippe abwaerts liegen. Im risk modus kann er nicht nach hause
gehen, da er den heimatweg nicht kennt. Der risk modus wird erst wieder abgeschaltet, sobald er auf ein
feld trifft, von dem aus er den heimatweg kennt, da er durch dieses, dann auch den heimatweg von allen
feldern, die er vorher betreten hat rekursiv durch setHomepath rekonstruiert. Um dies abzusichern wird
die Heimatpfadlaenge bei feldern, fuer die man keinen heimatpfad festlegen kann auf einen unmoeglich
hohen wert gesetzt.<br>
Das setzen und abarbeiten der heimatpfadlaenge erfolgt durch {@link warhamster40K.Hexfield#setHomepath()}
Es ist identisch mit der variante aus {@link warhamster.Warhamster Warhamster} und {@link carson.Carson Carson}.<br>
Warhamster40K stoppt, wenn auch im Risk modus von {@link #Home} aus keine relevanten felder gefunden werden,
da dann keine relevanten felder mehr im labyrinth existieren.
*/
public class Warhamster40K
extends 	Hamster2
implements 	Constants,Hexfield_Constants{
    //********************MemberFunctions*********************
    /**siehe {@link warhamster.Warhamster#forward()}.*/
    public boolean  forward(){
        if ((Current.Border[Looking]==WALL)||
            (Current.Border[Looking]==CLIFFUP))
            return false;
        if (!super.forward()) throw new Error("Forward buggy!");
        Steps++;
        //if (verbose()) logln("Stepped forward towards: "+Current);
        Current=Current.get(Looking).explore();
        return true;
    }
    /**Selbsterklaerend.*/
    public String   getHamsterAuthor() {return "HerbaCasipulensisGigantis und des Pudels Kern";}
    /**Selbsterklaerend.*/
    public String   getHamsterVersion(){return "0.000000... RC1";}
    /**siehe {@link warhamster.Warhamster#goHome()}.*/
    public boolean  goHome(){
        GoHome=true;
        if (Current.Homepath==null) return false;
        while (!atHome()) stepTowards(Current.Homepath);
        return true;
    }
    /**
    Herzstueck des hamsters. Initialiesiert alle wichtigen noch nicht initialisierten variablen,
    welche nicht bei instanziierung initialisiert werden.Arbeitsweise der haptnachrichtenschleife in
    Strategieerklaerung erlaeutert.<br>
    */
    public void     run(){
    try{
        Hexfield.Map    = new HexfieldHash(3000,.95f);
        Hexfield.Home   = new Hexfield();
        Hexfield.Soldier= this;
        Current         = Hexfield.Home;
        Home            = Hexfield.Home;
        Map             = Current.Map;
        Current.explore();
        
        do{
            if (Path.empty()) Path=Current.SearchNext();
            if (Path.empty()) {Risk=true;Path=Current.SearchNext();}
            if (Path.empty()&&atHome()) break;
            if (Path.empty()) goHome();
            while (!Path.empty()){
                if (!stepTowards((Hexfield)Path.pop()))
                    throw new Error("SearchNext buggy!");
            }
            if (GoHome||Stop) goHome();
            if (!atHome()) Stop=false;
        }while(!Stop);
        //while ((Current!=Home));
    }
    catch (Throwable X){
       logln("\nRun :"+X.getClass()+" thrown. ->"+X.getMessage());
       X.printStackTrace();
       Stop=true;goHome();
    }
    }
    /**siehe {@link warhamster.Warhamster#stepTowards(Hexfield toGoTo)}.*/
    public boolean  stepTowards(Hexfield toGoTo){
        int Direction;
        byte i=0;
        if (verbose()) logln("Stepping towards: "+toGoTo);
        for (i=0;i<6;i++){
            if (Current.get(i)==toGoTo) break;
            if (i==5) throw new Error("Argument: "+toGoTo+" unreachable from: "+Current);
        }
        for (i=0;i<6;i++){
            turn(TURN_RIGHT);
            if (Current.get(Looking)==toGoTo) break;
        }
        if (!forward()){
            logln("Bug occured while going towards: "+toGoTo);
            return false;
        }
        return true;
    }
    /**siehe {@link warhamster.Warhamster#turn(int Direction)}.*/
    public void     turn(int Direction){
        //if (verbose()) logln("Turning towards: "+Direction+"\nLooking before turn: "+Looking);
        switch(Looking){
        case NORTH:
            if (Direction==TURN_LEFT) Looking=6;
            break;
        case NWEST:
            if (Direction==TURN_RIGHT) Looking=-1;
            break;
        }
        if (Direction==TURN_LEFT) --Looking;
        else ++Looking;
        super.turn(Direction);
        //if (verbose()) logln("Looking after  turn: "+Looking);
    }

    //*************************Fields*************************
    /**Speichert himmelsrichtung, in welche hamster gerade schaut.*/
    public byte         Looking = NORTH;
    /**Zaehlt schritte seit das letzte korn aufgehoben wurde, oder heimatfeld besucht wurde.*/
    public long         Steps   = 0;
    /**Speichert pfad zum naechsten feld.*/
    public Stack        Path    = new Stack();
    /**Flagge die gesetzt wird, um Warhamster40K nach hause zu schicken.*/
    public boolean      GoHome  = false;
    /**Flagge wird gesetzt, um Risk modus einzuschalten. Hamster springt nur im risk modues klippen runter.*/
    public boolean      Risk    = false;
    /**Flagge die gesetzt wird, um Warhamster40K zu stoppen.*/
    public boolean      Stop    = false;
    /**Speichert hexfeld, auf dem sich der hamster gerade befindet.*/
    public Hexfield     Current = null;
    /**Speichert heimatfeld.*/
    public Hexfield     Home    = null;
    /**Hash, in dem sich alle hexfelder befinden, von denen gewiss ist, dass sie existieren.
    Als schluessel werden punkte mit Koordinaten des Hexfelds benutzt.*/
    public HexfieldHash Map     = null;
}